# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import hashlib, re

from htmlentitydefs import name2codepoint

from threading import Lock

from threading import Thread as BaseThread



STATUS_KEY = u'status'


ENCODING = u'UTF-8'



def md5sum(url):
    md5 = hashlib.md5()

    md5.update(url)

    return md5.hexdigest()


def split(text, pattern):
    if not text:
        return []

    words = re.split(pattern, text, flags=re.UNICODE)

    if not words[-1]:
        words.pop()

    if not words[0]:
        words.pop(0)

    return words


def escape(text):
    start = 0

    slices = []

    while True:
        ampersand_index = text.find(u'&', start)

        if ampersand_index == -1:
            slices.append(text[start:])

            break

        slices.append(text[start:ampersand_index])

        semicolon_index = text.find(u';', ampersand_index + 1)

        if semicolon_index != -1:
            name = text[(ampersand_index + 1):semicolon_index]

            if name in [u'quot', u'amp', u'lt', u'gt']:
                slices.append(text[ampersand_index:(semicolon_index + 1)])

                start = semicolon_index + 1

                continue

            if name.startswith(u'#'):
                try:
                    codepoint = long(name[1:])
                except ValueError:
                    codepoint = None
            else:
                try:
                    codepoint = name2codepoint[name]
                except KeyError:
                    codepoint = None

            if codepoint is not None:
                try:
                    char = unichr(codepoint)
                except (ValueError, OverflowError):
                    pass
                else:
                    slices.append(char)

                    start = semicolon_index + 1

                    continue

        slices.append(u'&amp;')

        start = ampersand_index + 1

    return u''.join(slices)


def rescape(text):
    start = 0

    slices = []

    while True:
        ampersand_index = text.find(u'&', start)

        if ampersand_index == -1:
            slices.append(text[start:])

            break

        slices.append(text[start:ampersand_index])

        slices.append(u'&amp;')

        start = ampersand_index + 1

    return u''.join(slices)


def unescape(text):
    start = 0

    slices = []

    while True:
        ampersand_index = text.find(u'&', start)

        if ampersand_index == -1:
            slices.append(text[start:])

            break

        slices.append(text[start:ampersand_index])

        semicolon_index = text.find(u';', ampersand_index + 1)

        if semicolon_index != -1:
            name = text[(ampersand_index + 1):semicolon_index]

            if name == u'quot':
                char = u'"'
            elif name == u'amp':
                char = u'&'
            elif name == u'lt':
                char = u'<'
            elif name == u'gt':
                char = u'>'
            else:
                char = None

            if char is not None:
                slices.append(char)

                start = semicolon_index + 1

                continue

        slices.append(u'&')

        start = ampersand_index + 1

    return u''.join(slices)


def morescape(text):
    start = 0

    end = 0

    slices = []

    while end < len(text):
        char = text[end]

        if char == u'"':
            name = u'quot'
        elif char == u'<':
            name = u'lt'
        elif char == u'>':
            name = u'gt'
        else:
            name = None

        if name is not None:
            slices.append(text[start:end])

            slices.append(u'&{};'.format(name))

            start = end + 1

        end += 1

    if start < len(text):
        slices.append(text[start:end])

    return u''.join(slices)


def dispatch(callback, *args):
    thread = Thread(callback, *args)

    thread.start()

    return thread



class Thread(BaseThread):
    def __init__(self, callback, *args):
        super(Thread, self).__init__()

        self.callback = callback

        self.args = args

        self.daemon = True


    def run(self):
        self.callback(*self.args)

        self.args = None

        self.callback = None



class Notifier(object):
    def __init__(self):
        self.callbacks = {}

        self.lock = Lock()


    def connect(self, callback):
        uid = id(callback)

        self.lock.acquire()

        self.callbacks[uid] = callback

        self.lock.release()

        return uid


    def disconnect(self, uid):
        self.lock.acquire()

        del self.callbacks[uid]

        self.lock.release()


    def notify(self):
        for callback in self.callbacks.viewvalues():
            callback()



class Updatable(object):
    def _acquire_update(self):
        self.update_lock.acquire()

        self.block_lock.acquire()
 
        self.block_lock.release()


    def _release_update(self):
        self.update_lock.release()


    def __init__(self):
        self.block_count = 0

        self.block_count_lock = Lock()

        self.block_lock = Lock()

        self.update_lock = Lock()


    def block_updates(self):
        self.update_lock.acquire()

        self.block_count_lock.acquire()

        previous = self.block_count

        self.block_count += 1

        self.block_count_lock.release()

        if previous == 0:
            self.block_lock.acquire()

        self.update_lock.release()


    def allow_updates(self):
        self.block_count_lock.acquire()

        self.block_count -= 1

        current = self.block_count

        self.block_count_lock.release()

        if current == 0:
            self.block_lock.release()



class BaseCache(object):
    def __init__(self):
        self.cache = {}

        self.lock = Lock()



import external, setting, proxy, shortener, iconset, twitter, gui
