# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from __future__ import division


import os

from urllib import urlretrieve


from glib import GError

from gtk import gdk as Gdk

from socks import ProxyError


from polly import ENCODING, md5sum, dispatch, Notifier, Updatable, BaseCache

from polly.twitter.lead import user

from polly.gui.basic import activate_pixbuf, resize_pixbuf, transparent_pixbuf, FilesystemPool



class Framebuffer(object):
    def __init__(self, animation, size):
        timestamp = 0.001

        iter = animation.get_iter(timestamp)

        self.pixbufs = []

        self.delays = []

        while True:
            pixbuf = iter.get_pixbuf()

            if pixbuf in self.pixbufs:
                break

            self.pixbufs.append(pixbuf)

            delay = iter.get_delay_time()

            self.delays.append(delay)

            timestamp += (delay / 1000) + 0.001

            iter.advance(timestamp)

        self.length = len(self.pixbufs)

        self.pixbufs_active = []

        for index in range(self.length):
            self.pixbufs[index] = resize_pixbuf(self.pixbufs[index], size)

            self.pixbufs_active.append(activate_pixbuf(self.pixbufs[index]))



class Backend(Updatable):
#    def __del__(self):
#        print u'debug: avatar backend deleted'
    def _update_url(self):
        Gdk.threads_enter()

        normal_url = self.permanent_url

        Gdk.threads_leave()

        if self.normal_url == normal_url:
            return False

        self.normal_url = normal_url

        self.normal_hash = u'{}-{}'.format(self.user_backend.id, md5sum(self.normal_url))

        self.normal_path = os.path.join(self.cache_path, self.normal_hash)

        index = self.normal_url.rfind(u'.')

        if index < self.normal_url.rfind(u'/'):
            index = len(self.normal_url)

        self.bigger_url = u''.join([self.normal_url[:(index - 7)], u'_bigger', self.normal_url[index:]])

        self.bigger_hash = u'{}-{}'.format(self.user_backend.id, md5sum(self.bigger_url))

        self.bigger_path = os.path.join(self.cache_path, self.bigger_hash)

        return True


    def _update_size(self):
        Gdk.threads_enter()

        size = self.permanent_size

        Gdk.threads_leave()

        if self.size == size:
            return False

        self.size = size

        return True


    def _update_pixbufs(self):
        if self.size > 48:
            url = self.bigger_url

            hash = self.bigger_hash

            path = self.bigger_path
        else:
            url = self.normal_url

            hash = self.normal_hash

            path = self.normal_path

        self.filesystem_pool.acquire(hash)

        success = True

        if not os.path.exists(path):
            try:
                self.cache_path_lock.acquire()

                try:
                    if not os.path.isdir(self.cache_path):
                        os.makedirs(self.cache_path)
                finally:
                    self.cache_path_lock.release()

                self.connection_pool.acquire()

                try:
                    urlretrieve(url.encode(ENCODING), path)
                finally:
                    self.connection_pool.release()
            except (OSError, IOError, ProxyError):
                success = False

        Gdk.threads_enter()

        if success:
            try:
                animation = Gdk.PixbufAnimation(path)

                self.pixbuf = resize_pixbuf(animation.get_static_image(), self.size)

                self.pixbuf_active = activate_pixbuf(self.pixbuf)

                if animation.is_static_image():
                    self.framebuffer = None
                else:
                    self.framebuffer = Framebuffer(animation, self.size)

            except GError:
                success = False

                try:
                    os.remove(path)
                except OSError:
                    pass

        if not success:
            self.pixbuf = transparent_pixbuf(self.size)

            self.pixbuf_active = self.pixbuf.copy()

            self.framebuffer = None

        Gdk.threads_leave()

        self.filesystem_pool.release(hash)


    def _update(self, previous_thread, callback):
        previous_thread.join()

        self._acquire_update()

        if callback is None:
            success = True
        else:
            success = callback()

        if success:
            self._update_pixbufs()

            self.notifier.notify()

        self._release_update()


    def _first_update(self):
        self.normal_url = None

        self.size = None

        self._update_url()

        self._update_size()

        self._update_pixbufs()


    def _update_permanent_url(self):
        self.permanent_url = self.user_backend.data[user.IMAGE_URL_KEY]


    def _update_permanent_size(self):
        unit = self.dimension * self.font_scheme.unit

        void = self.spacing * self.font_scheme.void

        self.permanent_size = unit + void


    def _update_permanent(self, permanent_callback, callback):
        Gdk.threads_enter()

        permanent_callback()

        self.thread = dispatch(self._update, self.thread, callback)

        Gdk.threads_leave()


    def _on_user_image_url_changed(self):
        self._update_permanent(self._update_permanent_url, self._update_url)


    def _on_font_changed(self):
        self._update_permanent(self._update_permanent_size, self._update_size)


    def __init__(self, user_backend, dimension, spacing, cache_path, cache_path_lock, font_scheme, filesystem_pool, connection_pool, backend_cache):
        super(Backend, self).__init__()

        self.user_backend = user_backend

        self.dimension = dimension

        self.spacing = spacing

        self.cache_path = cache_path

        self.cache_path_lock = cache_path_lock

        self.font_scheme = font_scheme

        self.filesystem_pool = filesystem_pool

        self.connection_pool = connection_pool

        self.backend_cache = backend_cache

        self.reference_count = 0

        self.notifier = Notifier()

        self.user_backend.block_updates()

        self.font_scheme.block_updates()

        self._update_permanent_url()

        self._update_permanent_size()

        self.first_thread = dispatch(self._first_update)

        self.thread = self.first_thread

        self.user_image_url_uid = self.user_backend.image_url_notifier.connect(self._on_user_image_url_changed)

        self.font_uid = self.font_scheme.notifier.connect(self._on_font_changed)

        self.user_backend.allow_updates()

        self.font_scheme.allow_updates()


    def isolate(self):
        self.font_scheme.block_updates()

        self.user_backend.block_updates()

        self.font_scheme.notifier.disconnect(self.font_uid)

        self.user_backend.image_url_notifier.disconnect(self.user_image_url_uid)

        self.font_scheme.allow_updates()

        self.user_backend.allow_updates()

        self.thread.join()


    def reload(self):
        Gdk.threads_enter()

        self.thread = dispatch(self._update, self.thread, None)

        thread = self.thread

        Gdk.threads_leave()

        return thread


    def release(self):
        self.backend_cache.release(self)



class BackendCache(BaseCache):
    def __init__(self, cache_path, cache_path_lock, font_scheme, connection_pool):
        super(BackendCache, self).__init__()

        self.cache_path = os.path.join(cache_path, u'avatars')

        self.cache_path_lock = cache_path_lock

        self.font_scheme = font_scheme

        self.connection_pool = connection_pool

        self.filesystem_pool = FilesystemPool()


    def reload(self):
        threads = []

        self.lock.acquire()

        for user_id in self.cache.viewkeys():
            for dimension in self.cache[user_id].viewkeys():
                for backend in self.cache[user_id][dimension].viewvalues():
                    threads.append(backend.reload())

        self.lock.release()

        for thread in threads:
            thread.join()


    def acquire(self, user_backend, dimension, spacing):
        user_id = user_backend.id

        self.lock.acquire()

        if user_id not in self.cache.viewkeys():
            self.cache[user_id] = {}

        if dimension not in self.cache[user_id].viewkeys():
            self.cache[user_id][dimension] = {}

        if spacing not in self.cache[user_id][dimension].viewkeys():
            self.cache[user_id][dimension][spacing] = Backend(user_backend, dimension, spacing, self.cache_path, self.cache_path_lock, self.font_scheme, self.filesystem_pool, self.connection_pool, self)

        backend = self.cache[user_id][dimension][spacing]

        backend.reference_count += 1

        self.lock.release()

        return backend


    def release(self, backend):
        self.lock.acquire()

        backend.reference_count -= 1

        if backend.reference_count == 0:
            user_id = backend.user_backend.id

            dimension = backend.dimension

            del self.cache[user_id][dimension][backend.spacing]

            if not self.cache[user_id][dimension]:
                del self.cache[user_id][dimension]

            if not self.cache[user_id]:
                del self.cache[user_id]

            backend.isolate()

        self.lock.release()
