# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from glib import GError

from gtk import gdk as Gdk


from polly import Notifier, Updatable, BaseCache

from polly.gui.basic import transparent_pixbuf



class Backend(Updatable):
    def _update_paths(self):
        self.paths = []

        for string in self.iconset_controller.backend.strings:
            self.paths.append(string.format(self.name))


    def _update_size(self):
        self.size = self.font_scheme.unit


    def _update_pixbuf(self):
        Gdk.threads_enter()

        self.broken = True

        for path in self.paths:
            try:
                self.pixbuf = Gdk.pixbuf_new_from_file_at_size(path, self.size, self.size)
            except GError:
                pass
            else:
                self.broken = False

                break

        if self.broken:
            self.pixbuf = transparent_pixbuf(self.size)

        Gdk.threads_leave()


    def _update(self, callback):
        self._acquire_update()

        callback()

        self._update_pixbuf()

        self.notifier.notify()

        self._release_update()


    def _on_iconset_changed(self):
        self._update(self._update_paths)


    def _on_font_changed(self):
        self._update(self._update_size)


    def __init__(self, name, iconset_controller, font_scheme):
        super(Backend, self).__init__()

        self.name = name

        self.iconset_controller = iconset_controller

        self.font_scheme = font_scheme

        self.notifier = Notifier()

        self.iconset_controller.backend.block_updates()

        self.font_scheme.block_updates()

        self._update_paths()

        self._update_size()

        self._update_pixbuf()

        self.iconset_uid = self.iconset_controller.backend.notifier.connect(self._on_iconset_changed)

        self.font_uid = self.font_scheme.notifier.connect(self._on_font_changed)

        self.iconset_controller.backend.allow_updates()

        self.font_scheme.allow_updates()


    def isolate(self):
        self.font_scheme.block_updates()

        self.iconset_controller.backend.block_updates()

        self.font_scheme.notifier.disconnect(self.font_uid)

        self.iconset_controller.backend.notifier.disconnect(self.iconset_uid)

        self.font_scheme.allow_updates()

        self.iconset_controller.backend.allow_updates()



class BackendCache(BaseCache):
    def __init__(self, iconset_controller, font_scheme):
        super(BackendCache, self).__init__()

        self.iconset_controller = iconset_controller

        self.font_scheme = font_scheme


    def clear(self):
        self.lock.acquire()

        for backend in self.cache.viewvalues():
            backend.isolate()

        self.cache.clear()

        self.lock.release()


    def acquire(self, name):
        self.lock.acquire()

        if name not in self.cache.viewkeys():
            self.cache[name] = Backend(name, self.iconset_controller, self.font_scheme)

        backend = self.cache[name]

        self.lock.release()

        return backend
