# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import os

from urllib import urlretrieve


from glib import GError

from gtk import gdk as Gdk

from socks import ProxyError


from polly import ENCODING, md5sum, dispatch, Notifier, Updatable, BaseCache

from polly.twitter.lead import user

from polly.gui.basic import resize_pixbuf, transparent_pixbuf



class Backend(Updatable):
#    def __del__(self):
#        print u'debug: identifier backend deleted'
    def _update_url(self):
        Gdk.threads_enter()

        url = self.permanent_url

        Gdk.threads_leave()

        if self.url == url:
            return False

        self.url = url

        self.hash = u'{}-{}'.format(self.user_backend.id, md5sum(self.url))

        self.path = os.path.join(self.cache_path, self.hash)

        return True


    def _update_size(self):
        Gdk.threads_enter()

        size = self.permanent_size

        Gdk.threads_leave()

        if self.size == size:
            return False

        self.size = size

        return True


    def _update_pixbuf(self):
        success = True

        if not os.path.exists(self.path):
            try:
                self.cache_path_lock.acquire()

                try:
                    if not os.path.isdir(self.cache_path):
                        os.makedirs(self.cache_path)
                finally:
                    self.cache_path_lock.release()

                self.connection_pool.acquire()

                try:
                    urlretrieve(self.url.encode(ENCODING), self.path)
                finally:
                    self.connection_pool.release()
            except (OSError, IOError, ProxyError):
                success = False

        Gdk.threads_enter()

        if success:
            try:
                animation = Gdk.PixbufAnimation(self.path)

                self.pixbuf = resize_pixbuf(animation.get_static_image(), self.size)
            except GError:
                success = False

                try:
                    os.remove(self.path)
                except OSError:
                    pass

        if not success:
            self.pixbuf = transparent_pixbuf(self.size)

        Gdk.threads_leave()


    def _update(self, previous_thread, callback):
        previous_thread.join()

        self._acquire_update()

        if callback is None:
            success = True
        else:
            success = callback()

        if success:
            self._update_pixbuf()

            self.notifier.notify()

        self._release_update()


    def _first_update(self):
        self.url = None

        self.size = None

        self._update_url()

        self._update_size()

        self._update_pixbuf()


    def _update_permanent_url(self):
        self.permanent_url = self.user_backend.data[user.IMAGE_URL_KEY]


    def _update_permanent_size(self):
        self.permanent_size = self.font_scheme.unit


    def _update_permanent(self, permanent_callback, callback):
        Gdk.threads_enter()

        permanent_callback()

        self.thread = dispatch(self._update, self.thread, callback)

        Gdk.threads_leave()


    def _on_user_image_url_changed(self):
        self._update_permanent(self._update_permanent_url, self._update_url)


    def _on_font_changed(self):
        self._update_permanent(self._update_permanent_size, self._update_size)


    def __init__(self, user_backend, cache_path, cache_path_lock, font_scheme, connection_pool, backend_cache):
        super(Backend, self).__init__()

        self.user_backend = user_backend

        self.cache_path = cache_path

        self.cache_path_lock = cache_path_lock

        self.font_scheme = font_scheme

        self.connection_pool = connection_pool

        self.backend_cache = backend_cache

        self.reference_count = 0

        self.notifier = Notifier()

        self.user_backend.block_updates()

        self.font_scheme.block_updates()

        self._update_permanent_url()

        self._update_permanent_size()

        self.first_thread = dispatch(self._first_update)

        self.thread = self.first_thread

        self.user_image_url_uid = self.user_backend.image_url_notifier.connect(self._on_user_image_url_changed)

        self.font_uid = self.font_scheme.notifier.connect(self._on_font_changed)

        self.user_backend.allow_updates()

        self.font_scheme.allow_updates()


    def isolate(self):
        self.font_scheme.block_updates()

        self.user_backend.block_updates()

        self.font_scheme.notifier.disconnect(self.font_uid)

        self.user_backend.image_url_notifier.disconnect(self.user_image_url_uid)

        self.font_scheme.allow_updates()

        self.user_backend.allow_updates()

        self.thread.join()


    def reload(self):
        Gdk.threads_enter()

        self.thread = dispatch(self._update, self.thread, None)

        thread = self.thread

        Gdk.threads_leave()

        return thread


    def release(self):
        self.backend_cache.release(self)



class BackendCache(BaseCache):
    def __init__(self, cache_path, cache_path_lock, font_scheme, connection_pool):
        super(BackendCache, self).__init__()

        self.cache_path = os.path.join(cache_path, u'identifiers')

        self.cache_path_lock = cache_path_lock

        self.font_scheme = font_scheme

        self.connection_pool = connection_pool


    def reload(self):
        threads = []

        self.lock.acquire()

        for backend in self.cache.viewvalues():
            threads.append(backend.reload())

        self.lock.release()

        for thread in threads:
            thread.join()


    def acquire(self, user_backend):
        user_id = user_backend.id

        self.lock.acquire()

        if user_id not in self.cache.viewkeys():
            self.cache[user_id] = Backend(user_backend, self.cache_path, self.cache_path_lock, self.font_scheme, self.connection_pool, self)

        backend = self.cache[user_id]

        backend.reference_count += 1

        self.lock.release()

        return backend


    def release(self, backend):
        self.lock.acquire()

        backend.reference_count -= 1

        if backend.reference_count == 0:
            del self.cache[backend.user_backend.id]

            backend.isolate()

        self.lock.release()
