# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import os

from urllib import urlretrieve


from glib import GError

from gtk import gdk as Gdk

from socks import ProxyError


from polly import ENCODING, dispatch, Notifier, Updatable, BaseCache

from polly.gui.basic import activate_pixbuf, resize_pixbuf



class Backend(Updatable):
#    def __del__(self):
#        print u'debug: preview backend deleted'
    def _update_pixbufs(self):
        success = True

        if not os.path.exists(self.path):
            try:
                self.cache_path_lock.acquire()

                try:
                    if not os.path.isdir(self.cache_path):
                        os.makedirs(self.cache_path)
                finally:
                    self.cache_path_lock.release()

                self.connection_pool.acquire()

                try:
                    urlretrieve(self.small_url, self.path)
                finally:
                    self.connection_pool.release()
            except (OSError, IOError, ProxyError):
                success = False

        Gdk.threads_enter()

        if success:
            try:
                animation = Gdk.PixbufAnimation(self.path)

                if self.name == u'twitter' or self.name == u'twitpic' or self.name == u'twitvid':
                    self.pixbuf = resize_pixbuf(animation.get_static_image(), 150)
                elif self.name == u'imgur':
                    self.pixbuf = resize_pixbuf(animation.get_static_image(), 160)
                else:
                    self.pixbuf = animation.get_static_image()

                self.pixbuf_active = activate_pixbuf(self.pixbuf)
            except GError:
                success = False

                try:
                    os.remove(self.path)
                except OSError:
                    pass

        if not success:
            self.pixbuf = None

            self.pixbuf_active = None

        Gdk.threads_leave()


    def _update(self, previous_thread):
        previous_thread.join()

        self._acquire_update()

        self._update_pixbufs()

        self.notifier.notify()

        self._release_update()


    def __init__(self, name, hash, small_url, cache_path, cache_path_lock, connection_pool, backend_cache):
        super(Backend, self).__init__()

        self.name = name

        self.hash = hash

        self.small_url = small_url.encode(ENCODING)

        self.cache_path = os.path.join(cache_path, self.name)

        self.cache_path_lock = cache_path_lock

        self.connection_pool = connection_pool

        self.backend_cache = backend_cache

        self.reference_count = 0

        self.path = os.path.join(self.cache_path, self.hash)

        self.notifier = Notifier()

        self.first_thread = dispatch(self._update_pixbufs)

        self.thread = self.first_thread


    def isolate(self):
        self.thread.join()


    def reload(self):
        Gdk.threads_enter()

        self.thread = dispatch(self._update, self.thread)

        thread = self.thread

        Gdk.threads_leave()

        return thread


    def release(self):
        self.backend_cache.release(self)



class BackendCache(BaseCache):
    def __init__(self, cache_path, cache_path_lock, connection_pool):
        super(BackendCache, self).__init__()

        self.cache_path = os.path.join(cache_path, u'previews')

        self.cache_path_lock = cache_path_lock

        self.connection_pool = connection_pool


    def reload(self):
        threads = []

        self.lock.acquire()

        for name in self.cache.viewkeys():
            for backend in self.cache[name].viewvalues():
                threads.append(backend.reload())

        self.lock.release()

        for thread in threads:
            thread.join()


    def acquire(self, name, hash, small_url):
        self.lock.acquire()

        if name not in self.cache.viewkeys():
            self.cache[name] = {}

        if hash not in self.cache[name].viewkeys():
            self.cache[name][hash] = Backend(name, hash, small_url, self.cache_path, self.cache_path_lock, self.connection_pool, self)

        backend = self.cache[name][hash]

        backend.reference_count += 1

        self.lock.release()

        return backend


    def release(self, backend):
        self.lock.acquire()

        backend.reference_count -= 1

        if backend.reference_count == 0:
            name = backend.name

            del self.cache[name][backend.hash]

            if not self.cache[name]:
                del self.cache[name]

            backend.isolate()

        self.lock.release()
