# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import os

from urllib import urlretrieve


import numpy

from glib import GError

from gtk import gdk as Gdk

from socks import ProxyError


from polly import ENCODING, md5sum, dispatch, Notifier, Updatable, BaseCache

from polly.gui.basic import resize_pixbuf, transparent_pixbuf, FilesystemPool



def _fade_value(value):
    return long(0.5 * value + 0.5)


def fade_pixbuf(pixbuf_normal):
    array_normal = pixbuf_normal.get_pixels_array()

    pixbuf = Gdk.Pixbuf(pixbuf_normal.get_colorspace(),
                        True,
                        pixbuf_normal.get_bits_per_sample(),
                        pixbuf_normal.get_width(),
                        pixbuf_normal.get_height())

    array = pixbuf.get_pixels_array()

    array[:,:,:3] = array_normal[:,:,:3]

    if pixbuf_normal.get_has_alpha():
        fade_array = numpy.vectorize(_fade_value, [numpy.uint8])

        array[:,:,3] = fade_array(array_normal[:,:,3])
    else:
        array[:,:,3] = 128

    return pixbuf



class Backend(Updatable):
#    def __del__(self):
#        print u'debug: stamp backend deleted'
    def _update_size(self):
        Gdk.threads_enter()

        size = self.permanent_size

        Gdk.threads_leave()

        if self.size == size:
            return False

        self.size = size

        return True


    def _update_pixbufs(self):
        if self.size > 48:
            url = self.bigger_url

            hash = self.bigger_hash

            path = self.bigger_path
        else:
            url = self.normal_url

            hash = self.normal_hash

            path = self.normal_path

        self.filesystem_pool.acquire(hash)

        success = True

        if not os.path.exists(path):
            try:
                self.cache_path_lock.acquire()

                try:
                    if not os.path.isdir(self.cache_path):
                        os.makedirs(self.cache_path)
                finally:
                    self.cache_path_lock.release()

                self.connection_pool.acquire()

                try:
                    urlretrieve(url.encode(ENCODING), path)
                finally:
                    self.connection_pool.release()
            except (OSError, IOError, ProxyError):
                success = False

        Gdk.threads_enter()

        if success:
            try:
                animation = Gdk.PixbufAnimation(path)

                pixbuf_normal = resize_pixbuf(animation.get_static_image(), self.size)

                self.pixbuf = fade_pixbuf(pixbuf_normal)
            except GError:
                success = False

                try:
                    os.remove(path)
                except OSError:
                    pass

        if not success:
            self.pixbuf = transparent_pixbuf(self.size)

        Gdk.threads_leave()

        self.filesystem_pool.release(hash)


    def _update(self, previous_thread, callback):
        previous_thread.join()

        self._acquire_update()

        if callback is None:
            success = True
        else:
            success = callback()

        if success:
            self._update_pixbufs()

            self.notifier.notify()

        self._release_update()


    def _first_update(self):
        self.size = None

        self._update_size()

        self._update_pixbufs()


    def _update_permanent_size(self):
        self.permanent_size = self.dimension * self.font_scheme.unit


    def _on_font_changed(self):
        Gdk.threads_enter()

        self._update_permanent_size()

        self.thread = dispatch(self._update, self.thread, self._update_size)

        Gdk.threads_leave()


    def __init__(self, url, dimension, cache_path, cache_path_lock, font_scheme, filesystem_pool, connection_pool, backend_cache):
        super(Backend, self).__init__()

        self.dimension = dimension

        self.cache_path = cache_path

        self.cache_path_lock = cache_path_lock

        self.font_scheme = font_scheme

        self.filesystem_pool = filesystem_pool

        self.connection_pool = connection_pool

        self.backend_cache = backend_cache

        self.reference_count = 0

        self.normal_url = url

        self.normal_hash = md5sum(self.normal_url)

        self.normal_path = os.path.join(self.cache_path, self.normal_hash)

        index = self.normal_url.rfind(u'.')

        if index < self.normal_url.rfind(u'/'):
            index = len(self.normal_url)

        self.bigger_url = u''.join([self.normal_url[:(index - 7)], u'_bigger', self.normal_url[index:]])

        self.bigger_hash = md5sum(self.bigger_url)

        self.bigger_path = os.path.join(self.cache_path, self.bigger_hash)

        self.notifier = Notifier()

        self.font_scheme.block_updates()

        self._update_permanent_size()

        self.first_thread = dispatch(self._first_update)

        self.thread = self.first_thread

        self.font_uid = self.font_scheme.notifier.connect(self._on_font_changed)

        self.font_scheme.allow_updates()


    def isolate(self):
        self.font_scheme.block_updates()

        self.font_scheme.notifier.disconnect(self.font_uid)

        self.font_scheme.allow_updates()

        self.thread.join()


    def reload(self):
        Gdk.threads_enter()

        self.thread = dispatch(self._update, self.thread, None)

        thread = self.thread

        Gdk.threads_leave()

        return thread


    def release(self):
        self.backend_cache.release(self)



class BackendCache(BaseCache):
    def __init__(self, cache_path, cache_path_lock, font_scheme, connection_pool):
        super(BackendCache, self).__init__()

        self.cache_path = os.path.join(cache_path, u'stamps')

        self.cache_path_lock = cache_path_lock

        self.font_scheme = font_scheme

        self.connection_pool = connection_pool

        self.filesystem_pool = FilesystemPool()


    def reload(self):
        threads = []

        self.lock.acquire()

        for url in self.cache.viewkeys():
            for backend in self.cache[url].viewvalues():
                threads.append(backend.reload())

        self.lock.release()

        for thread in threads:
            thread.join()


    def acquire(self, url, dimension):
        self.lock.acquire()

        if url not in self.cache.viewkeys():
            self.cache[url] = {}

        if dimension not in self.cache[url].viewkeys():
            self.cache[url][dimension] = Backend(url, dimension, self.cache_path, self.cache_path_lock, self.font_scheme, self.filesystem_pool, self.connection_pool, self)

        backend = self.cache[url][dimension]

        backend.reference_count += 1

        self.lock.release()

        return backend


    def release(self, backend):
        self.lock.acquire()

        backend.reference_count -= 1

        if backend.reference_count == 0:
            url = backend.normal_url

            del self.cache[url][backend.dimension]

            if not self.cache[url]:
                del self.cache[url]

            backend.isolate()

        self.lock.release()
