# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from __future__ import division


from gtk import gdk as Gdk


from polly import split, dispatch, Notifier, Updatable

from polly.gui import PROPERTY_SIGNAL



PROPERTY_COLOR_SCHEME = u'gtk-color-scheme'


NORMAL_FG_KEY_TREEVIEW = u'text'

NORMAL_FG_KEY_WINDOW = u'fg'

NORMAL_BG_KEY_TREEVIEW = u'base'

NORMAL_BG_KEY_WINDOW = u'bg'

SELECT_FG_KEY = u'selected_fg'

SELECT_BG_KEY = u'selected_bg'


# NOTE: The defaults are based on Gnome 2.

DEFAULT_NORMAL_FG_SPEC_TREEVIEW = u'#1a1a1a1a1a1a'

DEFAULT_NORMAL_FG_SPEC_WINDOW = '#000000000000'

DEFAULT_NORMAL_BG_SPEC_TREEVIEW = u'#ffffffffffff'

DEFAULT_NORMAL_BG_SPEC_WINDOW = u'#ededececebeb'

DEFAULT_SELECT_FG_SPEC = u'#ffffffffffff'

DEFAULT_SELECT_BG_SPEC = u'#8686ababd9d9'



class Scheme(Updatable):
    def _color_mean(self, fg, bg):
        red = (fg.red + bg.red) // 2

        green = (fg.green + bg.green) // 2

        blue = (fg.blue + bg.blue) // 2

        return Gdk.Color(red, green, blue)


    def _update_specs(self):
        Gdk.threads_enter()

        value = self.gtk_settings.get_property(PROPERTY_COLOR_SCHEME)

        Gdk.threads_leave()

        words = split(value, self.pattern)

        self.specs = {}

        if len(words) % 2 == 0:
            for index in xrange(0, len(words), 2):
                key = words[index][0:-7]

                if key in self.keys:
                    spec = words[index + 1]

                    try:
                        Gdk.color_parse(spec)
                    except ValueError:
                        pass
                    else:
                        self.specs[key] = spec

        if len(self.specs) < len(self.keys):
            self.specs = {self.normal_fg_key: self.default_normal_fg_spec,
                          self.normal_bg_key: self.default_normal_bg_spec,
                          SELECT_FG_KEY: DEFAULT_SELECT_FG_SPEC,
                          SELECT_BG_KEY: DEFAULT_SELECT_BG_SPEC}


    def _update_normal_fg(self):
        if self.normal_fg_spec == self.specs[self.normal_fg_key]:
            return False

        self.normal_fg_spec = self.specs[self.normal_fg_key]

        Gdk.threads_enter()

        self.normal_fg = Gdk.color_parse(self.normal_fg_spec)

        Gdk.threads_leave()

        return True


    def _update_normal_bg(self):
        if self.normal_bg_spec == self.specs[self.normal_bg_key]:
            return False

        self.normal_bg_spec = self.specs[self.normal_bg_key]

        Gdk.threads_enter()

        self.normal_bg = Gdk.color_parse(self.normal_bg_spec)

        Gdk.threads_leave()

        return True


    def _update_normal_mg(self):
        Gdk.threads_enter()

        self.normal_mg = self._color_mean(self.normal_fg, self.normal_bg)

        Gdk.threads_leave()


    def _update_select_fg(self):
        if self.select_fg_spec == self.specs[SELECT_FG_KEY]:
            return False

        self.select_fg_spec = self.specs[SELECT_FG_KEY]

        Gdk.threads_enter()

        self.select_fg = Gdk.color_parse(self.select_fg_spec)

        Gdk.threads_leave()

        return True


    def _update_select_bg(self):
        if self.select_bg_spec == self.specs[SELECT_BG_KEY]:
            return False

        self.select_bg_spec = self.specs[SELECT_BG_KEY]

        Gdk.threads_enter()

        self.select_bg = Gdk.color_parse(self.select_bg_spec)

        Gdk.threads_leave()

        return True


    def _update_select_mg(self):
        Gdk.threads_enter()

        self.select_mg = self._color_mean(self.select_fg, self.select_bg)

        Gdk.threads_leave()


    def _update(self, previous_thread):
        if previous_thread is not None:
            previous_thread.join()

        self._acquire_update()

        self._update_specs()

        normal_fg_changed = self._update_normal_fg()

        if normal_fg_changed:
            self.normal_fg_notifier.notify()

        normal_bg_changed = self._update_normal_bg()

        if normal_bg_changed:
            self.normal_bg_notifier.notify()

        if normal_fg_changed or normal_bg_changed:
            self._update_normal_mg()

            self.normal_mg_notifier.notify()

        select_fg_changed = self._update_select_fg()

        if select_fg_changed:
            self.select_fg_notifier.notify()

        select_bg_changed = self._update_select_bg()

        if select_bg_changed:
            self.select_bg_notifier.notify()

        if select_fg_changed or select_bg_changed:
            self._update_select_mg()

            self.select_mg_notifier.notify()

        self._release_update()


    def _on_changed(self, object, pspec):
        self.thread = dispatch(self._update, self.thread)


    def __init__(self, gtk_settings, windowcolor):
        super(Scheme, self).__init__()

        self.gtk_settings = gtk_settings

        self.pattern = u'[\s\\;]+'

        if windowcolor:
            self.default_normal_fg_spec = DEFAULT_NORMAL_FG_SPEC_WINDOW

            self.default_normal_bg_spec = DEFAULT_NORMAL_BG_SPEC_WINDOW

            self.normal_fg_key = NORMAL_FG_KEY_WINDOW

            self.normal_bg_key = NORMAL_BG_KEY_WINDOW
        else:
            self.default_normal_fg_spec = DEFAULT_NORMAL_FG_SPEC_TREEVIEW

            self.default_normal_bg_spec = DEFAULT_NORMAL_BG_SPEC_TREEVIEW

            self.normal_fg_key = NORMAL_FG_KEY_TREEVIEW

            self.normal_bg_key = NORMAL_BG_KEY_TREEVIEW

        self.normal_fg_spec = self.default_normal_fg_spec

        self.normal_bg_spec = self.default_normal_bg_spec

        self.select_fg_spec = DEFAULT_SELECT_FG_SPEC

        self.select_bg_spec = DEFAULT_SELECT_BG_SPEC

        self.keys = {self.normal_fg_key, self.normal_bg_key, SELECT_FG_KEY, SELECT_BG_KEY}

        self.normal_fg = Gdk.color_parse(self.normal_fg_spec)

        self.normal_bg = Gdk.color_parse(self.normal_bg_spec)

        self.select_fg = Gdk.color_parse(self.select_fg_spec)

        self.select_bg = Gdk.color_parse(self.select_bg_spec)

        self.normal_fg_notifier = Notifier()

        self.normal_bg_notifier = Notifier()

        self.normal_mg_notifier = Notifier()

        self.select_fg_notifier = Notifier()

        self.select_bg_notifier = Notifier()

        self.select_mg_notifier = Notifier()

        self._update_specs()

        self._update_normal_fg()

        self._update_normal_bg()

        self._update_normal_mg()

        self._update_select_fg()

        self._update_select_bg()

        self._update_select_mg()

        self.thread = None

        self.uid = self.gtk_settings.connect(PROPERTY_SIGNAL + PROPERTY_COLOR_SCHEME, self._on_changed)


    def isolate(self):
        self.gtk_settings.disconnect(self.uid)

        if self.thread is not None:
            self.thread.join()
