# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from __future__ import division


import pango as Pango

from gtk import gdk as Gdk


from polly import split, dispatch, Notifier, Updatable

from polly.gui import PROPERTY_SIGNAL



PROPERTY_FONT_NAME = u'gtk-font-name'

PROPERTY_XFT_DPI = u'gtk-xft-dpi'


# NOTE: The defaults are based on Gnome 2.

DEFAULT_SIZE = 10

DEFAULT_DPI = 50

DEFAULT_UNIT = 10



class Scheme(Updatable):
    def _update_size(self):
        Gdk.threads_enter()

        value = self.gtk_settings.get_property(PROPERTY_FONT_NAME)

        Gdk.threads_leave()

        words = split(value, self.pattern)

        try:
            if not words:
                raise ValueError

            size = long(words[-1])

            if size < 1:
                raise ValueError
        except ValueError:
            size = DEFAULT_SIZE

        if self.size == size:
            return False

        self.size = size

        return True


    def _update_dpi(self):
        Gdk.threads_enter()

        value = self.gtk_settings.get_property(PROPERTY_XFT_DPI)

        Gdk.threads_leave()

        dpi = value // Pango.SCALE

        if dpi < 1:
            dpi = DEFAULT_DPI

        if self.dpi == dpi:
            return False

        self.dpi = dpi

        return True


    def _update_unit(self):
        em = (self.size / 72) * self.dpi

        unit = long(1.5 * em + 0.5)

        if unit < 3:
            unit = DEFAULT_UNIT

        if self.unit == unit:
            return False

        self.unit = unit

        self.char = 2 * (self.unit // 3)

        self.void = self.unit // 8

        return True


    def _update(self, previous_thread, callback):
        if previous_thread is not None:
            previous_thread.join()

        self._acquire_update()

        if callback():
            if self._update_unit():
                self.notifier.notify()

        self._release_update()


    def _on_font_name_changed(self, object, pspec):
        self.thread = dispatch(self._update, self.thread, self._update_size)


    def _on_xft_dpi_changed(self, object, pspec):
        self.thread = dispatch(self._update, self.thread, self._update_dpi)


    def __init__(self, gtk_settings):
        super(Scheme, self).__init__()

        self.gtk_settings = gtk_settings

        self.size = DEFAULT_SIZE

        self.dpi = DEFAULT_DPI

        self.unit = DEFAULT_UNIT

        self.char = 6

        self.void = 1

        self.pattern = u'[\s]+'

        self.notifier = Notifier()

        self._update_size()

        self._update_dpi()

        self._update_unit()

        self.thread = None

        self.font_name_uid = self.gtk_settings.connect(PROPERTY_SIGNAL + PROPERTY_FONT_NAME, self._on_font_name_changed)

        self.xft_dpi_uid = self.gtk_settings.connect(PROPERTY_SIGNAL + PROPERTY_XFT_DPI, self._on_xft_dpi_changed)


    def isolate(self):
        self.gtk_settings.disconnect(self.xft_dpi_uid)

        self.gtk_settings.disconnect(self.font_name_uid)

        if self.thread is not None:
            self.thread.join()
