# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import gtk as Gtk

from gtk import gdk as Gdk


from polly.gui import run_open_dialog, run_confirmation_dialog, run_error_dialog

from polly.iconset import Error, ControllerBackend



ADD_TITLE = u'Add new iconset'

REMOVE_TITLE = u'Remove current iconset'



class Controller(object):
    def _update_name_combobox(self):
        if self.backend.name in self.backend.names:
            index = self.backend.names.index(self.backend.name)
        else:
            index = -1

        self.updating_names = True

        self.name_combobox.set_active(index)

        self.updating_names = False


    def _update_remove_button(self):
        added = (self.backend.name in self.backend.added_names)

        self.remove_button.set_sensitive(added)


    def _added_names_callback(self):
        self._update_remove_button()


    def _names_callback(self):
        self.updating_names = True

        self.name_combobox.get_model().clear()

        for name in self.backend.names:
            self.name_combobox.append_text(name)

        self.updating_names = False

        self._update_name_combobox()

        self._update_remove_button()


    def _name_callback(self):
        self._update_name_combobox()

        self._update_remove_button()


    def _on_name_combobox_changed(self, widget):
        if not self.updating_names:
            name = self.name_combobox.get_active_text()

            self.backend.set_name(name)


    def _on_add_button_clicked(self, button):
        toplevel = self.add_button.get_toplevel()

        path = run_open_dialog(toplevel, ADD_TITLE,
                               [(u'Iconsets', [u'*.tar.gz', u'*.tar.bz2'])],
                               self.previous_path)

        if path is not None:
            self.previous_path = path

            Gdk.threads_leave()

            try:
                self.backend.add_new(path)
            except Error as error:
                message = error.message
            else:
                message = None

            Gdk.threads_enter()

            if message is not None:
                run_error_dialog(toplevel, ADD_TITLE, u'Could not add new iconset', message)


    def _on_remove_button_clicked(self, button):
        toplevel = self.remove_button.get_toplevel()

        success = run_confirmation_dialog(toplevel, REMOVE_TITLE,
                                          u'Are you sure you want to remove the iconset <i>{}</i>?'.format(self.backend.name),
                                          u'You will not be able to restore it.')

        if success:
            Gdk.threads_leave()

            try:
                self.backend.remove_current()
            except Error as error:
                message = error.message
            else:
                message = None

            Gdk.threads_enter()

            if message is not None:
                run_error_dialog(toplevel, REMOVE_TITLE, u'Could not remove current iconset', message)


    def __init__(self, cache_path, data_paths, cache_path_lock, setting_factory):
        self.backend = ControllerBackend(self._added_names_callback, self._names_callback, self._name_callback, cache_path, data_paths, cache_path_lock, setting_factory)

        self.name_combobox = Gtk.combo_box_new_text()
        self.name_combobox.connect(u'changed', self._on_name_combobox_changed)
        self.name_combobox.show()

        add_image = Gtk.image_new_from_stock(Gtk.STOCK_ADD, Gtk.ICON_SIZE_BUTTON)
        add_image.show()

        self.add_button = Gtk.Button(None, None, False)
        self.add_button.set_tooltip_text(ADD_TITLE)
        self.add_button.connect(u'clicked', self._on_add_button_clicked)
        self.add_button.add(add_image)
        self.add_button.show()

        remove_image = Gtk.image_new_from_stock(Gtk.STOCK_REMOVE, Gtk.ICON_SIZE_BUTTON)
        remove_image.show()

        self.remove_button = Gtk.Button(None, None, False)
        self.remove_button.set_tooltip_text(REMOVE_TITLE)
        self.remove_button.connect(u'clicked', self._on_remove_button_clicked)
        self.remove_button.add(remove_image)
        self.remove_button.show()

        self.previous_path = None

        self._names_callback()


    def isolate(self):
        self.backend.isolate()
