# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import gtk as Gtk

from gtk import gdk as Gdk


from polly import ENCODING, dispatch

from polly.twitter.account import Error, UnauthorizedError

from polly.gui import DEFAULT_BORDER, CONNECTING_TEXT, show_url



class ProgressPage(Gtk.Alignment):
    def __init__(self):
        self.type = Gtk.ASSISTANT_PAGE_PROGRESS

        self.title = CONNECTING_TEXT

        self.spinner = Gtk.Spinner()
        self.spinner.show()

        super(ProgressPage, self).__init__(0.5, 0.5, 1.0, 1.0)
        self.set_border_width(2 * DEFAULT_BORDER)
        self.add(self.spinner)
        self.show()


    def prepare(self, assistant):
        self.spinner.start()

        assistant.set_page_complete(self, False)

        dispatch(self._update, assistant)



class IntroPage(Gtk.Alignment):
    def __init__(self):
        self.type = Gtk.ASSISTANT_PAGE_INTRO

        self.title = u'Welcome'

        label = Gtk.Label(u'This assistant will guide you through the process of adding a new account.')
        label.set_justify(Gtk.JUSTIFY_CENTER)
        label.set_line_wrap(True)
        label.show()

        super(IntroPage, self).__init__(0.5, 0.5, 0.0, 0.0)
        self.set_border_width(2 * DEFAULT_BORDER)
        self.add(label)
        self.show()


    def prepare(self, assistant):
        assistant.set_page_complete(self, True)



class RequestProgressPage(ProgressPage):
    def _update(self, assistant):
        if assistant.thread is not None:
            assistant.thread.join()

        try:
            assistant.request_data = assistant.account_manager.backend.get_request_data()
        except Error as error:
            message = error.message
        else:
            message = None

        Gdk.threads_enter()

        if message is None:
            assistant.set_current_page(2)
        else:
            assistant.pages[5].label.set_text(message)

            assistant.set_current_page(5)

        self.spinner.stop()

        Gdk.threads_leave()



class AuthorizePage(Gtk.Alignment):
    def _on_activate_link(self, label, uri):
        show_url(uri.decode(ENCODING))

        self.entry.grab_focus()

        return True


    def __init__(self, entry):
        self.entry = entry

        self.type = Gtk.ASSISTANT_PAGE_CONTENT

        self.title = u'Authorization'

        self.markup = (u'<a href="{}">Click here</a> to start the authorization process in your default web browser. Once you complete it, you will receive a PIN that should be typed below.')

        self.label = Gtk.Label(None)
        self.label.set_justify(Gtk.JUSTIFY_CENTER)
        self.label.set_line_wrap(True)
        self.label.set_markup(self.markup)
        self.label.connect(u'activate-link', self._on_activate_link)
        self.label.show()

        alignment = Gtk.Alignment(0.5, 0.5, 0.0, 0.0)
        alignment.add(self.entry)
        alignment.show()

        box = Gtk.VBox(False, 2 * DEFAULT_BORDER)
        box.pack_start(self.label, False, False, 0)
        box.pack_start(alignment, False, False, 0)
        box.show()

        super(AuthorizePage, self).__init__(0.5, 0.5, 0.0, 0.0)
        self.set_border_width(2 * DEFAULT_BORDER)
        self.add(box)
        self.show()


    def prepare(self, assistant):
        url = assistant.account_manager.backend.get_authorize_url(assistant.request_data)

        self.label.set_markup(self.markup.format(url))



class AccessProgressPage(ProgressPage):
    def _update(self, assistant):
        if assistant.thread is not None:
            assistant.thread.join()

        Gdk.threads_enter()

        pin = assistant.entry.get_text()

        Gdk.threads_leave()

        try:
            data = assistant.account_manager.backend.get_access_data(assistant.request_data, pin)

            account = assistant.account_manager.add(data)
        except UnauthorizedError as error:
            message = u'Twitter did not accept the PIN typed.'
        except Error as error:
            message = error.message
        else:
            message = None

        Gdk.threads_enter()

        if message is None:
            assistant.account_manager.stream_manager.new_tweets_home_mentions(account)

            assistant.set_current_page(4)
        else:
            assistant.pages[5].label.set_text(message)

            assistant.set_current_page(5)

        self.spinner.stop()

        Gdk.threads_leave()



class SummaryPage(Gtk.Alignment):
    def _on_activate_link(self, label, uri):
        self.assistant.entrybuffer.delete_text(0, -1)

        self.assistant.set_current_page(0)

        return True


    def _on_populate_popup(self, label, menu):
        menuitems = menu.get_children()

        menuitems[0].set_sensitive(False)
        menuitems[1].set_sensitive(False)


    def __init__(self):
        self.type = Gtk.ASSISTANT_PAGE_SUMMARY

        self.title = u'Thank you'

        label = Gtk.Label(u'The account has been successfully added. You can now close this assistant.')
        label.set_justify(Gtk.JUSTIFY_CENTER)
        label.set_line_wrap(True)
        label.show()

        link = Gtk.Label(None)
        link.set_justify(Gtk.JUSTIFY_CENTER)
        link.set_markup(u'<a href="">add another account</a>')
        link.connect(u'activate-link', self._on_activate_link)
        link.connect(u'populate-popup', self._on_populate_popup)
        link.show()

        box = Gtk.VBox(False, 2 * DEFAULT_BORDER)
        box.pack_start(label, False, False, 0)
        box.pack_start(link, False, False, 0)
        box.show()

        super(SummaryPage, self).__init__(0.5, 0.5, 0.0, 0.0)
        self.set_border_width(2 * DEFAULT_BORDER)
        self.add(box)
        self.show()


    def prepare(self, assistant):
        self.assistant = assistant

        self.assistant.set_page_complete(self, True)



class ErrorPage(Gtk.Alignment):
    def _on_activate_link(self, label, uri):
        self.assistant.entrybuffer.delete_text(0, -1)

        self.assistant.set_current_page(0)

        return True


    def _on_populate_popup(self, label, menu):
        menuitems = menu.get_children()

        menuitems[0].set_sensitive(False)
        menuitems[1].set_sensitive(False)


    def __init__(self):
        self.type = Gtk.ASSISTANT_PAGE_CONTENT

        self.title = u'Error'

        self.label = Gtk.Label(None)
        self.label.set_justify(Gtk.JUSTIFY_CENTER)
        self.label.show()

        link = Gtk.Label(None)
        link.set_justify(Gtk.JUSTIFY_CENTER)
        link.set_markup(u'<a href="">restart authorization process</a>')
        link.connect(u'activate-link', self._on_activate_link)
        link.connect(u'populate-popup', self._on_populate_popup)
        link.show()

        box = Gtk.VBox(True, 2 * DEFAULT_BORDER)
        box.pack_start(self.label, True, True, 0)
        box.pack_start(link, True, True, 0)
        box.show()

        super(ErrorPage, self).__init__(0.5, 0.5, 0.0, 0.0)
        self.set_border_width(2 * DEFAULT_BORDER)
        self.add(box)
        self.show()


    def prepare(self, assistant):
        self.assistant = assistant

        self.assistant.set_page_complete(self, False)



class Window(Gtk.Assistant):
    def __del__(self):
        print u'DEBUG: ASSISTANT DELETED'
    def _on_activate(self, entry):
        if self.get_page_complete(self.pages[2]):
            self.set_current_page(3)


    def _on_inserted_text(self, buffer, position, chars, n_chars):
        self.set_page_complete(self.pages[2], True)


    def _on_deleted_text(self, buffer, position, n_chars):
        if self.entrybuffer.get_length() == 0:
            self.set_page_complete(self.pages[2], False)


    def _on_accel(self, accel_group, acceleratable, keyval, modifier):
        page = self.get_nth_page(self.get_current_page())

        if (self.get_page_type(page) != Gtk.ASSISTANT_PAGE_PROGRESS or
            self.get_page_complete(page)):
            self.quit()


    def _on_prepare(self, assistant, page):
        page.prepare(self)


    def _on_close(self, assistant):
        self.quit()


    def _on_cancel(self, assistant):
        self.quit()


    def _on_delete_event(self, widget, event):
        page = self.get_nth_page(self.get_current_page())

        if (self.get_page_type(page) != Gtk.ASSISTANT_PAGE_PROGRESS or
            self.get_page_complete(page)):
            self.quit()

        return False


    def __init__(self, title, account_manager):
        self.account_manager = account_manager

        self.entry = Gtk.Entry(0)
        self.entry.connect(u'activate', self._on_activate)
        self.entry.show()

        self.entrybuffer = self.entry.get_buffer()
        self.entrybuffer.connect(u'inserted-text', self._on_inserted_text)
        self.entrybuffer.connect(u'deleted-text', self._on_deleted_text)

        self.pages = [IntroPage(),
                      RequestProgressPage(),
                      AuthorizePage(self.entry),
                      AccessProgressPage(),
                      SummaryPage(),
                      ErrorPage()]

        accel_group = Gtk.AccelGroup()
        accel_group.connect_group(Gtk.keysyms.Escape, 0, Gtk.ACCEL_VISIBLE, self._on_accel)

        super(Window, self).__init__()
        self.set_resizable(False)
        self.set_title(title)
        self.connect(u'prepare', self._on_prepare)
        self.connect(u'close', self._on_close)
        self.connect(u'cancel', self._on_cancel)
        self.connect(u'delete-event', self._on_delete_event)
        self.add_accel_group(accel_group)

        for page in self.pages:
            self.append_page(page)

            self.set_page_type(page, page.type)

            self.set_page_title(page, page.title)

        self.thread = None


    def quit(self):
        self.hide()

        if self.thread is not None:
            self.thread.join()

        self.destroy()

        for page in self.pages:
            page.destroy()

        del self.pages[:]

        self.entrybuffer = None
