# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import re

from string import ascii_letters, digits, punctuation


import gobject as GObject

import gtk as Gtk

import pango as Pango

from glib import GError

from gtk import gdk as Gdk

from gtkspell import Spell as GtkSpell


from polly import ENCODING, dispatch

from polly.shortener import Exit, Error

from polly.gui import NICK_DIMENSION, run_generic_error_dialog



PAGE_SIZE = 5


TWITTER_LIMIT = 140



class Completer(Gtk.Window):
    def __init__(self):
        self.liststore = Gtk.ListStore(GObject.TYPE_STRING)

        cellrenderer = Gtk.CellRendererText()

        self.treeviewcolumn = Gtk.TreeViewColumn(None, cellrenderer, text=0)

        self.treeview = Gtk.TreeView(self.liststore)
        self.treeview.set_headers_visible(False)
        self.treeview.append_column(self.treeviewcolumn)
        self.treeview.show()

        self.treeselection = self.treeview.get_selection()

        self.adjustment = self.treeview.get_vadjustment()

        scrolledwindow = Gtk.ScrolledWindow(self.treeview.get_hadjustment(), self.adjustment)
        scrolledwindow.add(self.treeview)
        scrolledwindow.show()

        super(Completer, self).__init__(Gtk.WINDOW_POPUP)
        self.add(scrolledwindow)


    def get_nick(self):
        model, iter = self.treeselection.get_selected()

        return model.get_value(iter, 0).decode(ENCODING)


    def set_nicks(self, nicks):
        self.liststore.clear()

        for nick in nicks:
            self.liststore.append([nick])

        self.treeselection.select_path(0)

        self.adjustment.set_value(0.0)


    def arrow_up(self):
        selected = self.treeselection.get_selected()

        path = self.liststore.get_path(selected[1])

        if path[-1] > 0:
            path = path[-1] - 1

            self.treeselection.select_path(path)

            self.treeview.scroll_to_cell(path, None, False, 0.0, 0.0)


    def arrow_down(self):
        selected = self.treeselection.get_selected()

        iter = self.liststore.iter_next(selected[1])

        if iter is not None:
            path = self.liststore.get_path(iter)

            self.treeselection.select_iter(iter)

            self.treeview.scroll_to_cell(path[-1], None, False, 0.0, 0.0)


    def page_up(self):
        selected = self.treeselection.get_selected()

        path = self.liststore.get_path(selected[1])

        area = self.treeview.get_background_area(path[-1], self.treeviewcolumn)

        page_size = long(self.adjustment.get_page_size()) / area.height

        path = path[-1] - page_size

        if path > 0:
            # NOTE: The conversion to int will not be necessary
            # once the planned port to Python 3 is completed.

            self.treeselection.select_path(int(path))

            value = self.adjustment.get_value() - float(page_size * area.height)

            self.adjustment.set_value(max(value, 0.0))
        else:
            self.treeselection.select_path(0)

            self.treeview.scroll_to_cell(0, None, False, 0.0, 0.0)


    def page_down(self):
        selected = self.treeselection.get_selected()

        path = self.liststore.get_path(selected[1])

        last = self.liststore.iter_n_children(None) - 1

        area = self.treeview.get_background_area(path[-1], self.treeviewcolumn)

        page_size = long(self.adjustment.get_page_size()) / area.height

        path = path[-1] + page_size

        if path < last:
            # NOTE: The conversion to int will not be necessary
            # once the planned port to Python 3 is completed.

            self.treeselection.select_path(int(path))

            value = self.adjustment.get_value() + float(page_size * area.height)

            upper = self.adjustment.get_upper() - self.adjustment.get_page_size()

            self.adjustment.set_value(min(value, upper))
        else:
            self.treeselection.select_path(last)

            self.treeview.scroll_to_cell(last, None, False, 0.0, 0.0)


    def home(self):
        selected = self.treeselection.get_selected()

        path = self.liststore.get_path(selected[1])

        if path[-1] > 0:
            self.treeselection.select_path(0)

            self.treeview.scroll_to_cell(0, None, False, 0.0, 0.0)


    def end(self):
        selected = self.treeselection.get_selected()

        path = self.liststore.get_path(selected[1])

        last = self.liststore.iter_n_children(None) - 1

        if path[-1] < last:
            self.treeselection.select_path(last)

            self.treeview.scroll_to_cell(last, None, False, 0.0, 0.0)



class Controller(object):
    def _move_completer(self):
        x, y = self.textview.get_window(Gtk.TEXT_WINDOW_WIDGET).get_origin()

        location = self.textview.get_iter_location(self.completing_start)

        value = long(self.adjustment.get_value())

        x += location.x

        y += location.y + self.unit - value

        self.completer.move(x, y)


    def _hide_completer(self):
        self.completer.hide()

        del self.nicks[:]


    def _shorten_task(self, previous_thread):
        if previous_thread is not None:
            previous_thread.join()

        try:
            short_url = self.shortener_controller.backend.shorten(self.url)
        except Exit:
            short_url = None
        except Error as error:
            short_url = None

            self.status_bar.popup(error.message)

        Gdk.threads_enter()

        self.textbuffer.remove_tag(self.texttag, self.shortening_start, self.shortening_end)

        if short_url is not None:
            self.textbuffer.delete(self.shortening_start, self.shortening_end)

            self.textbuffer.insert_at_cursor(short_url)

            self.shortening_end = self.textbuffer.get_iter_at_mark(self.textbuffer.get_insert())

            self.shortening_start = self.textbuffer.get_iter_at_offset(self.shortening_end.get_offset() - len(short_url))

        self.textview.set_sensitive(True)

        self.textview.grab_focus()

        self.send_button.set_spinning(False)

        # NOTE: The lines below fix a GTK+ bug where
        # a widget remains painted as insensitive if
        # it regained sensitivity while off-screen.

        if self.send_button.get_sensitive():
            self.send_button.set_sensitive(False)

            Gdk.flush()

            self.send_button.set_sensitive(True)

        self.compose_button.set_sensitive(True)

        self.shortening = False

        if short_url is not None:
            self.unshorten_button.show()

        Gdk.threads_leave()


    def _update_count(self):
        char_text = self.get_text()

        self.char_count = len(char_text)

        self.real_count = len(char_text.strip())

        link_count = 0

        matches = re.findall(self.pattern_findall, char_text, re.UNICODE | re.IGNORECASE)

        for _, _, url, protocol, domain, _, path, _ in matches:
            if protocol:
                lengths = self.account_manager.get_lengths()

                if protocol[4] == u's':
                    self.char_count += lengths[1] - len(url)
                else:
                    self.char_count += lengths[0] - len(url)

                link_count += 1
            else:
                if (re.match(self.subpattern1, domain, re.UNICODE | re.IGNORECASE) is not None or
                    re.match(self.subpattern2, domain, re.UNICODE | re.IGNORECASE) is not None or
                   (re.match(self.subpattern3, domain, re.UNICODE | re.IGNORECASE) is not None and path)):
                    lengths = self.account_manager.get_lengths()

                    self.char_count += lengths[0] - len(url)

                    link_count += 1

        if link_count == 0:
            self.link_label.hide()
        else:
            if link_count == 1:
                link_text = u'1 link'
            else:
                link_text = u'{} links'.format(link_count)

            self.link_label.set_text(u' ({})'.format(link_text))

            self.link_label.show()


    def _update_label(self):
        self.char_label.set_markup(u'<b>{}</b>'.format(self.limit - self.char_count))


    def _update_alert(self):
        if self.limited and (self.real_count == 0 or self.char_count > self.limit):
            self.char_label.modify_fg(Gtk.STATE_NORMAL, self.color_active)

            self.send_button.set_sensitive(False)
        else:
            self.send_button.set_sensitive(True)

            self.char_label.modify_fg(Gtk.STATE_NORMAL, self.color)


    def _update_check(self):
        check = self.check_setting.get_value()

        if self.check == check:
            return False

        self.check = check

        return True


    def _update_override(self):
        override = self.override_setting.get_value()

        if self.override == override:
            return False

        self.override = override

        return True


    def _check_callback(self):
        if self.spell is not None:
            self.spell.detach()

        if self.check:
            try:
                self.spell = GtkSpell(self.textview, None)
            except GError:
                self.spell = None
        else:
            self.spell = None

        # NOTE: The following line fixes a bug where
        # the GtkSpell underlining is not refreshed.

        self.textview.queue_draw()

        self.check_checkbutton.set_active(self.check)


    def _override_callback(self):
        index = 1 if self.override else 0

        self.override_radiobuttons[index].set_active(True)


    def _on_unshorten_button_clicked(self):
        self.textbuffer.delete(self.shortening_start, self.shortening_end)

        self.textbuffer.insert_at_cursor(self.url)


    def _on_textview_button_press_event(self, widget, event):
        if event.button == 2:
            self.pasting = True

            self.slices = []

        return False


    def _on_textview_focus_in_event(self, widget, event):
        self.textview.queue_draw()

        return False


    def _on_textview_focus_out_event(self, widget, event):
        self._hide_completer()

        return False


    def _on_textview_key_press_event(self, widget, event):
        if event.keyval == Gtk.keysyms.KP_Enter or event.keyval == Gtk.keysyms.Return:
            if self.nicks:
                self.textbuffer.delete(self.completing_start, self.completing_end)

                self.textbuffer.insert_at_cursor(self.completer.get_nick())

                self.completing_end = None

                self.completing_start = None

                self._hide_completer()

                return True
            else:
                shift = event.state & Gdk.SHIFT_MASK

                if (self.override and not shift) or (not self.override and shift):
                    if self.send_button.get_sensitive():
                        self.send_button.callback()

                    return True
        elif event.keyval == Gtk.keysyms.KP_Up or event.keyval == Gtk.keysyms.Up:
            if self.nicks:
                self.completer.arrow_up()

                return True
        elif event.keyval == Gtk.keysyms.KP_Down or event.keyval == Gtk.keysyms.Down:
            if self.nicks:
                self.completer.arrow_down()

                return True
        elif event.keyval == Gtk.keysyms.KP_Page_Up or event.keyval == Gtk.keysyms.Page_Up:
            if self.nicks:
                self.completer.page_up()

                return True
        elif event.keyval == Gtk.keysyms.KP_Page_Down or event.keyval == Gtk.keysyms.Page_Down:
            if self.nicks:
                self.completer.page_down()

                return True
        elif event.keyval == Gtk.keysyms.KP_Home or event.keyval == Gtk.keysyms.Home:
            if self.nicks:
                self.completer.home()

                return True
        elif event.keyval == Gtk.keysyms.KP_End or event.keyval == Gtk.keysyms.End:
            if self.nicks:
                self.completer.end()

                return True
        elif (event.keyval == Gtk.keysyms.KP_Left or event.keyval == Gtk.keysyms.Left or
              event.keyval == Gtk.keysyms.KP_Right or event.keyval == Gtk.keysyms.Right or
              event.keyval == Gtk.keysyms.Escape):
            self._hide_completer()

        return False


    def _on_textview_paste_clipboard(self, text_view):
        self.pasting = True

        self.slices = []


    def _on_textbuffer_mark_set(self, textbuffer, location, mark):
        # NOTE: The following line fixes a bug where
        # the GtkSpell underlining is not refreshed.

        self.textview.queue_draw()


    def _on_textbuffer_changed(self, textbuffer):
        self._update_count()

        self._update_label()

        self._update_alert()

        # NOTE: The following line fixes a bug where
        # the GtkSpell underlining is not refreshed.

        self.textview.queue_draw()

        if self.inserted:
            self.inserted = False
        else:
            end = self.textbuffer.get_iter_at_mark(self.textbuffer.get_insert())

            start = end.copy()

            start.backward_char()

            success = False

            if start.get_char() in self.allowed:
                while start.backward_char():
                    char = start.get_char()

                    if char not in self.allowed:
                        if char == u'@' or char == u'\uFF20':
                            start.forward_char()

                            success = True

                        break

            if success:
                del self.nicks[:]

                account = self.account_manager.account

                if account is not None:
                    bytes = self.textbuffer.get_text(start, end)

                    prefix = bytes.decode(ENCODING)

                    account.database.select(prefix, self.nicks)

                if self.nicks:
                    self.completing_start = start

                    self.completing_end = end

                    self.completer.set_nicks(self.nicks)

                    self._move_completer()

                    self.completer.show()
                else:
                    self._hide_completer()
            else:
                self._hide_completer()

        self.unshorten_button.hide()


    def _on_textbuffer_insert_text(self, textbuffer, location, text, len):
        if self.pasting:
            self.slices.append(text)


    def _on_textbuffer_paste_done(self, textbuffer, arg1):
        self.url = b''.join(self.slices).decode(ENCODING)

        if re.match(self.pattern_match, self.url, re.UNICODE | re.IGNORECASE) is not None:
            self.shortening = True

            self.compose_button.set_sensitive(False)

            self.send_button.spinner.set_tooltip_text(u'Shortening\u2026')

            self.send_button.set_spinning(True)

            self.textview.set_sensitive(False)

            self.shortening_end = self.textbuffer.get_iter_at_mark(self.textbuffer.get_insert())

            self.shortening_start = self.textbuffer.get_iter_at_offset(self.shortening_end.get_offset() - len(self.url))

            self.textbuffer.apply_tag(self.texttag, self.shortening_start, self.shortening_end)

            self.thread = dispatch(self._shorten_task, self.thread)

        self.pasting = False


    def _on_check_checkbutton_toggled(self, togglebutton):
        check = self.check_checkbutton.get_active()

        if self.check != check:
            self.check_setting.set_value(check)


    def _on_override_radiobutton_toggled(self, togglebutton):
        if togglebutton.get_active():
            index = self.override_radiobuttons.index(togglebutton)

            override = True if index == 1 else False

            if self.override != override:
                self.override_setting.set_value(override)


    def _on_check_changed(self):
        if self._update_check():
            self._check_callback()


    def _on_override_changed(self):
        if self._update_override():
            self._override_callback()


    def _on_font_changed(self):
        self.unit = self.font_scheme.unit

        self.scrolledwindow.set_size_request(-1, 2 * self.unit)

        self.completer.resize(NICK_DIMENSION * self.font_scheme.char, PAGE_SIZE * self.unit)

        if self.nicks:
            self._move_completer()


    def __init__(self, compose_button, send_button, setting_factory, shortener_controller, font_scheme, button_factory, status_bar, account_manager):
        self.compose_button = compose_button

        self.send_button = send_button

        self.shortener_controller = shortener_controller

        self.font_scheme = font_scheme

        self.status_bar = status_bar

        self.account_manager = account_manager

        self.check = True

        self.override = True

        self.inserted = False

        self.pasting = False

        self.shortening = False

        self.limited = True

        self.limit = TWITTER_LIMIT

        self.allowed = u''.join([digits, ascii_letters, u'_'])

        # October 03, 2011
        # https://github.com/twitter/twitter-text-rb
        # https://dev.twitter.com/blog/twitter-text-library-and-tco-links-wrapping-update

        self.invalid = u'\uFFFE\uFEFF\uFFFF\u202A\u202B\u202C\u202D\u202E'

        preceding = u'(?:[^{}a-z0-9{}]|^)'.format(re.escape(u'-/"\'!=_@\uFF20.'), self.invalid)

        control = [unichr(127)]

        for code in range(32):
            control.append(unichr(code))

        domain_char = u'[^{}\s{}{}]'.format(re.escape(punctuation), u''.join(control), self.invalid)

        subdomain = u'(?:(?:{0}(?:[\\_\\-]|{0})*)?{0}\\.)'.format(domain_char)

        domain_name = u'(?:(?:{0}(?:\\-|{0})*)?{0}\\.)'.format(domain_char)

        gtld = (u'(?:(?:aero|asia|biz|cat|com|coop|edu|gov|info|int|jobs|mil|'
                u'mobi|museum|name|net|org|pro|tel|travel)(?=[^a-z]|$))')

        cctld = (u'(?:(?:ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|'
                 u'bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|'
                 u'cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|'
                 u'eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|'
                 u'gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|'
                 u'kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|'
                 u'mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|'
                 u'nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|'
                 u'rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|sk|sl|sm|sn|so|sr|ss|st|su|sv|sy|sz|'
                 u'tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|'
                 u'vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|za|zm|zw)(?=[^a-z]|$))')

        punycode = u'(?:xn\\-\\-[0-9a-z]+)'

        domain = u'(?:{}*{}(?:{}|{}|{}))'.format(subdomain, domain_name, gtld, cctld, punycode)

        port = u'[0-9]+'

        latin = [unichr(351)]

        for code in range(192, 215):
            latin.append(unichr(code))

        for code in range(216, 247):
            latin.append(unichr(code))

        for code in range(248, 256):
            latin.append(unichr(code))

        path_char = u'[a-z0-9{}{}]'.format(re.escape(u'!*\';:=+,.$/%#[]-_~&|'), u''.join(latin))

        path_parens = u'\\({}+\\)'.format(path_char)

        path_ending = u'[a-z0-9{}{}]|(?:{})'.format(re.escape(u'=_#/+-'), u''.join(latin), path_parens)

        path = u'(?:(?:{0}*(?:{1}{0}*)*{2})|(?:@{0}+\\/))'.format(path_char, path_parens, path_ending)

        query_char = u'[a-z0-9{}]'.format(re.escape(u'!*\'();:&=+$/%#[]-_.,~|'))

        query_ending = u'[a-z0-9{}]'.format(re.escape(u'_&=#/'))

        self.pattern_match = u'^https?\\:\\/\\/{}(?:\\:{})?(?:\\/{}*)?(?:\\?{}*{})?$'.format(domain, port, path, query_char, query_ending)

        self.pattern_findall = u'(({})((https?\\:\\/\\/)?({})(?:\\:({}))?(\\/{}*)?(\\?{}*{})?))'.format(preceding, domain, port, path, query_char, query_ending)

        self.subpattern1 = u'^{}*{}(?:{}|{})$'.format(subdomain, domain_name, gtld, punycode)

        self.subpattern2 = u'^{}+{}{}$'.format(subdomain, domain_name, cctld)

        self.subpattern3 = u'^{}{}$'.format(domain_name, cctld)

        self.nicks = []

        self.color = None

        self.color_active = Gdk.color_parse(u'red')

        self.char_label = Gtk.Label(None)

        self.texttag = Gtk.TextTag(None)
        self.texttag.set_property(u'underline', Pango.UNDERLINE_SINGLE)

        self.textview = Gtk.TextView(None)
        self.textview.set_wrap_mode(Gtk.WRAP_WORD_CHAR)
        self.textview.connect(u'button-press-event', self._on_textview_button_press_event)
        self.textview.connect(u'focus-in-event', self._on_textview_focus_in_event)
        self.textview.connect(u'focus-out-event', self._on_textview_focus_out_event)
        self.textview.connect(u'key-press-event', self._on_textview_key_press_event)
        self.textview.connect(u'paste-clipboard', self._on_textview_paste_clipboard)
        self.textview.show()

        self.textbuffer = self.textview.get_buffer()
        self.textbuffer.connect(u'mark-set', self._on_textbuffer_mark_set)
        self.textbuffer.connect(u'changed', self._on_textbuffer_changed)
        self.textbuffer.connect(u'insert-text', self._on_textbuffer_insert_text)
        self.textbuffer.connect(u'paste-done', self._on_textbuffer_paste_done)
        self.textbuffer.get_tag_table().add(self.texttag)

        self.spell = None

        self.scrolledwindow = Gtk.ScrolledWindow(None, None)
        self.scrolledwindow.set_policy(Gtk.POLICY_AUTOMATIC, Gtk.POLICY_AUTOMATIC)
        self.scrolledwindow.set_shadow_type(Gtk.SHADOW_IN)
        self.scrolledwindow.add(self.textview)
        self.scrolledwindow.show()

        self.adjustment = self.scrolledwindow.get_vadjustment()

        self.link_label = Gtk.Label(None)

        self.unshorten_button = button_factory.build(u'unshorten', self._on_unshorten_button_clicked)
        self.unshorten_button.set_toggled(True)
        self.unshorten_button.box.set_tooltip_text(u'Undo last shortening')
        self.unshorten_button.hide()

        self.check_checkbutton = Gtk.CheckButton(u'Check entry for spelling mistakes', False)
        self.check_checkbutton.connect(u'toggled', self._on_check_checkbutton_toggled)
        self.check_checkbutton.show()

        self.override_radiobuttons = []

        self.override_radiobuttons.append(Gtk.RadioButton(None,
                                                          u'Enter inserts linebreak, Shift+Enter sends', False))
        self.override_radiobuttons[0].connect(u'toggled', self._on_override_radiobutton_toggled)
        self.override_radiobuttons[0].show()

        self.override_radiobuttons.append(Gtk.RadioButton(self.override_radiobuttons[0],
                                                          u'Enter sends, Shift+Enter inserts linebreak', False))
        self.override_radiobuttons[1].connect(u'toggled', self._on_override_radiobutton_toggled)
        self.override_radiobuttons[1].show()

        self.check_setting = setting_factory.build_boolean(u'entry', u'check', self._on_check_changed)

        self.override_setting = setting_factory.build_boolean(u'entry', u'override', self._on_override_changed)

        self.completer = Completer()

        self._update_count()

        self._update_label()

        self._update_alert()

        self._update_check()

        self._update_override()

        self._check_callback()

        self._override_callback()

        self.font_scheme.block_updates()

        self._on_font_changed()

        self.font_uid = self.font_scheme.notifier.connect(self._on_font_changed)

        self.font_scheme.allow_updates()

        self.thread = None


    def isolate(self):
        if self.thread is not None:
            self.thread.join()

        self.font_scheme.block_updates()

        self.font_scheme.notifier.disconnect(self.font_uid)

        self.font_scheme.allow_updates()

        self.completer.destroy()

        self.override_setting.isolate()

        self.check_setting.isolate()

        self.unshorten_button.isolate()

        self.textbuffer = None


    def get_shortening(self):
        if self.shortening:
            run_generic_error_dialog(self.scrolledwindow.get_toplevel(), u'Please wait until the shortening process is finished and try again.')

        return self.shortening


    def get_text(self):
        start = self.textbuffer.get_start_iter()

        end = self.textbuffer.get_end_iter()

        bytes = self.textbuffer.get_text(start, end)

        return bytes.decode(ENCODING)


    def set_color(self, color):
        self.color = color

        self.link_label.modify_fg(Gtk.STATE_NORMAL, self.color)

        self._update_alert()


    def enable_limited(self):
        self.limited = True

        self.limit = TWITTER_LIMIT

        self._update_label()

        self._update_alert()


    def disable_limited(self):
        self.limited = False

        lengths = self.account_manager.get_lengths()

        self.limit = TWITTER_LIMIT - lengths[0] - 1

        self._update_label()

        self._update_alert()


    def insert(self, text):
        self.inserted = True

        iter = self.textbuffer.get_iter_at_mark(self.textbuffer.get_insert())

        self.textbuffer.insert(iter, text)


    def append(self, text):
        end = self.textbuffer.get_end_iter()

        self.textbuffer.insert(end, u' {}'.format(text))


    def clear(self):
        start = self.textbuffer.get_start_iter()

        end = self.textbuffer.get_end_iter()

        self.textbuffer.delete(start, end)


    def overwrite(self, text):
        start = self.textbuffer.get_start_iter()

        end = self.textbuffer.get_end_iter()

        self.textbuffer.delete(start, end)

        start = self.textbuffer.get_start_iter()

        self.textbuffer.insert(start, text)


    def home(self):
        start = self.textbuffer.get_start_iter()

        self.textbuffer.place_cursor(start)


    def move_completer(self):
        if self.nicks:
            self._move_completer()
