# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from __future__ import division


import gtk as Gtk

from gtk import gdk as Gdk


from polly import escape, dispatch

from polly.twitter.account import Exit, Error, ForbiddenError

from polly.gui import DEFAULT_BORDER, run_generic_error_dialog, NOTIMPLEMENTEDYET

from polly.gui.twitter import CANCEL_MESSAGE_TOOLTIP

from polly.gui.twitter.lead.group.dialog import CreateGroupDialog

from polly.gui.twitter.post import tweet

from polly.gui.header import assistant, media, entry



SHOW_COMPOSE_TOOLTIP = u'Show compose box'


TWEET_TOOLTIP = u'What\u2019s happening?'

REPLY_STRING = u'Reply to {}'

MESSAGE_STRING = u'Message to {}'

MEDIA_STRING = u'Description for {}'



class Panel(Gtk.EventBox):
    def _cancel_reply(self):
        if not self.media_button.toggled:
            self.reply_button.hide()

            self._set_label_text(TWEET_TOOLTIP)

        self.entry_controller.clear()

        self.replied_tweet.set_replying(False)

        self.replied_tweet = None


    def _cancel_message_to_user(self):
        if not self.media_button.toggled:
            self.message_button.hide()

            self._set_label_text(TWEET_TOOLTIP)

        self.entry_controller.clear()

        self.messaged_user.set_messaging(False)

        self.messaged_user = None


    def _cancel_message_to_message(self):
        if not self.media_button.toggled:
            self.message_button.hide()

            self._set_label_text(TWEET_TOOLTIP)

        self.entry_controller.clear()

        self.messaged_message.set_messaging(False)

        self.messaged_message = None


    def _cancel_media(self):
        self.media_button.box.set_tooltip_text(u'{}\u2026'.format(media.TITLE))

        self.media_button.set_toggled(False)

        if self.replied_tweet is not None:
            self.reply_button.show()
        elif self.messaged_user is not None or self.messaged_message is not None:
            self.message_button.show()

        if self.replied_tweet is not None:
            self._set_label_text(REPLY_STRING.format(self.replied_tweet.user_nick))
        elif self.messaged_user is not None:
            self._set_label_text(MESSAGE_STRING.format(self.messaged_user.user_nick))
        elif self.messaged_message is not None:
            self._set_label_text(MESSAGE_STRING.format(self.messaged_message.user_nick))
        else:
            self._set_label_text(TWEET_TOOLTIP)

        self.entry_controller.enable_limited()

        self.provider = None


    def _send_post(self):
        try:
            if self.replied_tweet is not None:
                data = self.account.backend.send_tweet(self.text, self.replied_tweet.tweet_backend.id)
            elif self.messaged_user is not None:
                data = self.account.backend.send_message(self.messaged_user.user_backend.id, self.text)
            elif self.messaged_message is not None:
                data = self.account.backend.send_message(self.messaged_message.user_backend.id, self.text)
            else:
                data = self.account.backend.send_tweet(self.text, None)
        except ForbiddenError as error:
            success = False

            if self.messaged_user is None and self.messaged_message is None:
                self.status_bar.popup(error.message)
            else:
                self.status_bar.popup(u'Twitter refused the request. Check if the user is following you and try again.')
        except Error as error:
            success = False

            self.status_bar.popup(error.message)
        else:
            success = True

            if data is not None:
                if self.messaged_user is None and self.messaged_message is None:
                    self.account.backend.user_backend.set_data(data[2])
                else:
                    self.account.backend.user_backend_cache.set_data(data[1])

                    self.account.backend.user_backend.set_data(data[2])

        Gdk.threads_enter()

        self.send_button.set_spinning(False)

        self.media_button.set_sensitive(True)

        if self.replied_tweet is not None:
            self.reply_button.set_sensitive(True)
        elif self.messaged_user is not None or self.messaged_message is not None:
            self.message_button.set_sensitive(True)

        if success:
            if self.replied_tweet is not None:
                self._cancel_reply()
            elif self.messaged_user is not None:
                self._cancel_message_to_user()
            elif self.messaged_message is not None:
                self._cancel_message_to_message()
            else:
                self.entry_controller.clear()

            if self.hide_on_send:
                self.set_composing(False)

        return False


    def _send_media(self):
        try:
            short_url = self.provider.send(self.account, self.text, self.path, self._update_progressbar)
        except Exit:
            short_url = None
        except Error as error:
            short_url = None

            self.status_bar.popup(error.message)

        Gdk.threads_enter()

        self.progress_box.hide()

        self.footer_box.show()

        self.cancel_button.set_spinning(False)

        if short_url is not None:
            self.entry_controller.append(short_url)

            self._cancel_media()

            self.path = None

        return True


    def _send(self, previous_thread, callback):
        if previous_thread is not None:
            previous_thread.join()

        grab_focus = callback()

        self.text = None

        self.entry_controller.textview.set_sensitive(True)

        if grab_focus:
            self.entry_controller.textview.grab_focus()

        self.entry_controller.unshorten_button.set_sensitive(True)

        self.compose_button.set_sensitive(True)

        self.account_manager.set_sensitive(True)

        self.sending = False

        self.account = None

        Gdk.threads_leave()


    def _set_label_text(self, text):
        self.label.set_markup(u'<i>{}</i> '.format(text))


    def _update_progressbar(self, download_total, download_current, upload_total, upload_current):
        if upload_total == 0:
            fraction = 0.0
        else:
            fraction = upload_current / upload_total

        percentage = u'{}%'.format(long(100 * fraction + 0.5))

        Gdk.threads_enter()

        self.progressbar.set_fraction(fraction)

        self.progressbar.set_text(percentage)

        cancel = self.cancel

        Gdk.threads_leave()

        return cancel


    def _update_hide_on_send(self):
        hide_on_send = self.hide_on_send_setting.get_value()

        if self.hide_on_send == hide_on_send:
            return False

        self.hide_on_send = hide_on_send

        return True


    def _update_hide_on_cancel(self):
        hide_on_cancel = self.hide_on_cancel_setting.get_value()

        if self.hide_on_cancel == hide_on_cancel:
            return False

        self.hide_on_cancel = hide_on_cancel

        return True


    def _on_compose_button_clicked(self):
        self.set_composing(not self.compose_button.toggled)


    def _on_reply_button_clicked(self):
        self.cancel_reply()


    def _on_message_button_clicked(self):
        if self.messaged_user is not None:
            self.cancel_message_to_user()
        else:
            self.cancel_message_to_message()


    def _on_media_button_clicked(self):
        name, self.path = self.media_runner.run(self.get_toplevel())

        if self.path is None:
            self._cancel_media()
        else:
            self.provider = self.media_runner.backend.providers[name]

            self.entry_controller.disable_limited()

            self._set_label_text(MEDIA_STRING.format(name))

            if self.replied_tweet is not None:
                self.reply_button.hide()
            elif self.messaged_user is not None or self.messaged_message is not None:
                self.message_button.hide()

            self.media_button.set_toggled(True)

            self.media_button.box.set_tooltip_text(escape(self.path))


    def _on_send_button_clicked(self):
        for char in self.entry_controller.get_text():
            if char in self.entry_controller.invalid:
                run_generic_error_dialog(self.get_toplevel(), u'The text contains invalid characters.')

                return

        account = self.get_current_account()

        if account is not None:
            self.account = account

            self.sending = True

            self.account_manager.set_sensitive(False)

            self.compose_button.set_sensitive(False)

            self.entry_controller.unshorten_button.set_sensitive(False)

            self.entry_controller.textview.set_sensitive(False)

            self.text = self.entry_controller.get_text()

            if self.media_button.toggled:
                self.progressbar.set_fraction(0.0)

                self.progressbar.set_text(u'0%')

                self.cancel = False

                self.footer_box.hide()

                self.progress_box.show()

                self.send_thread = dispatch(self._send, self.send_thread, self._send_media)
            else:
                if self.replied_tweet is not None:
                    self.reply_button.set_sensitive(False)
                elif self.messaged_user is not None or self.messaged_message is not None:
                    self.message_button.set_sensitive(False)

                self.media_button.set_sensitive(False)

                self.send_button.spinner.set_tooltip_text(u'Sending\u2026')

                self.send_button.set_spinning(True)

                self.send_thread = dispatch(self._send, self.send_thread, self._send_post)


    def _on_cancel_button_clicked(self):
        self.cancel_button.set_spinning(True)

        self.cancel = True


    def _on_hide_on_send_checkbutton_toggled(self, togglebutton):
        hide_on_send = self.hide_on_send_checkbutton.get_active()

        if self.hide_on_send != hide_on_send:
            self.hide_on_send_setting.set_value(hide_on_send)


    def _on_hide_on_cancel_checkbutton_toggled(self, togglebutton):
        hide_on_cancel = self.hide_on_cancel_checkbutton.get_active()

        if self.hide_on_cancel != hide_on_cancel:
            self.hide_on_cancel_setting.set_value(hide_on_cancel)


    def _on_hide_on_send_changed(self):
        if self._update_hide_on_send():
            self.hide_on_send_checkbutton.set_active(self.hide_on_send)


    def _on_hide_on_cancel_changed(self):
        if self._update_hide_on_cancel():
            self.hide_on_cancel_checkbutton.set_active(self.hide_on_cancel)


    def _on_font_changed(self):
        unit = self.font_scheme.unit

        Gdk.threads_enter()

        self.progressbar.set_size_request(-1, unit)

        Gdk.threads_leave()


    def _on_theme_changed(self):
        style = self.theme_scheme.style

        Gdk.threads_enter()

        fg = style.fg[Gtk.STATE_NORMAL]

        bg = style.bg[Gtk.STATE_NORMAL]

        self.account_manager.set_colors(fg, style)

        self.entry_controller.set_color(fg)

        self.label.modify_fg(Gtk.STATE_NORMAL, fg)

        self.send_button.spinner.modify_fg(Gtk.STATE_NORMAL, fg)

        self.progressbar.modify_fg(Gtk.STATE_NORMAL, fg)
        self.progressbar.modify_bg(Gtk.STATE_ACTIVE, bg)

        self.cancel_button.spinner.modify_fg(Gtk.STATE_NORMAL, fg)

        self.modify_fg(Gtk.STATE_NORMAL, fg)
        self.modify_bg(Gtk.STATE_NORMAL, bg)

        self.queue_draw()

        Gdk.threads_leave()


    def _on_assistant_window_hide(self, widget):
        self.assistant_window = None

        self.adding = False


    def __init__(self, setting_factory, shortener_controller, font_scheme, theme_scheme, button_factory, status_bar, account_manager, media_runner):
        self.font_scheme = font_scheme

        self.theme_scheme = theme_scheme

        self.status_bar = status_bar

        self.account_manager = account_manager

        self.media_runner = media_runner

        self.adding = False

        self.sending = False

        self.hide_on_send = True

        self.hide_on_cancel = True

        header_alignment = Gtk.Alignment(0.0, 0.5, 0.0, 0.0)
        header_alignment.add(self.account_manager.togglebutton)
        header_alignment.show()

        self.compose_button = button_factory.build(u'compose', self._on_compose_button_clicked)
        self.compose_button.box.set_tooltip_text(SHOW_COMPOSE_TOOLTIP)

        self.send_button = button_factory.build(u'send', self._on_send_button_clicked)
        self.send_button.set_toggled(True)
        self.send_button.box.set_tooltip_text(u'Send')

        self.entry_controller = entry.Controller(self.compose_button, self.send_button, setting_factory, shortener_controller, self.font_scheme, button_factory, status_bar, account_manager)

        header_box = Gtk.HBox(False, DEFAULT_BORDER)
        header_box.pack_start(self.account_manager.identifier_box, False, False, 0)
        header_box.pack_start(header_alignment, True, True, 0)
        header_box.pack_start(self.entry_controller.char_label, False, False, 0)
        header_box.pack_start(self.compose_button, False, False, 0)
        header_box.show()

        self.label = Gtk.Label(None)
        self.label.show()

        label_box = Gtk.HBox(False, 0)
        label_box.pack_start(self.label, True, True, 0)
        label_box.pack_start(self.entry_controller.link_label, True, True, 0)
        label_box.show()

        footer_alignment = Gtk.Alignment(0.0, 0.5, 0.0, 0.0)
        footer_alignment.add(self.entry_controller.unshorten_button)
        footer_alignment.show()

        self.reply_button = button_factory.build(u'reply', self._on_reply_button_clicked)
        self.reply_button.set_toggled(True)
        self.reply_button.box.set_tooltip_text(tweet.CANCEL_REPLY_TOOLTIP)
        self.reply_button.hide()

        self.message_button = button_factory.build(u'message', self._on_message_button_clicked)
        self.message_button.set_toggled(True)
        self.message_button.box.set_tooltip_text(CANCEL_MESSAGE_TOOLTIP)
        self.message_button.hide()

        self.media_button = button_factory.build(u'media', self._on_media_button_clicked)
        self.media_button.box.set_tooltip_text(u'{}\u2026'.format(media.TITLE))

        self.footer_box = Gtk.HBox(False, DEFAULT_BORDER)
        self.footer_box.pack_start(label_box, False, False, 0)
        self.footer_box.pack_start(footer_alignment, True, True, 0)
        self.footer_box.pack_start(self.reply_button, False, False, 0)
        self.footer_box.pack_start(self.message_button, False, False, 0)
        self.footer_box.pack_start(self.media_button, False, False, 0)
        self.footer_box.pack_start(self.send_button, False, False, 0)
        self.footer_box.show()

        self.progressbar = Gtk.ProgressBar()
        self.progressbar.show()

        self.cancel_button = button_factory.build(u'cancel', self._on_cancel_button_clicked)
        self.cancel_button.set_toggled(True)
        self.cancel_button.box.set_tooltip_text(u'Cancel')
        self.cancel_button.spinner.set_tooltip_text(u'Cancelling\u2026')

        self.progress_box = Gtk.HBox(False, DEFAULT_BORDER)
        self.progress_box.pack_start(self.progressbar, True, True, 0)
        self.progress_box.pack_start(self.cancel_button, False, False, 0)

        switch_box = Gtk.HBox(False, 0)
        switch_box.pack_start(self.footer_box, True, True, 0)
        switch_box.pack_start(self.progress_box, True, True, 0)
        switch_box.show()

        self.compose_box = Gtk.VBox(False, DEFAULT_BORDER)
        self.compose_box.pack_start(self.entry_controller.scrolledwindow, True, True, 0)
        self.compose_box.pack_start(switch_box, False, False, 0)

        box = Gtk.VBox(False, DEFAULT_BORDER)
        box.set_border_width(DEFAULT_BORDER)
        box.pack_start(header_box, False, False, 0)
        box.pack_start(self.compose_box, True, True, 0)
        box.show()

        super(Panel, self).__init__()
        self.add(box)
        self.show()

        self.hide_on_send_checkbutton = Gtk.CheckButton(u'Hide compose box after sending a tweet or message', False)
        self.hide_on_send_checkbutton.connect(u'toggled', self._on_hide_on_send_checkbutton_toggled)
        self.hide_on_send_checkbutton.show()

        self.hide_on_cancel_checkbutton = Gtk.CheckButton(u'Hide compose box after cancelling a reply or message', False)
        self.hide_on_cancel_checkbutton.connect(u'toggled', self._on_hide_on_cancel_checkbutton_toggled)
        self.hide_on_cancel_checkbutton.show()

        self.hide_on_send_setting = setting_factory.build_boolean(u'panel', u'hide_on_send', self._on_hide_on_send_changed)

        self.hide_on_cancel_setting = setting_factory.build_boolean(u'panel', u'hide_on_cancel', self._on_hide_on_cancel_changed)

        self._set_label_text(TWEET_TOOLTIP)

        self._update_hide_on_send()

        self._update_hide_on_cancel()

        self.hide_on_send_checkbutton.set_active(self.hide_on_send)

        self.hide_on_cancel_checkbutton.set_active(self.hide_on_cancel)

        self.font_scheme.block_updates()

        self.theme_scheme.block_updates()

        self._on_font_changed()

        self._on_theme_changed()

        self.font_uid = self.font_scheme.notifier.connect(self._on_font_changed)

        self.theme_uid = self.theme_scheme.notifier.connect(self._on_theme_changed)

        self.font_scheme.allow_updates()

        self.theme_scheme.allow_updates()

        self.replied_tweet = None

        self.messaged_user = None

        self.messaged_message = None

        self.send_thread = None

        self.create_thread = None


    def isolate(self):
        if self.create_thread is not None:
            self.create_thread.join()

        if self.send_thread is not None:
            self.send_thread.join()

        self.messaged_message = None

        self.messaged_user = None

        self.replied_tweet = None

        self.theme_scheme.block_updates()

        self.font_scheme.block_updates()

        self.theme_scheme.notifier.disconnect(self.theme_uid)

        self.font_scheme.notifier.disconnect(self.font_uid)

        self.theme_scheme.allow_updates()

        self.font_scheme.allow_updates()

        self.hide_on_cancel_setting.isolate()

        self.hide_on_send_setting.isolate()

        self.cancel_button.isolate()

        self.media_button.isolate()

        self.message_button.isolate()

        self.reply_button.isolate()

        self.entry_controller.isolate()

        self.send_button.isolate()

        self.compose_button.isolate()


    def get_adding(self):
        if self.adding:
            run_generic_error_dialog(self.get_toplevel(), u'Please close the account adding window and try again.')

        return self.adding


    def get_sending(self):
        if self.sending:
            run_generic_error_dialog(self.get_toplevel(), u'Please wait until the sending process is finished and try again.')

        return self.sending


    def get_current_account(self):
        if not self.get_adding():
            return self.account_manager.get_current_account()


    def get_other_accounts(self, account):
        accounts = []

        if not self.get_adding():
            for other_account in self.account_manager.accounts:
                if other_account != account:
                    accounts.append(other_account)

        return accounts


    def set_composing(self, composing):
        self.compose_button.set_toggled(composing)

        if composing:
            self.entry_controller.char_label.show()

            self.compose_button.box.set_tooltip_text(u'Hide compose box')

            self.compose_box.show()

            self.entry_controller.textview.grab_focus()
        else:
            self.compose_box.hide()

            self.compose_button.box.set_tooltip_text(SHOW_COMPOSE_TOOLTIP)

            self.entry_controller.char_label.hide()


    def create_group(self, stream_manager, account):
        if account is None:
            account = self.get_current_account()

        if account is not None:
            dialog = CreateGroupDialog(self.get_toplevel(), account, stream_manager)

            self.create_thread = dialog.run_and_destroy()


    def change_image(self):
        account = self.get_current_account()

        if account is not None:
            NOTIMPLEMENTEDYET(self.get_toplevel())


    def change_details(self):
        account = self.get_current_account()

        if account is not None:
            NOTIMPLEMENTEDYET(self.get_toplevel())


    def add_account(self):
        if not self.get_sending():
            if self.adding:
                self.assistant_window.present()
            else:
                self.adding = True

                self.assistant_window = assistant.Window(u'Add new account', self.account_manager)
                self.assistant_window.connect(u'hide', self._on_assistant_window_hide)

                self.assistant_window.present()


    def insert_mention(self, nick, account):
        if not self.entry_controller.get_shortening():
            if not self.get_adding():
                if not self.get_sending():
                    self.account_manager.set_account(account)

                    text = u'@{}'.format(nick)

                    self.entry_controller.insert(text)

                    self.set_composing(True)


    def compose_retweet(self, retweeter_nick, user_nick, tweet):
        if not self.entry_controller.get_shortening():
            if not self.get_adding():
                if not self.get_sending():
                    if self.replied_tweet is not None:
                        self._cancel_reply()
                    elif self.messaged_user is not None:
                        self._cancel_message_to_user()
                    elif self.messaged_message is not None:
                        self._cancel_message_to_message()

                    self.account_manager.set_account(tweet.stream.account)

                    text = u'RT @{}: {}'.format(user_nick, tweet.original_text)

                    if retweeter_nick is not None:
                        text = u'RT @{}: {}'.format(retweeter_nick, text)

                    self.entry_controller.overwrite(text)

                    self.entry_controller.home()

                    self.set_composing(True)


    def compose_reply(self, replied_tweet):
        if not self.entry_controller.get_shortening():
            if not self.get_adding():
                if not self.get_sending():
                    if self.replied_tweet is not None:
                        self._cancel_reply()
                    elif self.messaged_user is not None:
                        self._cancel_message_to_user()
                    elif self.messaged_message is not None:
                        self._cancel_message_to_message()

                    self.replied_tweet = replied_tweet

                    self.replied_tweet.set_replying(True)

                    self.account_manager.set_account(self.replied_tweet.stream.account)

                    nicks = self.replied_tweet.get_nicks()

                    words = []

                    for nick in nicks:
                        words.append(u'@{} '.format(nick))

                    self.entry_controller.overwrite(u''.join(words))

                    if not self.media_button.toggled:
                        self._set_label_text(REPLY_STRING.format(self.replied_tweet.user_nick))

                        self.reply_button.show()

                    self.set_composing(True)


    def compose_message_to_user(self, messaged_user):
        if not self.entry_controller.get_shortening():
            if not self.get_adding():
                if not self.get_sending():
                    if self.replied_tweet is not None:
                        self._cancel_reply()
                    elif self.messaged_user is not None:
                        self._cancel_message_to_user()
                    elif self.messaged_message is not None:
                        self._cancel_message_to_message()

                    self.messaged_user = messaged_user

                    self.messaged_user.set_messaging(True)

                    self.account_manager.set_account(self.messaged_user.stream.account)

                    self.entry_controller.clear()

                    if not self.media_button.toggled:
                        self._set_label_text(MESSAGE_STRING.format(self.messaged_user.user_nick))

                        self.message_button.show()

                    self.set_composing(True)


    def compose_message_to_message(self, messaged_message):
        if not self.entry_controller.get_shortening():
            if not self.get_adding():
                if not self.get_sending():
                    if self.replied_tweet is not None:
                        self._cancel_reply()
                    elif self.messaged_user is not None:
                        self._cancel_message_to_user()
                    elif self.messaged_message is not None:
                        self._cancel_message_to_message()

                    self.messaged_message = messaged_message

                    self.messaged_message.set_messaging(True)

                    self.account_manager.set_account(self.messaged_message.stream.account)

                    self.entry_controller.clear()

                    if not self.media_button.toggled:
                        self._set_label_text(MESSAGE_STRING.format(self.messaged_message.user_nick))

                        self.message_button.show()

                    self.set_composing(True)


    def cancel_reply(self):
        if not self.entry_controller.get_shortening():
            if not self.get_sending():
                self._cancel_reply()

                if self.hide_on_cancel:
                    self.set_composing(False)


    def cancel_message_to_user(self):
        if not self.entry_controller.get_shortening():
            if not self.get_sending():
                self._cancel_message_to_user()

                if self.hide_on_cancel:
                    self.set_composing(False)


    def cancel_message_to_message(self):
        if not self.entry_controller.get_shortening():
            if not self.get_sending():
                self._cancel_message_to_message()

                if self.hide_on_cancel:
                    self.set_composing(False)
