# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import gtk as Gtk

from gtk import gdk as Gdk


from polly.gui import DEFAULT_BORDER

from polly.gui.shell import menu



WINDOW_DIR = u'window'


DEFAULT_MAXIMIZED = False

DEFAULT_WIDTH = 0

DEFAULT_HEIGHT = 0



class Window(Gtk.Window):
    def run_about_dialog(self):
        self.about_runner.run(self)


    def _on_refresh_accel(self, accel_group, acceleratable, keyval, modifier):
        self.stream_manager.refresh_all()


    def _on_compose_accel(self, accel_group, acceleratable, keyval, modifier):
        toggled = self.main_panel.compose_button.toggled

        self.main_panel.set_composing(not toggled)


    def _on_move_up_accel(self, accel_group, acceleratable, keyval, modifier):
        self.account_manager.move_up()


    def _on_move_down_accel(self, accel_group, acceleratable, keyval, modifier):
        self.account_manager.move_down()


    def _on_window_state_event(self, widget, event):
        maximized = bool(event.new_window_state & Gdk.WINDOW_STATE_MAXIMIZED)

        if self.maximized != maximized:
            self.maximized = maximized

            self.maximized_setting.set_value(self.maximized)

        return False


    def _on_size_allocate(self, widget, allocation):
        self.main_panel.entry_controller.move_completer()

        if not self.maximized:
            if self.width != allocation.width:
                self.width = allocation.width

                self.width_setting.set_value(self.width)

            if self.height != allocation.height:
                self.height = allocation.height

                self.height_setting.set_value(self.height)


    def _on_focus_in_event(self, widget, event):
        self.stream_manager.reset_indicators()

        self.focused = False

        return False


    def _on_focus_in_event_after(self, widget, event):
        self.focused = True

        return False


    def _on_maximized_changed(self):
        if self.maximized_setting.get_value():
            self.maximize()
        else:
            self.unmaximize()


    def _on_width_changed(self):
        if not self.maximized:
            width = self.width_setting.get_value()

            self.set_property(u'default-width', width)


    def _on_height_changed(self):
        if not self.maximized:
            height = self.height_setting.get_value()

            self.set_property(u'default-height', height)


    def __init__(self, title, minimized, compose, setting_factory, status_bar, account_manager, main_panel, indicator_server, stream_manager, setting_window, about_runner):
        self.account_manager = account_manager

        self.main_panel = main_panel

        self.indicator_server = indicator_server

        self.stream_manager = stream_manager

        self.setting_window = setting_window

        self.about_runner = about_runner

        self.maximized = DEFAULT_MAXIMIZED

        self.width = DEFAULT_WIDTH

        self.height = DEFAULT_HEIGHT

        accel_group = Gtk.AccelGroup()
        accel_group.connect_group(Gtk.keysyms.F5, 0, Gtk.ACCEL_VISIBLE, self._on_refresh_accel)
        accel_group.connect_group(Gtk.keysyms.n, Gdk.CONTROL_MASK, Gtk.ACCEL_VISIBLE, self._on_compose_accel)
        accel_group.connect_group(Gtk.keysyms.comma, Gdk.CONTROL_MASK, Gtk.ACCEL_VISIBLE, self._on_move_up_accel)
        accel_group.connect_group(Gtk.keysyms.period, Gdk.CONTROL_MASK, Gtk.ACCEL_VISIBLE, self._on_move_down_accel)

        self.menu_bar = menu.Bar(self, accel_group)

        separator = Gtk.HSeparator()
        separator.show()

        box = Gtk.VBox(False, 0)
        box.pack_start(self.menu_bar, False, False, 0)
        box.pack_start(self.main_panel, False, False, 0)
        box.pack_start(separator, False, False, 0)
        box.pack_start(self.stream_manager.scrolledwindow, True, True, 0)
        box.pack_start(status_bar, False, False, 0)
        box.show()

        super(Window, self).__init__(Gtk.WINDOW_TOPLEVEL)
        self.set_title(title)
        self.connect(u'window-state-event', self._on_window_state_event)
        self.connect(u'size-allocate', self._on_size_allocate)
        self.connect(u'delete-event', self.hide_on_delete)
        self.connect(u'focus-in-event', self._on_focus_in_event)
        self.connect_after(u'focus-in-event', self._on_focus_in_event_after)
        self.add_accel_group(accel_group)
        self.add(box)

        self.maximized_setting = setting_factory.build_boolean(WINDOW_DIR, u'maximized', self._on_maximized_changed)

        self.width_setting = setting_factory.build_integer(WINDOW_DIR, u'width', self._on_width_changed)

        self.height_setting = setting_factory.build_integer(WINDOW_DIR, u'height', self._on_height_changed)

        self._on_maximized_changed()

        self._on_width_changed()

        self._on_height_changed()

        if self.account_manager.accounts:
            if not minimized:
                self.present()

            if compose:
                self.main_panel.set_composing(True)
        else:
            self.present()

            self.main_panel.add_account()

        self.running = True

        self.indicator_server.main_window = self


    def quit(self):
        self.stream_manager.hide_indicators()

        self.indicator_server.main_window = None

        self.setting_window.hide()

        self.main_panel.entry_controller.completer.hide()

        self.running = False

        self.hide()

        self.height_setting.isolate()

        self.width_setting.isolate()

        self.maximized_setting.isolate()

        self.menu_bar.isolate()

        Gtk.main_quit()


    def on_dialog_run(self):
        self.menu_bar.set_sensitive(False)


    def on_dialog_destroy(self):
        self.menu_bar.set_sensitive(True)


    def press(self, event):
        if event.keyval == Gtk.keysyms.Tab:
            return False

        streams = self.stream_manager.box.get_children()

        success = False

        for stream in streams:
            if stream.get_focus_child() is not None:
                if event.keyval == Gtk.keysyms.KP_Left or event.keyval == Gtk.keysyms.Left:
                    if event.state & Gdk.SHIFT_MASK:
                        self.stream_manager.move_left(stream)
                    else:
                        index = streams.index(stream)

                        if index > 0:
                            streams[index - 1].key_box.grab_focus()

                    success = True
                elif event.keyval == Gtk.keysyms.KP_Right or event.keyval == Gtk.keysyms.Right:
                    if event.state & Gdk.SHIFT_MASK:
                        self.stream_manager.move_right(stream)
                    else:
                        index = streams.index(stream)

                        if index < len(streams) - 1:
                            streams[index + 1].key_box.grab_focus()

                    success = True
                elif event.keyval == Gtk.keysyms.KP_Up or event.keyval == Gtk.keysyms.Up:
                    value = stream.adjustment.get_value()

                    if value > 0.0:
                        shift = float(10 * DEFAULT_BORDER)

                        stream.adjustment.set_value(max(value - shift, 0.0))

                    success = True
                elif event.keyval == Gtk.keysyms.KP_Down or event.keyval == Gtk.keysyms.Down:
                    value = stream.adjustment.get_value()

                    upper = stream.adjustment.get_upper() - stream.adjustment.get_page_size()

                    if value < upper:
                        shift = float(10 * DEFAULT_BORDER)

                        stream.adjustment.set_value(min(value + shift, upper))

                    success = True
                elif event.keyval == Gtk.keysyms.KP_Page_Up or event.keyval == Gtk.keysyms.Page_Up:
                    value = stream.adjustment.get_value()

                    if value > 0.0:
                        shift = stream.adjustment.get_page_size()

                        stream.adjustment.set_value(max(value - shift, 0.0))

                    success = True
                elif event.keyval == Gtk.keysyms.KP_Page_Down or event.keyval == Gtk.keysyms.Page_Down:
                    value = stream.adjustment.get_value()

                    shift = stream.adjustment.get_page_size()

                    upper = stream.adjustment.get_upper() - shift

                    if value < upper:
                        stream.adjustment.set_value(min(value + shift, upper))

                    success = True
                elif event.keyval == Gtk.keysyms.KP_Home or event.keyval == Gtk.keysyms.Home:
                    stream.adjustment.set_value(0.0)

                    success = True
                elif event.keyval == Gtk.keysyms.KP_End or event.keyval == Gtk.keysyms.End:
                    upper = stream.adjustment.get_upper() - stream.adjustment.get_page_size()

                    stream.adjustment.set_value(upper)

                    success = True

                break

        return success
