# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from threading import Lock


import gtk as Gtk

from gtk import gdk as Gdk


from polly import dispatch

from polly.twitter.account import Error

from polly.gui import DEFAULT_BORDER

from polly.gui.stream import Stream



PAGE_STRING = u'Page {}'



class Leads(Stream):
    def _build_menu(self):
        super(Leads, self)._build_menu()

        self.reset_menuitem = Gtk.MenuItem(u'Reset this stream', False)
        self.reset_menuitem.connect(u'activate', self._on_reset_menuitem_activate)
        self.reset_menuitem.show()

        self.menu.append(self.reset_menuitem)


    def _build_buttons(self):
        self.previous_button = self.button_factory.build(u'previous', self._on_previous_button_clicked)

        self.next_button = self.button_factory.build(u'next', self._on_next_button_clicked)

        Gdk.threads_enter()

        self.previous_button.set_sensitive(False)
        self.previous_button.set_toggled(True)
        self.previous_button.box.set_tooltip_text(u'Go to previous page')

        self.previous_button.box.connect(u'button-press-event', self._on_child_button_press_event)
        self.previous_button.box.connect(u'focus-in-event', self._on_child_focus_in_event)

        self.next_button.set_sensitive(False)
        self.next_button.set_toggled(True)
        self.next_button.box.set_tooltip_text(u'Go to next page')

        self.next_button.box.connect(u'button-press-event', self._on_child_button_press_event)
        self.next_button.box.connect(u'focus-in-event', self._on_child_focus_in_event)

        self.page_box.pack_start(self.previous_button, False, False, 0)
        self.page_box.reorder_child(self.previous_button, 0)
        self.page_box.pack_start(self.next_button, False, False, 0)

        Gdk.threads_leave()


    def _isolate_buttons(self):
        self.next_button.isolate()

        self.previous_button.isolate()


    def _update_reset(self):
        self.resetted = True

        self.page_number = 0


    def _update_previous(self):
        self.page_number -= 1


    def _update_next(self):
        self.page_number += 1


    def _update_page(self, previous_thread, callback, resetting):
        if previous_thread is not None:
            previous_thread.join()

        self.leads_lock.acquire()

        callback()

        page_text = PAGE_STRING.format(self.page_number + 1)

        Gdk.threads_enter()

        leads = self.lead_box.get_children()

        for lead in leads:
            self.lead_box.remove(lead)

        self.page_label.set_text(page_text)

        Gdk.threads_leave()

        self.leads_lock.release()

        self.refresh(False)

        Gdk.threads_enter()

        self.flippings -= 1

        if self.flippings == 0:
            self.previous_button.set_sensitive(False)

            self.next_button.set_sensitive(False)

            self.page_box.set_sensitive(True)

            self.lead_box.show()

        if resetting:
            self.reset_menuitem.set_sensitive(True)

        Gdk.threads_leave()

        for lead in leads:
            lead.isolate()

            Gdk.threads_enter()

            lead.destroy()

            Gdk.threads_leave()


    def _update(self):
        self.leads_lock.acquire()

        Gdk.threads_enter()

        leads = self.lead_box.get_children()

        Gdk.threads_leave()

        if not leads:
            try:
                dataset = self._get_processed_dataset()
            except Error as error:
                self.status_bar.popup(error.message)
            else:
                for data in dataset:
                    lead = self._build(data)

                    Gdk.threads_enter()

                    lead.reorder_child(lead.outer_box, 1)

                    self.lead_box.pack_start(lead, False, False, 0)

                    Gdk.threads_leave()

        self.leads_lock.release()


    def _on_flip(self, callback, resetting):
        self.lead_box.hide()

        self.page_box.set_sensitive(False)

        self.flippings += 1

        self.page_thread = dispatch(self._update_page, self.page_thread, callback, resetting)


    def _on_previous_button_clicked(self):
        self.key_box.grab_focus()

        self._on_flip(self._update_previous, False)


    def _on_next_button_clicked(self):
        self.key_box.grab_focus()

        self._on_flip(self._update_next, False)


    def _on_reset_menuitem_activate(self, menuitem):
        self.reset_menuitem.set_sensitive(False)

        self._on_flip(self._update_reset, True)


    def _on_child_button_press_event(self, widget, event):
        self.child_pressed = True

        return False


    def _on_child_focus_in_event(self, widget, event):
        if self.child_pressed:
            self.child_pressed = False
        else:
            toplevel = self.get_toplevel()

            if toplevel.focused:
                self.popup_child(self.page_box, 0)

        return False


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(Leads, self).__init__(key, account, parameter, True, color_scheme, font_scheme, button_factory, status_bar, manager)

        self.emblem_factory = emblem_factory

        self.avatar_factory = avatar_factory

        self.main_panel = main_panel

        self.resetted = False

        self.page_number = 0

        self.flippings = 0

        self.leads_lock = Lock()

        page_text = PAGE_STRING.format(1)

        self.page_label = Gtk.Label(page_text)
        self.page_label.show()

        self.page_box = Gtk.HBox(False, DEFAULT_BORDER)
        self.page_box.set_border_width(DEFAULT_BORDER)
        self.page_box.pack_start(self.page_label, True, True, 0)
        self.page_box.show()

        self.lead_box = Gtk.VBox(False, 0)
        self.lead_box.show()

        self.inner_box.pack_start(self.page_box, False, False, 0)
        self.inner_box.pack_start(self.lead_box, False, False, 0)

        self.user_backend_cache = account.backend.user_backend_cache

        self.page_thread = None

        self.child_pressed = False


    def isolate(self):
        if self.page_thread is not None:
            self.page_thread.join()

        Gdk.threads_enter()

        leads = self.lead_box.get_children()

        Gdk.threads_leave()

        for lead in leads:
            lead.isolate()

        super(Leads, self).isolate()



import users, groups, suggestions, matches
