# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from gtk import gdk as Gdk


from polly import dispatch

from polly.twitter.lead import user

from polly.twitter.account import Error

from polly.gui.twitter.lead.group import Group

from polly.gui.stream.leads import Leads



MY_MEMBERSHIPS_NAME = u'My memberships'

MY_SUBSCRIPTIONS_NAME = u'My subscriptions'

MY_GROUPS_NAME = u'My lists'



class Groups(Leads):
    def _get_processed_dataset(self):
        self.previous_cursor, self.next_cursor, group_dataset = self._get_dataset()

        Gdk.threads_enter()

        self.previous_button.set_sensitive(self.previous_cursor != 0)

        self.next_button.set_sensitive(self.next_cursor != 0)

        Gdk.threads_leave()

        return group_dataset


    def _build(self, data):
        return Group(data[0], data[1], self, self.user_backend_cache, self.group_backend_cache, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel)


    def _remove(self, previous_thread, lead):
        if previous_thread is not None:
            previous_thread.join()

        lead.isolate()

        Gdk.threads_enter()

        lead.destroy()

        Gdk.threads_leave()


    def _update_reset(self):
        super(Groups, self)._update_reset()

        self.cursor = -1


    def _update_previous(self):
        super(Groups, self)._update_previous()

        self.cursor = self.previous_cursor


    def _update_next(self):
        super(Groups, self)._update_next()

        self.cursor = self.next_cursor


    def _on_delete_lead(self):
        self.leads_lock.acquire()

        Gdk.threads_enter()

        leads = self.lead_box.get_children()

        Gdk.threads_leave()

        for lead in leads:
            if lead.group_backend.id == self.account.backend.deleted_id:
                Gdk.threads_enter()

                self.lead_box.remove(lead)

                Gdk.threads_leave()

                self.delete_thread = dispatch(self._remove, self.delete_thread, lead)

                break

        self.leads_lock.release()


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(Groups, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.cursor = -1

        self.group_backend_cache = account.backend.group_backend_cache

        self.account.backend.block_updates()

        self.delete_lead_uid = self.account.backend.delete_group_notifier.connect(self._on_delete_lead)

        self.account.backend.allow_updates()

        self.delete_thread = None


    def isolate(self):
        if self.delete_thread is not None:
            self.delete_thread.join()

        self.account.backend.block_updates()

        self.account.backend.delete_group_notifier.disconnect(self.delete_lead_uid)

        self.account.backend.allow_updates()

        super(Groups, self).isolate()



class MyMemberships(Groups):
    def _get_dataset(self):
        return self.account.backend.get_group_dataset_memberships(self.account.backend.user_backend.id, self.cursor, None)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(MyMemberships, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.label.set_text(MY_MEMBERSHIPS_NAME)
        self.label.set_tooltip_text(MY_MEMBERSHIPS_NAME)



class MySubscriptions(Groups):
    def _get_dataset(self):
        return self.account.backend.get_group_dataset_subscriptions(self.account.backend.user_backend.id, self.cursor)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(MySubscriptions, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.label.set_text(MY_SUBSCRIPTIONS_NAME)
        self.label.set_tooltip_text(MY_SUBSCRIPTIONS_NAME)



class MyGroups(Groups):
    def _get_dataset(self):
        return self.account.backend.get_group_dataset_groups(self.account.backend.user_backend.id, self.cursor)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(MyGroups, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.label.set_text(MY_GROUPS_NAME)
        self.label.set_tooltip_text(MY_GROUPS_NAME)



class UserGroups(Groups):
    def _finalize(self):
        if self.user_backend is not None:
            self.user_backend.block_updates()

            self.user_backend.nick_notifier.disconnect(self.user_nick_uid)

            self.user_backend.allow_updates()

            self.user_backend.release()


    def _update(self):
        if self.user_backend is None:
            try:
                user_data = self.account.backend.get_user_data(self.parameter)
            except Error as error:
                self.status_bar.popup(error.message)
            else:
                self.user_backend = self.user_backend_cache.acquire(user_data)

                self.user_backend.block_updates()

                self._on_user_nick_changed()

                self.user_nick_uid = self.user_backend.nick_notifier.connect(self._on_user_nick_changed)

                self.user_backend.allow_updates()

        if self.user_backend is not None:
            super(UserGroups, self)._update()


    def _on_user_nick_changed(self):
        user_nick = self.user_backend.data[user.NICK_KEY]

        markup = self.string.format(user_nick)

        Gdk.threads_enter()

        self.label.set_markup(markup)
        self.label.set_tooltip_markup(markup)

        Gdk.threads_leave()


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        # NOTE: The presence of long will not be necessary
        # after the planned port to Python 3 is completed.

        if not isinstance(parameter, (int, long)):
            raise TypeError

        super(UserGroups, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.user_backend = None



class MembershipsOfUser(UserGroups):
    def _get_dataset(self):
        return self.account.backend.get_group_dataset_memberships(self.parameter, self.cursor, None)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(MembershipsOfUser, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.string = u'Memberships of @{}'



class SubscriptionsOfUser(UserGroups):
    def _get_dataset(self):
        return self.account.backend.get_group_dataset_subscriptions(self.parameter, self.cursor)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(SubscriptionsOfUser, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.string = u'Subscriptions of @{}'



class GroupsOfUser(UserGroups):
    def _get_dataset(self):
        return self.account.backend.get_group_dataset_groups(self.parameter, self.cursor)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(GroupsOfUser, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.string = u'Lists of @{}'
