# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from time import time


from gtk import gdk as Gdk


from polly.twitter.lead import user, group

from polly.twitter.account import PROTECTED_MESSAGE, Error, UnauthorizedError

from polly.gui.twitter.lead.user import User

from polly.gui.twitter.post.tweet import Tweet

from polly.gui.stream import PAGE_SIZE

from polly.gui.stream.leads import Leads



FOLLOWERS_NAME = u'My followers'

FRIENDS_NAME = u'My friends'

RETWEETERS_NAME = u'Retweeters'



class Users(Leads):
    def _get_processed_dataset(self):
        if self.resetted:
            self.resetted = False

            self.user_ids = None

        if self.user_ids is None:
            self.user_ids = self._get_user_ids()

        index = self.page_number * PAGE_SIZE

        Gdk.threads_enter()

        self.previous_button.set_sensitive(self.page_number > 0)

        self.next_button.set_sensitive(index + PAGE_SIZE < len(self.user_ids))

        Gdk.threads_leave()

        if index < len(self.user_ids):
            slice = self.user_ids[index:(index + PAGE_SIZE)]

            return self.account.backend.get_user_dataset(slice)
        else:
            return []


    def _build(self, data):
        return User(data, self, self.user_backend_cache, self.relationship_backend_cache, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(Users, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.user_ids = None

        self.relationship_backend_cache = account.backend.relationship_backend_cache



class MyFollowers(Users):
    def _get_user_ids(self):
        return self.account.backend.get_user_ids_followers(self.account.backend.user_backend.id)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(MyFollowers, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.label.set_text(FOLLOWERS_NAME)
        self.label.set_tooltip_text(FOLLOWERS_NAME)



class MyFriends(Users):
    def _get_user_ids(self):
        return self.account.backend.get_user_ids_friends(self.account.backend.user_backend.id)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(MyFriends, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.label.set_text(FRIENDS_NAME)
        self.label.set_tooltip_text(FRIENDS_NAME)



class UserUsers(Users):
    def _get_user_ids(self):
        try:
            ids = self._get_protected_user_ids()
        except UnauthorizedError:
            self.status_bar.popup(PROTECTED_MESSAGE)

            ids = []

        return ids


    def _finalize(self):
        if self.user_backend is not None:
            self.user_backend.block_updates()

            self.user_backend.nick_notifier.disconnect(self.user_nick_uid)

            self.user_backend.allow_updates()

            self.user_backend.release()


    def _update(self):
        if self.user_backend is None:
            try:
                user_data = self.account.backend.get_user_data(self.parameter)
            except Error as error:
                self.status_bar.popup(error.message)
            else:
                self.user_backend = self.user_backend_cache.acquire(user_data)

                self.user_backend.block_updates()

                self._on_user_nick_changed()

                self.user_nick_uid = self.user_backend.nick_notifier.connect(self._on_user_nick_changed)

                self.user_backend.allow_updates()

        if self.user_backend is not None:
            super(UserUsers, self)._update()


    def _on_user_nick_changed(self):
        user_nick = self.user_backend.data[user.NICK_KEY]

        markup = self.string.format(user_nick)

        Gdk.threads_enter()

        self.label.set_markup(markup)
        self.label.set_tooltip_markup(markup)

        Gdk.threads_leave()


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        # NOTE: The presence of long will not be necessary
        # after the planned port to Python 3 is completed.

        if not isinstance(parameter, (int, long)):
            raise TypeError

        super(UserUsers, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.user_backend = None



class FollowersOfUser(UserUsers):
    def _get_protected_user_ids(self):
        return self.account.backend.get_user_ids_followers(self.parameter)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(FollowersOfUser, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.string = u'Followers of @{}'



class FriendsOfUser(UserUsers):
    def _get_protected_user_ids(self):
        return self.account.backend.get_user_ids_friends(self.parameter)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(FriendsOfUser, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.string = u'Friends of @{}'



class Retweeters(Users):
    def _get_user_ids(self):
        return self.account.backend.get_user_ids_retweeters(self.parameter)


    def _finalize(self):
        if self.tweet is not None:
            self.tweet.isolate()

            Gdk.threads_enter()

            self.tweet = None

            Gdk.threads_leave()


    def _update(self):
        if self.tweet is None:
            try:
                data = self.account.backend.get_tweet_data(self.parameter)
            except Error as error:
                self.status_bar.popup(error.message)
            else:
                timestamp = time()

                tweet = Tweet(data[0], data[1], data[2], data[3], self, self.user_backend_cache, self.tweet_backend_cache, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel)

                Gdk.threads_enter()

                self.tweet = tweet

                Gdk.threads_leave()

                self.tweet.synchronize(timestamp)

                Gdk.threads_enter()

                self.inner_box.pack_start(self.tweet, False, False, 0)
                self.inner_box.reorder_child(self.tweet, 0)

                Gdk.threads_leave()

        if self.tweet is not None:
            super(Retweeters, self)._update()


    def _on_delete_post(self):
        if self.parameter == self.account.backend.deleted_id:
            Gdk.threads_enter()

            self.manager.close(self)

            Gdk.threads_leave()


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, preview_factory, status_bar, main_panel, manager):
        # NOTE: The presence of long will not be necessary
        # after the planned port to Python 3 is completed.

        if not isinstance(parameter, (int, long)):
            raise TypeError

        super(Retweeters, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.preview_factory = preview_factory

        self.tweet_backend_cache = account.backend.tweet_backend_cache

        self.tweet = None

        self.label.set_text(RETWEETERS_NAME)
        self.label.set_tooltip_text(RETWEETERS_NAME)

        self.account.backend.block_updates()

        self.delete_post_uid = self.account.backend.delete_tweet_notifier.connect(self._on_delete_post)

        self.account.backend.allow_updates()


    def isolate(self):
        self.account.backend.block_updates()

        self.account.backend.delete_tweet_notifier.disconnect(self.delete_post_uid)

        self.account.backend.allow_updates()

        super(Retweeters, self).isolate()


    def synchronize(self):
        Gdk.threads_enter()

        tweet = self.tweet

        Gdk.threads_leave()

        if tweet is not None:
            timestamp = time()

            tweet.synchronize(timestamp)



class GroupUsers(Users):
    def _get_processed_dataset(self):
        self.previous_cursor, self.next_cursor, user_dataset = self._get_dataset()

        Gdk.threads_enter()

        self.previous_button.set_sensitive(self.previous_cursor != 0)

        self.next_button.set_sensitive(self.next_cursor != 0)

        Gdk.threads_leave()

        return user_dataset


    def _update_reset(self):
        super(GroupUsers, self)._update_reset()

        self.cursor = -1


    def _update_previous(self):
        super(GroupUsers, self)._update_previous()

        self.cursor = self.previous_cursor


    def _update_next(self):
        super(GroupUsers, self)._update_next()

        self.cursor = self.next_cursor


    def _finalize(self):
        if self.group_backend is not None:
            self.group_backend.block_updates()

            self.group_backend.path_notifier.disconnect(self.group_path_uid)

            self.group_backend.allow_updates()

            self.group_backend.release()

            self.user_backend.release()


    def _update(self):
        if self.group_backend is None:
            try:
                data = self.account.backend.get_group_data(self.parameter)
            except Error as error:
                self.status_bar.popup(error.message)
            else:
                self.user_backend = self.user_backend_cache.acquire(data[0])

                self.group_backend = self.group_backend_cache.acquire(data[1])

                self.group_backend.block_updates()

                self._on_group_path_changed()

                self.group_path_uid = self.group_backend.path_notifier.connect(self._on_group_path_changed)

                self.group_backend.allow_updates()

        if self.group_backend is not None:
            super(GroupUsers, self)._update()


    def _on_group_path_changed(self):
        group_path = self.group_backend.data[group.PATH_KEY]

        markup = self.string.format(group_path)

        Gdk.threads_enter()

        self.label.set_markup(markup)
        self.label.set_tooltip_markup(markup)

        Gdk.threads_leave()


    def _on_delete_lead(self):
        if self.parameter == self.account.backend.deleted_id:
            Gdk.threads_enter()

            self.manager.close(self)

            Gdk.threads_leave()


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        # NOTE: The presence of long will not be necessary
        # after the planned port to Python 3 is completed.

        if not isinstance(parameter, (int, long)):
            raise TypeError

        super(GroupUsers, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.cursor = -1

        self.user_backend = None

        self.group_backend = None

        self.group_backend_cache = account.backend.group_backend_cache

        self.account.backend.block_updates()

        self.delete_lead_uid = self.account.backend.delete_group_notifier.connect(self._on_delete_lead)

        self.account.backend.allow_updates()


    def isolate(self):
        self.account.backend.block_updates()

        self.account.backend.delete_group_notifier.disconnect(self.delete_lead_uid)

        self.account.backend.allow_updates()

        super(GroupUsers, self).isolate()



class MembersOfGroup(GroupUsers):
    def _get_dataset(self):
        return self.account.backend.get_user_dataset_members(self.parameter, self.cursor)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(MembersOfGroup, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.string = u'Members of {}'



class SubscribersOfGroup(GroupUsers):
    def _get_dataset(self):
        return self.account.backend.get_user_dataset_subscribers(self.parameter, self.cursor)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager):
        super(SubscribersOfGroup, self).__init__(key, account, parameter, color_scheme, font_scheme, emblem_factory, button_factory, avatar_factory, status_bar, main_panel, manager)

        self.string = u'Subscribers of {}'
