# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import json, os, shutil

from threading import Lock, Condition


import glib as GLib

import gtk as Gtk

from gtk import gdk as Gdk


from polly import dispatch

from polly.gui import DEFAULT_BORDER

from polly.gui.stream.leads import users, groups

from polly.gui.stream.leads.suggestions import Suggestions

from polly.gui.stream.posts import tweets, messages, realtime

from polly.gui.stream.posts.results import Results

from polly.gui.stream.topics.categories import Categories

from polly.gui.stream.topics.trends import Trends



DEFAULT_INTERVAL = 5

LOWER_INTERVAL = 2

UPPER_INTERVAL = 60

DEFAULT_MAXIMUM = 40

LOWER_MAXIMUM = 20

UPPER_MAXIMUM = 800

DEFAULT_POLICY = 2


TWEETS_HOME_MENTIONS_KEY = u'tweets_home_mentions'

TWEETS_HOME_KEY = u'tweets_home'

TWEETS_MENTIONS_KEY = u'tweets_mentions'

TWEETS_SENT_BY_ME_KEY = u'tweets_sent_by_me'

TWEETS_FAVORITED_BY_ME_KEY = u'tweets_favorited_by_me'

TWEETS_RETWEETED_BY_ME_KEY = u'tweets_retweeted_by_me'

TWEETS_RETWEETED_TO_ME_KEY = u'tweets_retweeted_to_me'

TWEETS_RETWEETED_FROM_ME_KEY = u'tweets_retweeted_from_me'

TWEETS_CONTEXT_KEY = u'tweets_context'

TWEETS_HOME_OF_USER_KEY = u'tweets_home_of_user'

TWEETS_SENT_BY_USER_KEY = u'tweets_sent_by_user'

TWEETS_FAVORITED_BY_USER_KEY = u'tweets_favorited_by_user'

TWEETS_HOME_OF_GROUP_KEY = u'tweets_home_of_group'

TWEETS_TWEETS_OF_GROUP_KEY = u'tweets_tweets_of_group'

USERS_MY_FOLLOWERS_KEY = u'users_my_followers'

USERS_MY_FRIENDS_KEY = u'users_my_friends'

USERS_FOLLOWERS_OF_USER_KEY = u'users_followers_of_user'

USERS_FRIENDS_OF_USER_KEY = u'users_friends_of_user'

USERS_RETWEETERS_KEY = u'users_retweeters'

USERS_MEMBERS_OF_GROUP_KEY = u'users_members_of_group'

USERS_SUBSCRIBERS_OF_GROUP_KEY = u'users_subscribers_of_group'

GROUPS_MY_MEMBERSHIPS_KEY = u'groups_my_memberships'

GROUPS_MY_SUBSCRIPTIONS_KEY = u'groups_my_subscriptions'

GROUPS_MY_GROUPS_KEY = u'groups_my_groups'

GROUPS_MEMBERSHIPS_OF_USER_KEY = u'groups_memberships_of_user'

GROUPS_SUBSCRIPTIONS_OF_USER_KEY = u'groups_subscriptions_of_user'

GROUPS_GROUPS_OF_USER_KEY = u'groups_groups_of_user'

MESSAGES_ALL_KEY = u'messages_all'

MESSAGES_RECEIVED_KEY = u'messages_received'

MESSAGES_SENT_KEY = u'messages_sent'

CATEGORIES_KEY = u'categories'

SUGGESTIONS_KEY = u'suggestions'

TRENDS_KEY = u'trends'

RESULTS_KEY = u'results'



class Log(object):
    def __init__(self):
        self.counts = {}

        self.lock = Lock()


    def increment(self, id):
        self.lock.acquire()

        if id not in self.counts.viewkeys():
            self.counts[id] = 0

        self.counts[id] += 1

        count = self.counts[id]

        self.lock.release()

        return count


    def decrement(self, id):
        self.lock.acquire()

        self.counts[id] -= 1

        if self.counts[id] == 0:
            del self.counts[id]

        self.lock.release()



class StreamManager(object):
    def _add(self, key, account, parameter, old_stream, online):
        if key == TWEETS_HOME_MENTIONS_KEY:
            new_stream = realtime.tweets.HomeMentions(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self.indicator_server, self)
        elif key == TWEETS_HOME_KEY:
            new_stream = realtime.tweets.Home(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self.indicator_server, self)
        elif key == TWEETS_MENTIONS_KEY:
            new_stream = realtime.tweets.Mentions(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self.indicator_server, self)
        elif key == TWEETS_SENT_BY_ME_KEY:
            new_stream = realtime.tweets.SentByMe(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == TWEETS_FAVORITED_BY_ME_KEY:
            new_stream = tweets.FavoritedByMe(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == TWEETS_RETWEETED_BY_ME_KEY:
            new_stream = realtime.tweets.RetweetedByMe(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == TWEETS_RETWEETED_TO_ME_KEY:
            new_stream = realtime.tweets.RetweetedToMe(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == TWEETS_RETWEETED_FROM_ME_KEY:
            new_stream = tweets.RetweetedFromMe(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == TWEETS_CONTEXT_KEY:
            new_stream = tweets.Context(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == TWEETS_HOME_OF_USER_KEY:
            new_stream = tweets.HomeOfUser(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == TWEETS_SENT_BY_USER_KEY:
            new_stream = tweets.SentByUser(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == TWEETS_FAVORITED_BY_USER_KEY:
            new_stream = tweets.FavoritedByUser(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == TWEETS_HOME_OF_GROUP_KEY:
            new_stream = tweets.HomeOfGroup(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self.indicator_server, self)
        elif key == TWEETS_TWEETS_OF_GROUP_KEY:
            new_stream = tweets.TweetsOfGroup(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == USERS_MY_FOLLOWERS_KEY:
            new_stream = users.MyFollowers(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == USERS_MY_FRIENDS_KEY:
            new_stream = users.MyFriends(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == USERS_FOLLOWERS_OF_USER_KEY:
            new_stream = users.FollowersOfUser(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == USERS_FRIENDS_OF_USER_KEY:
            new_stream = users.FriendsOfUser(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == USERS_RETWEETERS_KEY:
            new_stream = users.Retweeters(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == USERS_MEMBERS_OF_GROUP_KEY:
            new_stream = users.MembersOfGroup(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == USERS_SUBSCRIBERS_OF_GROUP_KEY:
            new_stream = users.SubscribersOfGroup(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == GROUPS_MY_MEMBERSHIPS_KEY:
            new_stream = groups.MyMemberships(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == GROUPS_MY_SUBSCRIPTIONS_KEY:
            new_stream = groups.MySubscriptions(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == GROUPS_MY_GROUPS_KEY:
            new_stream = groups.MyGroups(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == GROUPS_MEMBERSHIPS_OF_USER_KEY:
            new_stream = groups.MembershipsOfUser(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == GROUPS_SUBSCRIPTIONS_OF_USER_KEY:
            new_stream = groups.SubscriptionsOfUser(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == GROUPS_GROUPS_OF_USER_KEY:
            new_stream = groups.GroupsOfUser(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == MESSAGES_ALL_KEY:
            new_stream = messages.All(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self.indicator_server, self)
        elif key == MESSAGES_RECEIVED_KEY:
            new_stream = messages.Received(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == MESSAGES_SENT_KEY:
            new_stream = messages.Sent(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        elif key == CATEGORIES_KEY:
            new_stream = Categories(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.status_bar, self)
        elif key == SUGGESTIONS_KEY:
            new_stream = Suggestions(key, account, parameter, self.color_scheme, self.font_scheme, self.emblem_factory, self.button_factory, self.avatar_factory, self.status_bar, self.main_panel, self)
        elif key == TRENDS_KEY:
            new_stream = Trends(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.status_bar, self)
        elif key == RESULTS_KEY:
            new_stream = Results(key, account, parameter, self.color_scheme, self.font_scheme, self.button_factory, self.stamp_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel, self)
        else:
            raise KeyError

        streams = self.box.get_children()

        if old_stream is None:
            index = len(streams)
        else:
            index = streams.index(old_stream) + 1

        if online:
            self.new_stream = new_stream

            self.new_stream.start()

        self.box.pack_start(new_stream, True, True, 0)
        self.box.reorder_child(new_stream, index)


    def _remove(self, previous_thread, stream):
        if previous_thread is not None:
            previous_thread.join()

        stream.stop()

        stream.isolate()

        Gdk.threads_enter()

        stream.destroy()

        Gdk.threads_leave()


    def _new(self, name, account, parameter, old_stream):
        self._add(name, account, parameter, old_stream, True)

        self.save()


    def _close(self, stream):
        streams = self.box.get_children()

        if stream in streams:
            stream.hide_indicator()

            self.box.remove(stream)

            self.close_thread = dispatch(self._remove, self.close_thread, stream)


    def _scroll_left(self):
        if self.drag_scroll == -1:
            value = self.adjustment.get_value()

            if value > 0.0:
                shift = float(4 * DEFAULT_BORDER)

                value = max(value - shift, 0.0)

                self.adjustment.set_value(value)

            return True

        return False


    def _scroll_right(self):
        if self.drag_scroll == 1:
            value = self.adjustment.get_value()

            upper = self.adjustment.get_upper() - self.adjustment.get_page_size()

            if value < upper:
                shift = float(4 * DEFAULT_BORDER)

                value = min(value + shift, upper)

                self.adjustment.set_value(value)

            return True

        return False


    def _update_interval(self):
        interval = self.interval_setting.get_value()

        if interval < LOWER_INTERVAL or interval > UPPER_INTERVAL:
            interval = DEFAULT_INTERVAL

        if self.interval == interval:
            return False

        self.interval_lock.acquire()

        self.interval = interval

        self.interval_lock.release()

        return True


    def _update_maximum(self):
        maximum = self.maximum_setting.get_value()

        if maximum < LOWER_MAXIMUM or maximum > UPPER_MAXIMUM:
            maximum = DEFAULT_MAXIMUM

        if self.maximum == maximum:
            return False

        self.maximum_lock.acquire()

        self.maximum = maximum

        self.maximum_lock.release()

        return True


    def _update_policy(self):
        policy = self.policy_setting.get_value()

        if policy < 0 or policy > 4:
            policy = DEFAULT_POLICY

        if self.policy == policy:
            return False

        self.policy_lock.acquire()

        self.policy = policy

        self.policy_lock.release()

        return True


    def _update_preview(self):
        preview = self.preview_setting.get_value()

        if self.preview == preview:
            return False

        self.preview_lock.acquire()

        self.preview = preview

        self.preview_lock.release()

        return True


    def _on_size_allocate(self, widget, allocation):
        if self.new_stream is not None:
            self.new_stream.popup()

            self.new_stream.key_box.grab_focus()

            self.new_stream = None


    def _on_drag_motion(self, widget, drag_context, x, y, time):
        if self.drag_source is not None:
            allocation = self.viewport.get_allocation()

            if x < 4 * DEFAULT_BORDER:
                if self.drag_scroll != -1:
                    self.drag_scroll = -1

                    GLib.timeout_add(40, self._scroll_left)
            elif x < allocation.width - 4 * DEFAULT_BORDER:
                self.drag_scroll = 0
            else:
                if self.drag_scroll != 1:
                    self.drag_scroll = 1

                    GLib.timeout_add(40, self._scroll_right)

            x += long(self.adjustment.get_value())

            for stream in self.box.get_children():
                if stream != self.drag_source:
                    allocation = stream.get_allocation()

                    if x >= allocation.x and x < allocation.x + allocation.width:
                        if self.drag_dest_candidate is None:
                            success = True
                        else:
                            if self.drag_dest_candidate == stream:
                                success = False
                            else:
                                success = True

                                self.drag_dest_candidate.drag_unhighlight()

                        if success:
                            self.drag_dest_candidate = stream

                            self.drag_dest_candidate.drag_highlight()

                        drag_context.drag_status(Gdk.ACTION_MOVE, time)

                        return True

        if self.drag_dest_candidate is not None:
            self.drag_dest_candidate.drag_unhighlight()

            self.drag_dest_candidate = None

        drag_context.drag_status(Gdk.ACTION_DEFAULT, time)

        return True


    def _on_drag_leave(self, widget, drag_context, time):
        if self.drag_dest_candidate is not None:
            self.drag_dest = self.drag_dest_candidate

            self.drag_dest_candidate.drag_unhighlight()

            self.drag_dest_candidate = None


    def _on_drag_drop(self, widget, drag_context, x, y, time):
        self.drag_move()

        drag_context.finish(True, False, time)

        return True


    def _on_interval_spinbutton_output(self, spin_button):
        interval = self.interval_spinbutton.get_value_as_int()

        if self.interval != interval:
            self.interval_setting.set_value(interval)

        return False


    def _on_maximum_spinbutton_output(self, spin_button):
        maximum = self.maximum_spinbutton.get_value_as_int()

        if self.maximum != maximum:
            self.maximum_setting.set_value(maximum)

        return False


    def _on_policy_radiobutton_toggled(self, togglebutton):
        if togglebutton.get_active():
            policy = self.policy_radiobuttons.index(togglebutton)

            if self.policy != policy:
                self.policy_setting.set_value(policy)


    def _on_preview_checkbutton_toggled(self, togglebutton):
        preview = self.preview_checkbutton.get_active()

        if self.preview != preview:
            self.preview_setting.set_value(preview)


    def _on_interval_changed(self):
        if self._update_interval():
            self.refresh_all()

            self.interval_spinbutton.set_value(self.interval)


    def _on_maximum_changed(self):
        if self._update_maximum():
            self.maximum_spinbutton.set_value(self.maximum)


    def _on_policy_changed(self):
        if self._update_policy():
            self.policy_radiobuttons[self.policy].set_active(True)


    def _on_preview_changed(self):
        if self._update_preview():
            self.preview_checkbutton.set_active(self.preview)


    def _update_loop(self):
        while True:
            Gdk.threads_enter()

            streams = self.box.get_children()

            Gdk.threads_leave()

            for stream in streams:
                stream.synchronize()

            self.update_condition.acquire()

            if not self.update_finished:
                self.update_condition.wait(60.0)

            update_finished = self.update_finished

            self.update_condition.release()

            if update_finished:
                break


    def __init__(self, config_path, config_path_lock, setting_factory, color_scheme, font_scheme, emblem_factory, button_factory, stamp_factory, avatar_factory, preview_factory, status_bar, account_manager, main_panel, indicator_server):
        self.config_path = config_path

        self.config_path_lock = config_path_lock

        self.color_scheme = color_scheme

        self.font_scheme = font_scheme

        self.emblem_factory = emblem_factory

        self.button_factory = button_factory

        self.stamp_factory = stamp_factory

        self.avatar_factory = avatar_factory

        self.preview_factory = preview_factory

        self.status_bar = status_bar

        self.main_panel = main_panel

        self.indicator_server = indicator_server

        self.interval = DEFAULT_INTERVAL

        self.maximum = DEFAULT_MAXIMUM

        self.policy = DEFAULT_POLICY

        self.preview = True

        self.update_finished = False

        self.path = os.path.join(self.config_path, u'streams')

        self.interval_lock = Lock()

        self.maximum_lock = Lock()

        self.policy_lock = Lock()

        self.preview_lock = Lock()

        self.update_condition = Condition()

        self.new_stream = None

        self.box = Gtk.HBox(True, DEFAULT_BORDER - 4)
        self.box.connect(u'size-allocate', self._on_size_allocate)
        self.box.show()

        self.scrolledwindow = Gtk.ScrolledWindow(None, None)
        self.scrolledwindow.set_border_width(DEFAULT_BORDER - 2)
        self.scrolledwindow.set_policy(Gtk.POLICY_AUTOMATIC, Gtk.POLICY_NEVER)
        self.scrolledwindow.add_with_viewport(self.box)
        self.scrolledwindow.show()

        self.adjustment = self.scrolledwindow.get_hadjustment()

        self.viewport = self.scrolledwindow.get_child()
        self.viewport.set_shadow_type(Gtk.SHADOW_NONE)
        self.viewport.drag_dest_set(0, [], 0)
        self.viewport.connect(u'drag-motion', self._on_drag_motion)
        self.viewport.connect(u'drag-leave', self._on_drag_leave)
        self.viewport.connect(u'drag-drop', self._on_drag_drop)

        interval_adjustment = Gtk.Adjustment(float(DEFAULT_INTERVAL), float(LOWER_INTERVAL), float(UPPER_INTERVAL), 1.0, 5.0, 0.0)

        self.interval_spinbutton = Gtk.SpinButton(interval_adjustment, 0.0, 0)
        self.interval_spinbutton.connect(u'output', self._on_interval_spinbutton_output)
        self.interval_spinbutton.show()

        maximum_adjustment = Gtk.Adjustment(float(DEFAULT_MAXIMUM), float(LOWER_MAXIMUM), float(UPPER_MAXIMUM), 20.0, 100.0, 0.0)

        self.maximum_spinbutton = Gtk.SpinButton(maximum_adjustment, 0.0, 0)
        self.maximum_spinbutton.connect(u'output', self._on_maximum_spinbutton_output)
        self.maximum_spinbutton.show()

        self.policy_radiobuttons = []

        self.policy_radiobuttons.append(Gtk.RadioButton(None,
                                                        u'Never show notification bubbles', False))
        self.policy_radiobuttons[0].connect(u'toggled', self._on_policy_radiobutton_toggled)
        self.policy_radiobuttons[0].show()

        self.policy_radiobuttons.append(Gtk.RadioButton(self.policy_radiobuttons[0],
                                                        u'Show notification bubbles for messages', False))
        self.policy_radiobuttons[1].connect(u'toggled', self._on_policy_radiobutton_toggled)
        self.policy_radiobuttons[1].show()

        self.policy_radiobuttons.append(Gtk.RadioButton(self.policy_radiobuttons[0],
                                                        u'Show notification bubbles for messages and mentions', False))
        self.policy_radiobuttons[2].connect(u'toggled', self._on_policy_radiobutton_toggled)
        self.policy_radiobuttons[2].show()

        self.policy_radiobuttons.append(Gtk.RadioButton(self.policy_radiobuttons[0],
                                                        u'Show notification bubbles for everything except retweets', False))
        self.policy_radiobuttons[3].connect(u'toggled', self._on_policy_radiobutton_toggled)
        self.policy_radiobuttons[3].show()

        self.policy_radiobuttons.append(Gtk.RadioButton(self.policy_radiobuttons[0],
                                                        u'Always show notification bubbles', False))
        self.policy_radiobuttons[4].connect(u'toggled', self._on_policy_radiobutton_toggled)
        self.policy_radiobuttons[4].show()

        self.preview_checkbutton = Gtk.CheckButton(u'Show image previews', False)
        self.preview_checkbutton.connect(u'toggled', self._on_preview_checkbutton_toggled)
        self.preview_checkbutton.show()

        self.interval_setting = setting_factory.build_integer(u'stream', u'interval', self._on_interval_changed)

        self.maximum_setting = setting_factory.build_integer(u'stream', u'maximum', self._on_maximum_changed)

        self.policy_setting = setting_factory.build_integer(u'stream', u'policy', self._on_policy_changed)

        self.preview_setting = setting_factory.build_boolean(u'stream', u'preview', self._on_preview_changed)

        self.tweet_log = Log()

        self.message_log = Log()

        try:
            with open(self.path) as file:
                dataset = json.load(file)

            for spec in dataset:
                try:
                    key, account_id, parameter = spec

                    # NOTE: The presence of long will not be necessary
                    # after the planned port to Python 3 is completed.

                    if not isinstance(account_id, (int, long)):
                        raise TypeError

                    account = account_manager.get_account(account_id)

                    if account is None:
                        raise ValueError

                    self._add(key, account, parameter, None, False)
                except (TypeError, KeyError, ValueError):
                    pass
        except (IOError, ValueError, TypeError):
            pass

        self.save()

        account_manager.stream_manager = self

        self._update_interval()

        self._update_maximum()

        self._update_policy()

        self._update_preview()

        self.interval_spinbutton.set_value(self.interval)

        self.maximum_spinbutton.set_value(self.maximum)

        self.policy_radiobuttons[self.policy].set_active(True)

        self.preview_checkbutton.set_active(self.preview)

        self.current_stream = None

        self.drag_source = None

        self.drag_scroll = 0

        self.drag_dest_candidate = None

        self.drag_dest = None

        self.close_thread = None


    def isolate(self):
        if self.close_thread is not None:
            self.close_thread.join()

        for stream in self.box.get_children():
            stream.isolate()

        self.preview_setting.isolate()

        self.policy_setting.isolate()

        self.maximum_setting.isolate()

        self.interval_setting.isolate()


    def start_all(self):
        self.update_thread = dispatch(self._update_loop)

        for stream in self.box.get_children():
            stream.start()


    def refresh_all(self):
        for stream in self.box.get_children():
            if stream.refreshable:
                stream.refresh(False)


    def stop_all(self):
        for stream in self.box.get_children():
            stream.stop()

        self.update_condition.acquire()

        self.update_finished = True

        self.update_condition.notify()

        self.update_condition.release()

        self.update_thread.join()


    def hide_indicators(self):
        for stream in self.box.get_children():
            stream.hide_indicator()


    def reset_indicators(self):
        for stream in self.box.get_children():
            stream.reset_indicator()


    def get_interval(self):
        self.interval_lock.acquire()

        interval = self.interval

        self.interval_lock.release()

        return interval


    def get_maximum(self):
        self.maximum_lock.acquire()

        maximum = self.maximum

        self.maximum_lock.release()

        return maximum


    def get_policy(self):
        self.policy_lock.acquire()

        policy = self.policy

        self.policy_lock.release()

        return policy


    def get_preview(self):
        self.preview_lock.acquire()

        preview = self.preview

        self.preview_lock.release()

        return preview


    def save(self):
        dataset = []

        for stream in self.box.get_children():
            dataset.append([stream.key, stream.account.id, stream.parameter])
 
        try:
            self.config_path_lock.acquire()

            try:
                if not os.path.isdir(self.config_path):
                    os.makedirs(self.config_path)
            finally:
                self.config_path_lock.release()

            if os.path.isdir(self.path):
                shutil.rmtree(self.path)

            with open(self.path, u'w') as file:
                json.dump(dataset, file)
        except (OSError, IOError):
            pass


    def new_tweets_home_mentions(self, account):
        self._new(TWEETS_HOME_MENTIONS_KEY, account, None, None)


    def new_tweets_home(self, account):
        self._new(TWEETS_HOME_KEY, account, None, None)


    def new_tweets_mentions(self, account):
        self._new(TWEETS_MENTIONS_KEY, account, None, None)


    def new_tweets_sent_by_me(self, account):
        self._new(TWEETS_SENT_BY_ME_KEY, account, None, None)


    def new_tweets_favorited_by_me(self, account):
        self._new(TWEETS_FAVORITED_BY_ME_KEY, account, None, None)


    def new_tweets_retweeted_by_me(self, account):
        self._new(TWEETS_RETWEETED_BY_ME_KEY, account, None, None)


    def new_tweets_retweeted_to_me(self, account):
        self._new(TWEETS_RETWEETED_TO_ME_KEY, account, None, None)


    def new_tweets_retweeted_from_me(self, account):
        self._new(TWEETS_RETWEETED_FROM_ME_KEY, account, None, None)


    def new_tweets_context(self, account, parameter, stream):
        self._new(TWEETS_CONTEXT_KEY, account, parameter, stream)


    def new_tweets_home_of_user(self, account, parameter, stream):
        self._new(TWEETS_HOME_OF_USER_KEY, account, parameter, stream)


    def new_tweets_sent_by_user(self, account, parameter, stream):
        self._new(TWEETS_SENT_BY_USER_KEY, account, parameter, stream)


    def new_tweets_favorited_by_user(self, account, parameter, stream):
        self._new(TWEETS_FAVORITED_BY_USER_KEY, account, parameter, stream)


    def new_tweets_home_of_group(self, account, parameter, stream):
        self._new(TWEETS_HOME_OF_GROUP_KEY, account, parameter, stream)


    def new_tweets_tweets_of_group(self, account, parameter, stream):
        self._new(TWEETS_TWEETS_OF_GROUP_KEY, account, parameter, stream)


    def new_users_my_followers(self, account):
        self._new(USERS_MY_FOLLOWERS_KEY, account, None, None)


    def new_users_my_friends(self, account):
        self._new(USERS_MY_FRIENDS_KEY, account, None, None)


    def new_users_followers_of_user(self, account, parameter, stream):
        self._new(USERS_FOLLOWERS_OF_USER_KEY, account, parameter, stream)


    def new_users_friends_of_user(self, account, parameter, stream):
        self._new(USERS_FRIENDS_OF_USER_KEY, account, parameter, stream)


    def new_users_retweeters(self, account, parameter, stream):
        self._new(USERS_RETWEETERS_KEY, account, parameter, stream)


    def new_users_members_of_group(self, account, parameter, stream):
        self._new(USERS_MEMBERS_OF_GROUP_KEY, account, parameter, stream)


    def new_users_subscribers_of_group(self, account, parameter, stream):
        self._new(USERS_SUBSCRIBERS_OF_GROUP_KEY, account, parameter, stream)


    def new_groups_my_memberships(self, account):
        self._new(GROUPS_MY_MEMBERSHIPS_KEY, account, None, None)


    def new_groups_my_subscriptions(self, account):
        self._new(GROUPS_MY_SUBSCRIPTIONS_KEY, account, None, None)


    def new_groups_my_groups(self, account):
        self._new(GROUPS_MY_GROUPS_KEY, account, None, None)


    def new_groups_memberships_of_user(self, account, parameter, stream):
        self._new(GROUPS_MEMBERSHIPS_OF_USER_KEY, account, parameter, stream)


    def new_groups_subscriptions_of_user(self, account, parameter, stream):
        self._new(GROUPS_SUBSCRIPTIONS_OF_USER_KEY, account, parameter, stream)


    def new_groups_groups_of_user(self, account, parameter, stream):
        self._new(GROUPS_GROUPS_OF_USER_KEY, account, parameter, stream)


    def new_messages_all(self, account):
        self._new(MESSAGES_ALL_KEY, account, None, None)


    def new_messages_received(self, account):
        self._new(MESSAGES_RECEIVED_KEY, account, None, None)


    def new_messages_sent(self, account):
        self._new(MESSAGES_SENT_KEY, account, None, None)


    def new_categories(self, account):
        self._new(CATEGORIES_KEY, account, None, None)


    def new_suggestions(self, account, parameter, stream):
        self._new(SUGGESTIONS_KEY, account, parameter, stream)


    def new_trends(self, account):
        self._new(TRENDS_KEY, account, None, None)


    def new_results(self, account, parameter, stream):
        self._new(RESULTS_KEY, account, parameter, stream)


    def close(self, stream):
        self._close(stream)

        self.save()


    def close_all(self, account):
        streams = [stream
                   for stream in self.box.get_children()
                   if stream.account == account]

        stream_thread = None

        for stream in streams:
            self._close(stream)

            stream_thread = self.close_thread

        self.save()

        return stream_thread


    def drag_move(self):
        streams = self.box.get_children()

        index = streams.index(self.drag_dest)

        self.box.reorder_child(self.drag_source, index)

        self.save()


    def move_left(self, stream):
        streams = self.box.get_children()

        index = streams.index(stream)

        if index > 0:
            streams[index - 1].popup()

            self.box.reorder_child(stream, index - 1)

        self.save()


    def move_right(self, stream):
        streams = self.box.get_children()

        index = streams.index(stream)

        if index < len(streams) - 1:
            streams[index + 1].popup()

            self.box.reorder_child(stream, index + 1)

        self.save()
