# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from time import time


from gtk import gdk as Gdk


from polly import escape, dispatch

from polly.twitter.account import Error, NotFoundError

from polly.gui.twitter.post.tweet import Tweet

from polly.gui.twitter.post.result import Result

from polly.gui.stream import PAGE_SIZE

from polly.gui.stream.posts import Posts



RECENT_BOX_TOOLTIP = u'Switch to recent results'

RECENT_SPINNER_TOOLTIP = u'Switching to recent results\u2026'

POPULAR_BOX_TOOLTIP = u'Switch to popular results'

POPULAR_SPINNER_TOOLTIP = u'Switching to popular results\u2026'



class Results(Posts):
    def _build_buttons(self):
        super(Results, self)._build_buttons()

        self.popular_button = self.button_factory.build(u'popular', self._on_popular_button_clicked)

        Gdk.threads_enter()

        if self.parameter[1]:
            self.popular_button.set_toggled(True)
            self.popular_button.box.set_tooltip_text(RECENT_BOX_TOOLTIP)
            self.popular_button.spinner.set_tooltip_text(RECENT_SPINNER_TOOLTIP)
        else:
            self.popular_button.box.set_tooltip_text(POPULAR_BOX_TOOLTIP)
            self.popular_button.spinner.set_tooltip_text(POPULAR_SPINNER_TOOLTIP)

        self.button_box.pack_end(self.popular_button, True, True, 0)

        Gdk.threads_leave()


    def _isolate_buttons(self):
        self.popular_button.isolate()

        super(Results, self)._isolate_buttons()


    def _type_task(self, previous_thread, toggled):
        if previous_thread is not None:
            previous_thread.join()

        self.posts_lock.acquire()

        self.parameter[1] = not toggled

        Gdk.threads_enter()

        self.manager.save()

        posts = self.post_box.get_children()

        for post in posts:
            self.post_box.remove(post)

        Gdk.threads_leave()

        self.max_id = None

        self.since_id = None

        self.posts_lock.release()

        self.refresh(False)

        Gdk.threads_enter()

        self.clearings -= 1

        if self.clearings == 0:
            self.post_box.show()

        if self.parameter[1]:
            self.popular_button.set_toggled(True)
            self.popular_button.box.set_tooltip_text(RECENT_BOX_TOOLTIP)
            self.popular_button.spinner.set_tooltip_text(RECENT_SPINNER_TOOLTIP)
        else:
            self.popular_button.set_toggled(False)
            self.popular_button.box.set_tooltip_text(POPULAR_BOX_TOOLTIP)
            self.popular_button.spinner.set_tooltip_text(POPULAR_SPINNER_TOOLTIP)

        self.popular_button.set_spinning(False)

        Gdk.threads_leave()

        for post in posts:
            post.isolate()

            Gdk.threads_enter()

            post.destroy()

            Gdk.threads_leave()


    def _update(self):
        self.posts_lock.acquire()

        if self.parameter[1]:
            try:
                dataset = self._get_dataset(None, None, PAGE_SIZE)
            except Error as error:
                destroyed_posts = []

                self.status_bar.popup(error.message)
            else:
                Gdk.threads_enter()

                destroyed_posts = self.post_box.get_children()

                for post in destroyed_posts:
                    self.post_box.remove(post)

                Gdk.threads_leave()

                if dataset:
                    timestamp = time()

                    for data in dataset:
                        post = self._build(data)

                        post.synchronize(timestamp)

                        self.post_box.pack_start(post, False, False, 0)
        else:
            _, destroyed_posts = self._update_new(None)

        self.posts_lock.release()

        for post in destroyed_posts:
            post.isolate()

            Gdk.threads_enter()

            post.destroy()

            Gdk.threads_leave()


    def _update_more(self, previous_thread):
        if previous_thread is not None:
            previous_thread.join()

        self.posts_lock.acquire()

        if not self.parameter[1]:
            self._update_old()

        self.posts_lock.release()

        Gdk.threads_enter()

        self.more_button.set_spinning(False)

        Gdk.threads_leave()


    def _get_dataset(self, since_id, max_id, count):
        result_type = u'popular' if self.parameter[1] else u'recent'

        try:
            dataset = self.account.manager.backend.get_result_dataset(self.parameter[0], since_id, max_id, count, result_type)
        except NotFoundError:
            dataset = []

        return dataset


    def _build(self, data):
        return Result(data[0], data[1], self, self.color_scheme, self.font_scheme, self.button_factory, self.stamp_factory, self.status_bar)


    def _on_popular_button_clicked(self):
        self.popular_button.set_spinning(True)

        self.post_box.hide()

        self.clearings += 1

        self.type_thread = dispatch(self._type_task, self.type_thread, self.popular_button.toggled)


    def __init__(self, key, account, parameter, color_scheme, font_scheme, button_factory, stamp_factory, avatar_factory, preview_factory, status_bar, main_panel, manager):
        if not isinstance(parameter, list):
            raise TypeError

        if len(parameter) != 2:
            raise ValueError

        if not isinstance(parameter[0], unicode):
            raise TypeError

        if not isinstance(parameter[1], bool):
            raise TypeError

        super(Results, self).__init__(key, account, parameter, True, color_scheme, font_scheme, button_factory, avatar_factory, preview_factory, status_bar, main_panel, None, manager)

        self.stamp_factory = stamp_factory

        self.tweet_backend_cache = account.backend.tweet_backend_cache

        markup = u'Search: {}'.format(escape(self.parameter[0]))

        self.label.set_markup(markup)
        self.label.set_tooltip_markup(markup)

        self.account.backend.block_updates()

        self.delete_post_uid = self.account.backend.delete_tweet_notifier.connect(self._on_delete_post)

        self.account.backend.allow_updates()

        self.type_thread = None


    def isolate(self):
        if self.type_thread is not None:
            self.type_thread.join()

        self.account.backend.block_updates()

        self.account.backend.delete_tweet_notifier.disconnect(self.delete_post_uid)

        self.account.backend.allow_updates()

        super(Results, self).isolate()


    def replace(self, result, data):
        self.posts_lock.acquire()

        Gdk.threads_enter()

        posts = self.post_box.get_children()

        Gdk.threads_leave()

        if result in posts:
            timestamp = time()

            tweet = Tweet(data[0], data[1], data[2], data[3], self, self.user_backend_cache, self.tweet_backend_cache, self.color_scheme, self.font_scheme, self.button_factory, self.avatar_factory, self.preview_factory, self.status_bar, self.main_panel)

            tweet.synchronize(timestamp)

            index = len(posts) - posts.index(result) - 1

            Gdk.threads_enter()

            self.post_box.remove(result)

            tweet.revert_top_scrollings()

            self.post_box.pack_end(tweet, False, False, 0)
            self.post_box.reorder_child(tweet, index)

            Gdk.threads_leave()

            self.delete_thread = dispatch(self._remove, self.delete_thread, result)

        self.posts_lock.release()
