# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import gtk as Gtk

from gtk import gdk as Gdk


from polly import ENCODING, escape, rescape

from polly.external.wraplabel import WrapLabel

from polly.twitter import ID_KEY, DISPLAY_URL_KEY, EXPANDED_URL_KEY

from polly.twitter.lead import user

from polly.gui import DEFAULT_BORDER, show_url



INFO_TOOLTIP = u'Show information'


HASHTAG_PREFIX = u'#'

MENTION_PREFIX = u'@'


DELETE_BOX_TOOLTIP = u'Delete'

DELETE_SPINNER_TOOLTIP = u'Deleting\u2026'


COMPOSE_MESSAGE_TOOLTIP = u'Compose message'

CANCEL_MESSAGE_TOOLTIP = u'Cancel message'



class Link(Gtk.Alignment):
    def _leave(self):
        self.inside = False

        self.get_window().set_cursor(self.cursor)

        self.label.modify_fg(Gtk.STATE_NORMAL, self.color)


    def _get_color_active(self):
        color_active = self.label.style_get_property(u'link-color')

        if color_active is None:
            color_active = Gdk.color_parse(u'blue')

        return color_active


    def _on_expose_event_after(self, widget, event):
        if self.box.is_focus():
            allocation = self.box.get_allocation()

            context = self.box.get_window().cairo_create()

            context.set_dash([1.0, 1.0])

            context.set_source_color(self.parent_style.fg[Gtk.STATE_NORMAL])
            context.rectangle(0, 0, allocation.width, allocation.height)
            context.stroke()

        return False


    def _on_expose_event(self, widget, event):
        while True:
            parent = widget.get_parent()

            if parent.get_has_window():
                break

            widget = parent

        self.parent_style = parent.get_style()

        context = self.box.get_window().cairo_create()

        context.set_source_color(self.parent_style.bg[Gtk.STATE_NORMAL])

        context.paint()

        return False


    def _on_enter_notify_event(self, widget, event):
        self.inside = True

        self.get_window().set_cursor(self.cursor_hover)

        if self.pressed:
            self.label.modify_fg(Gtk.STATE_NORMAL, self._get_color_active())

        return False


    def _on_leave_notify_event(self, widget, event):
        self._leave()

        return False


    def _on_button_press_event(self, widget, event):
        if event.button == 1:
            self.pressed = True

            self.box.grab_focus()

            self.label.modify_fg(Gtk.STATE_NORMAL, self._get_color_active())

        return False


    def _on_button_release_event(self, widget, event):
        if event.button == 1:
            if self.pressed:
                self.pressed = False

                if self.inside:
                    self.label.modify_fg(Gtk.STATE_NORMAL, self.color)

                    self.callback()

        return False


    def _on_focus_in_event(self, widget, event):
        self.box.queue_draw()

        return False


    def _on_focus_out_event(self, widget, event):
        self.box.queue_draw()

        return False


    def _on_key_press_event(self, widget, event):
        if (event.keyval == Gtk.keysyms.KP_Enter or
            event.keyval == Gtk.keysyms.Return or
            event.keyval == Gtk.keysyms.space):
            self.callback()

            return False

        toplevel = self.get_toplevel()

        return toplevel.press(event)


    def _on_unrealize(self, widget):
        if self.inside:
            self._leave()

        self.pressed = False


    def __init__(self, template, callback):
        self.template = template

        self.callback = callback

        self.inside = False

        self.pressed = False

        self.color_active = Gdk.color_parse(u'red')

        self.cursor = Gdk.Cursor(Gdk.ARROW)

        self.cursor_hover = Gdk.Cursor(Gdk.HAND2)

        self.label = Gtk.Label(None)
        self.label.show()

        self.box = Gtk.EventBox()
        self.box.set_app_paintable(True)
        self.box.set_can_focus(True)
        self.box.connect_after(u'expose-event', self._on_expose_event_after)
        self.box.connect(u'expose-event', self._on_expose_event)
        self.box.connect(u'enter-notify-event', self._on_enter_notify_event)
        self.box.connect(u'leave-notify-event', self._on_leave_notify_event)
        self.box.connect_after(u'button-press-event', self._on_button_press_event)
        self.box.connect(u'button-release-event', self._on_button_release_event)
        self.box.connect(u'focus-in-event', self._on_focus_in_event)
        self.box.connect(u'focus-out-event', self._on_focus_out_event)
        self.box.connect(u'key-press-event', self._on_key_press_event)
        self.box.connect(u'unrealize', self._on_unrealize)
        self.box.add(self.label)
        self.box.show()

        super(Link, self).__init__(0.0, 0.0, 0.0, 0.0)
        self.add(self.box)
        self.show()


    def isolate(self):
        self.callback = None


    def set_markup(self, markup):
        markup = self.template.format(markup)

        self.label.set_markup(markup)


    def set_color(self, color):
        self.color = color

        if not self.inside or not self.pressed:
            self.label.modify_fg(Gtk.STATE_NORMAL, self.color)



class Item(Gtk.VBox):
#    def __del__(self):
#        print u'DEBUG: ITEM DELETED'
    def _build_menu(self):
        mention_menuitem = Gtk.MenuItem(u'Mention this user', False)
        mention_menuitem.connect(u'activate', self._on_mention_menuitem_activate)
        mention_menuitem.show()

        self.menu = Gtk.Menu()
        self.menu.connect(u'selection-done', self._on_menu_selection_done)
        self.menu.append(mention_menuitem)


    def _show_action_buttons(self):
        self.action_box.show()


    def _hide_action_buttons(self):
        if self.action_box.get_focus_child() is not None:
            ancestor = self.action_box.get_ancestor(Gtk.Window)

            if ancestor is not None:
                ancestor.child_focus(Gtk.DIR_TAB_BACKWARD)

        self.action_box.hide()


    def _append_link(self, text, replacements, link):
        indices = self._get_indices(text, link)

        url = link[u'url']

        if not isinstance(url, unicode):
            raise TypeError

        try:
            display_url = link[DISPLAY_URL_KEY]

            if not isinstance(display_url, unicode):
                raise TypeError
        except (KeyError, TypeError):
            display_url = url

        try:
            large_url = link[EXPANDED_URL_KEY]

            if not isinstance(large_url, unicode):
                raise TypeError
        except (KeyError, TypeError):
            large_url = url

        expanded_url = rescape(escape(large_url))

        replacements.append((indices, u'<a href="{}" title="{}">{}</a>'.format(url, expanded_url, display_url)))

        return large_url


    def _append_hashtags(self, text, replacements, rtl, entities):
        try:
            for hashtag in entities[u'hashtags']:
                try:
                    indices = self._get_indices(text, hashtag)

                    tag = hashtag[u'text']

                    if not isinstance(tag, unicode):
                        raise TypeError

                    if rtl:
                        string = u'<a href="{0}{1}">{1}</a>{0}'
                    else:
                        string = u'{0}<a href="{0}{1}">{1}</a>'

                    replacements.append((indices, string.format(HASHTAG_PREFIX, tag)))
                except (TypeError, KeyError, ValueError):
                    pass
        except (KeyError, TypeError):
            pass


    def _append_mention(self, text, replacements, rtl, mention):
        indices = self._get_indices(text, mention)

        user_id = mention[ID_KEY]

        # NOTE: The presence of long will not be necessary
        # after the planned port to Python 3 is completed.

        if not isinstance(user_id, (int, long)):
            raise TypeError

        nick = mention[user.NICK_KEY]

        if not isinstance(nick, unicode):
            raise TypeError

        if rtl:
            string = u'<a href="{0}{1}">{2}</a>{0}'
        else:
            string = u'{0}<a href="{0}{1}">{2}</a>'

        replacements.append((indices, string.format(MENTION_PREFIX, user_id, nick)))

        return indices[0], user_id, nick


    def _join_replacements(self, text, replacements):
        replacements.sort()

        end = 0

        slices = []

        for replacement in replacements:
            start = replacement[0][0]

            if start >= end:
                slices.append(text[end:start])

                slices.append(replacement[1])

                end = replacement[0][1]

        slices.append(text[end:])

        return u''.join(slices)


    # NOTE: The presence of long will not be necessary
    # after the planned port to Python 3 is completed.

    def _get_indices(self, text, entity):
        indices = entity[u'indices']

        if not isinstance(indices, list):
            raise TypeError

        if len(indices) != 2:
            raise ValueError

        start = indices[0]

        if not isinstance(start, (int, long)):
            raise TypeError

        if start < 0 or start >= len(text):
            raise ValueError

        end = indices[1]

        if not isinstance(end, (int, long)):
            raise TypeError

        if end <= start or end > len(text):
            raise ValueError

        return indices


    def _set_selected(self, selected):
        if self.selected != selected:
            self.selected = selected

            if self.selected:
                self._set_normal_fg(self.select_fg)

                self._set_normal_bg(self.select_bg)

                self._set_normal_mg(self.select_mg)

                self._set_select_fg(self.normal_fg)

                self._set_select_bg(self.normal_bg)
            else:
                self._set_normal_fg(self.normal_fg)

                self._set_normal_bg(self.normal_bg)

                self._set_normal_mg(self.normal_mg)

                self._set_select_fg(self.select_fg)

                self._set_select_bg(self.select_bg)

            self.queue_draw()


    def _set_normal_fg(self, color):
        self.header_link.set_color(color)

        self.wraplabel.modify_fg(Gtk.STATE_NORMAL, color)

        self.outer_box.modify_fg(Gtk.STATE_NORMAL, color)

        for spinner in self.spinners:
            if spinner is not None:
                spinner.modify_fg(Gtk.STATE_NORMAL, color)


    def _set_normal_bg(self, color):
        self.wraplabel.modify_base(Gtk.STATE_PRELIGHT, color)

        self.outer_box.modify_bg(Gtk.STATE_NORMAL, color)


    def _set_normal_mg(self, color):
        self.footer_link.set_color(color)


    def _set_select_fg(self, color):
        self.wraplabel.modify_text(Gtk.STATE_ACTIVE, color)

        self.wraplabel.modify_text(Gtk.STATE_SELECTED, color)


    def _set_select_bg(self, color):
        self.wraplabel.modify_base(Gtk.STATE_ACTIVE, color)

        self.wraplabel.modify_base(Gtk.STATE_SELECTED, color)


    def _set_font(self, unit, void):
        self.header_link.box.set_size_request(-1, unit)

        self.vertical_box.child_set_property(self.inner_box, u'padding', void)

        self.footer_link.box.set_size_request(-1, unit)


    def _update_user_nick(self):
        self.user_avatar.box.set_tooltip_markup(self.user_nick)


    def _update_user_name(self):
        self.user_avatar.box.set_tooltip_markup(self.user_name)


    def _on_child_button_press_event(self, widget, event):
        self.child_pressed = True

        return False


    def _on_child_focus_in_event(self, widget, event):
        if self.child_pressed:
            self.child_pressed = False
        else:
            toplevel = self.get_toplevel()

            if toplevel.focused:
                self.popup()

                self.enter()

        return False


    def _on_avatar_clicked(self):
        self.stream.manager.new_tweets_home_of_user(self.stream.account, [self.user_backend.id, True], self.stream)


    def _on_info_button_clicked(self):
        self.info_button.set_toggled(not self.info_button.toggled)

        if self.info_button.toggled:
            self.user_avatar.box.set_tooltip_markup(self.user_name)

            self.info_button.box.set_tooltip_text(u'Hide information')
        else:
            self.user_avatar.box.set_tooltip_markup(self.user_nick)

            self.info_button.box.set_tooltip_text(INFO_TOOLTIP)


    def _on_label_key_press_event(self, widget, event):
        if (event.keyval == Gtk.keysyms.KP_Left or event.keyval == Gtk.keysyms.Left or
            event.keyval == Gtk.keysyms.KP_Right or event.keyval == Gtk.keysyms.Right):
            self.popup()

            return False

        toplevel = self.get_toplevel()

        return toplevel.press(event)


    def _on_label_populate_popup(self, label, menu):
        menuitems = menu.get_children()

        if len(menuitems) > 4:
            menuitems[0].destroy()
            menuitems[2].destroy()
            menuitems[3].destroy()
            menuitems[4].destroy()


    def _on_label_activate_link(self, label, uri):
        uri = uri.decode(ENCODING)

        if uri.startswith(HASHTAG_PREFIX):
            self.stream.manager.new_results(self.stream.account, [uri, False], self.stream)
        elif uri.startswith(MENTION_PREFIX):
            self.stream.manager.new_tweets_home_of_user(self.stream.account, [long(uri[1:]), True], self.stream)
        else:
            if not uri.startswith(u'http://') and not uri.startswith(u'https://'):
                uri = u'http://' + uri

            show_url(uri)

        return True


    def _on_enter_notify_event(self, widget, event):
        self.enter()

        return False


    def _on_leave_notify_event(self, widget, event):
        allocation = self.outer_box.get_allocation()

        if (event.x < 0 or event.x >= allocation.width or
            event.y < 0 or event.y >= allocation.height):
            self.leave()

        return False


    def _on_button_press_event(self, widget, event):
        if event.button == 1:
            if self.get_focus_child() is None:
                self.user_avatar.box.grab_focus()
        elif event.button == 3:
            self._build_menu()

            self.menu.popup(None, None, None, 3, event.time, None)

        return False


    def _on_key_press_event(self, widget, event):
        if event.keyval == Gtk.keysyms.Tab:
            self.popup()
        elif (event.keyval == Gtk.keysyms.Menu or
           (event.keyval == Gtk.keysyms.F10 and event.state & Gdk.SHIFT_MASK)):
            time = Gtk.get_current_event_time()

            self._build_menu()

            self.menu.popup(None, None, None, 0, time, None)

        return False


    def _on_unrealize(self, widget):
        if self.inside:
            self.leave()


    def _on_mention_menuitem_activate(self, menuitem):
        self.main_panel.insert_mention(self.user_nick, self.stream.account)


    def _on_menu_selection_done(self, menushell):
        self.menu.destroy()
        self.menu = None


    def _on_user_nick_changed(self):
        Gdk.threads_enter()

        self.user_nick = self.user_backend.data[user.NICK_KEY]

        if not self.info_button.toggled:
            self._update_user_nick()

        Gdk.threads_leave()


    def _on_user_name_changed(self):
        user_name = self.user_backend.data[user.NAME_KEY]

        if isinstance(user_name, unicode):
            user_name = user_name.strip()
        else:
            user_name = u''

        Gdk.threads_enter()

        if user_name:
            self.user_name = escape(user_name)
        else:
            self.user_name = self.user_nick

        if self.info_button.toggled:
            self._update_user_name()

        Gdk.threads_leave()


    def _on_color_normal_fg_changed(self):
        Gdk.threads_enter()

        self.normal_fg = self.color_scheme.normal_fg

        if self.selected:
            self._set_select_fg(self.normal_fg)
        else:
            self._set_normal_fg(self.normal_fg)

            self.queue_draw()

        Gdk.threads_leave()


    def _on_color_normal_bg_changed(self):
        Gdk.threads_enter()

        self.normal_bg = self.color_scheme.normal_bg

        if self.selected:
            self._set_select_bg(self.normal_bg)
        else:
            self._set_normal_bg(self.normal_bg)

            self.queue_draw()

        Gdk.threads_leave()


    def _on_color_normal_mg_changed(self):
        Gdk.threads_enter()

        self.normal_mg = self.color_scheme.normal_mg

        if not self.selected:
            self._set_normal_mg(self.normal_mg)

        Gdk.threads_leave()


    def _on_color_select_fg_changed(self):
        Gdk.threads_enter()

        self.select_fg = self.color_scheme.select_fg

        if self.selected:
            self._set_normal_fg(self.select_fg)

            self.queue_draw()
        else:
            self._set_select_fg(self.select_fg)

        Gdk.threads_leave()


    def _on_color_select_bg_changed(self):
        Gdk.threads_enter()

        self.select_bg = self.color_scheme.select_bg

        if self.selected:
            self._set_normal_bg(self.select_bg)

            self.queue_draw()
        else:
            self._set_select_bg(self.select_bg)

        Gdk.threads_leave()


    def _on_color_select_mg_changed(self):
        Gdk.threads_enter()

        self.select_mg = self.color_scheme.select_mg

        if self.selected:
            self._set_normal_mg(self.select_mg)

        Gdk.threads_leave()


    def _on_font_changed(self):
        unit = self.font_scheme.unit

        void = self.font_scheme.void

        Gdk.threads_enter()

        self._set_font(unit, void)

        Gdk.threads_leave()


    def _connect(self):
        self.user_backend.block_updates()

        self.color_scheme.block_updates()

        self.font_scheme.block_updates()

        self._on_user_nick_changed()

        self._on_user_name_changed()

        self._on_color_normal_fg_changed()

        self._on_color_normal_bg_changed()

        self._on_color_normal_mg_changed()

        self._on_color_select_fg_changed()

        self._on_color_select_bg_changed()

        self._on_color_select_mg_changed()

        self._on_font_changed()

        self.user_nick_uid = self.user_backend.nick_notifier.connect(self._on_user_nick_changed)

        self.user_name_uid = self.user_backend.name_notifier.connect(self._on_user_name_changed)

        self.color_normal_fg_uid = self.color_scheme.normal_fg_notifier.connect(self._on_color_normal_fg_changed)

        self.color_normal_bg_uid = self.color_scheme.normal_bg_notifier.connect(self._on_color_normal_bg_changed)

        self.color_normal_mg_uid = self.color_scheme.normal_mg_notifier.connect(self._on_color_normal_mg_changed)

        self.color_select_fg_uid = self.color_scheme.select_fg_notifier.connect(self._on_color_select_fg_changed)

        self.color_select_bg_uid = self.color_scheme.select_bg_notifier.connect(self._on_color_select_bg_changed)

        self.color_select_mg_uid = self.color_scheme.select_mg_notifier.connect(self._on_color_select_mg_changed)

        self.font_uid = self.font_scheme.notifier.connect(self._on_font_changed)

        self.user_backend.allow_updates()

        self.color_scheme.allow_updates()

        self.font_scheme.allow_updates()


    def _disconnect(self):
        self.font_scheme.block_updates()

        self.color_scheme.block_updates()

        self.user_backend.block_updates()

        self.font_scheme.notifier.disconnect(self.font_uid)

        self.color_scheme.select_mg_notifier.disconnect(self.color_select_mg_uid)

        self.color_scheme.select_bg_notifier.disconnect(self.color_select_bg_uid)

        self.color_scheme.select_fg_notifier.disconnect(self.color_select_fg_uid)

        self.color_scheme.normal_mg_notifier.disconnect(self.color_normal_mg_uid)

        self.color_scheme.normal_bg_notifier.disconnect(self.color_normal_bg_uid)

        self.color_scheme.normal_fg_notifier.disconnect(self.color_normal_fg_uid)

        self.user_backend.name_notifier.disconnect(self.user_name_uid)

        self.user_backend.nick_notifier.disconnect(self.user_nick_uid)

        self.font_scheme.allow_updates()

        self.color_scheme.allow_updates()

        self.user_backend.allow_updates()


    def __init__(self, dimension, spacing, user_data, stream, user_backend_cache, color_scheme, font_scheme, button_factory, avatar_factory, status_bar, main_panel):
        self.stream = stream

        self.color_scheme = color_scheme

        self.font_scheme = font_scheme

        self.status_bar = status_bar

        self.main_panel = main_panel

        self.inside = False

        self.selected = False

        self.user_own = user_data[user.OWN_KEY]

        self.user_backend = user_backend_cache.acquire(user_data)

        self.user_avatar = avatar_factory.build(self.user_backend, dimension, spacing, self._on_avatar_clicked)

        self.info_button = button_factory.build(u'info', self._on_info_button_clicked)

        self.spinners = []

        Gdk.threads_enter()

        self.spinners.append(self.user_avatar.spinner)

        self.user_avatar.box.connect(u'button-press-event', self._on_child_button_press_event)
        self.user_avatar.box.connect(u'focus-in-event', self._on_child_focus_in_event)

        self.header_link = Link(u'<b>{}</b>', self._on_header_clicked)

        self.header_link.box.connect(u'button-press-event', self._on_child_button_press_event)
        self.header_link.box.connect(u'focus-in-event', self._on_child_focus_in_event)

        self.info_button.box.set_tooltip_text(INFO_TOOLTIP)

        self.indicator_box = Gtk.HBox(True, DEFAULT_BORDER)
        self.indicator_box.pack_end(self.info_button, True, True, 0)

        self.header_box = Gtk.HBox(False, DEFAULT_BORDER)
        self.header_box.pack_start(self.header_link, True, True, 0)
        self.header_box.pack_start(self.indicator_box, False, False, 0)
        self.header_box.show()

        self.wraplabel = WrapLabel(None)
        self.wraplabel.set_selectable(True)
        self.wraplabel.connect(u'key-press-event', self._on_label_key_press_event)
        self.wraplabel.connect(u'populate-popup', self._on_label_populate_popup)
        self.wraplabel.show()

        self.inner_box = Gtk.VBox(False, 0)
        self.inner_box.pack_start(self.wraplabel, True, True, 0)
        self.inner_box.show()

        self.footer_link = Link(u'<small>{}</small>', self._on_footer_clicked)

        self.action_box = Gtk.HBox(True, DEFAULT_BORDER)

        self.footer_box = Gtk.HBox(False, DEFAULT_BORDER)
        self.footer_box.pack_end(self.action_box, False, False, 0)
        self.footer_box.show()

        self.vertical_box = Gtk.VBox(False, 0)
        self.vertical_box.pack_start(self.header_box, False, False, 0)
        self.vertical_box.pack_start(self.inner_box, True, True, 0)
        self.vertical_box.pack_start(self.footer_box, False, False, 0)
        self.vertical_box.show()

        horizontal_box = Gtk.HBox(False, DEFAULT_BORDER)
        horizontal_box.set_border_width(DEFAULT_BORDER)
        horizontal_box.pack_start(self.user_avatar, False, False, 0)
        horizontal_box.pack_start(self.vertical_box, True, True, 0)
        horizontal_box.show()

        self.outer_box = Gtk.EventBox()
        self.outer_box.connect(u'enter-notify-event', self._on_enter_notify_event)
        self.outer_box.connect(u'leave-notify-event', self._on_leave_notify_event)
        self.outer_box.connect(u'button-press-event', self._on_button_press_event)
        self.outer_box.connect(u'key-press-event', self._on_key_press_event)
        self.outer_box.connect(u'unrealize', self._on_unrealize)
        self.outer_box.add(horizontal_box)
        self.outer_box.show()

        separator = Gtk.HSeparator()
        separator.show()

        super(Item, self).__init__(False, 0)
        self.pack_start(self.outer_box, True, True, 0)
        self.pack_start(separator, False, False, 0)
        self.show()

        self.menu = None

        Gdk.threads_leave()

        self.child_pressed = False


    def isolate(self):
        Gdk.threads_enter()

        if self.menu is not None:
            self.menu.destroy()

        Gdk.threads_leave()

        self.footer_link.isolate()

        self.header_link.isolate()

        self.info_button.isolate()

        self.user_avatar.isolate()

        self.user_backend.release()


    def enter(self):
        if not self.inside:
            if self.stream.current_item is not None:
                self.stream.current_item.leave()

            self.stream.current_item = self

            self.inside = True

            self.indicator_box.show()

            self._show_action_buttons()


    def leave(self):
        self._hide_action_buttons()

        if self.indicator_box.get_focus_child() is not None:
            ancestor = self.indicator_box.get_ancestor(Gtk.Window)

            if ancestor is not None:
                ancestor.child_focus(Gtk.DIR_TAB_BACKWARD)

        self.indicator_box.hide()

        self.inside = False

        if self.stream.current_item == self:
            self.stream.current_item = None


    def popup(self):
        self.stream.popup()

        allocation = self.outer_box.get_allocation()

        adjustment = self.stream.adjustment

        top = float(allocation.y)

        value = adjustment.get_value()

        if top < value:
            adjustment.set_value(top)
        else:
            bottom = top + float(allocation.height)

            page_size = adjustment.get_page_size()

            if bottom > value + page_size:
                adjustment.set_value(bottom - page_size)



import lead, post, topic, shelf, database, account
