# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import json, os, shutil

from random import randint

from threading import Lock, Condition

from time import time


import keyring

import gtk as Gtk

from gtk import gdk as Gdk


from polly import ENCODING, dispatch

from polly.twitter.lead import user

from polly.twitter.account import SUBTOKEN_KEY, SUBTOKEN_SECRET_KEY, Error, ConnectionError, ManagerBackend

from polly.twitter.realtime import Producer

from polly.gui import DEFAULT_BORDER, NICK_DIMENSION, run_generic_error_dialog

from polly.gui.twitter.shelf import Shelf

from polly.gui.twitter.database import Database



class Account(object):
    def __del__(self):
        print u'DEBUG: ACCOUNT DELETED'
    def _on_menuitem_activate(self, menuitem):
        self.manager.set_account(self)


    def _on_user_nick_changed(self):
        self.user_nick_lock.acquire()

        self.user_nick = self.backend.user_backend.data[user.NICK_KEY]

        self.user_nick_lock.release()

        Gdk.threads_enter()

        self.label.set_markup(u'<b>{}</b>'.format(self.user_nick))

        self.menuitem.set_label(self.user_nick)

        Gdk.threads_leave()


    def _on_user_changed(self):
        try:
            self.config_path_lock.acquire()

            try:
                if not os.path.isdir(self.user_config_path):
                    os.makedirs(self.user_config_path)
            finally:
                self.config_path_lock.release()

            if os.path.isdir(self.user_path):
                shutil.rmtree(self.user_path)

            with open(self.user_path, u'w') as file:
                json.dump(self.backend.user_backend.data, file)
        except (OSError, IOError):
            pass


    def __init__(self, id, data, offline, user_config_path, shelf_config_path, database_config_path, config_path_lock, proxy_controller, identifier_factory, status_bar, manager):
        self.id = id

        self.user_config_path = user_config_path

        self.config_path_lock = config_path_lock

        self.identifier_factory = identifier_factory

        self.manager = manager

        self.user_path = os.path.join(self.user_config_path, bytes(self.id).decode(ENCODING))

        if offline:
            with open(self.user_path) as file:
                user_data = json.load(file)
        else:
            user_data = None

        self.backend = self.manager.backend.build(data, user_data)

        self.user_nick_lock = Lock()

        Gdk.threads_enter()

        self.label = Gtk.Label(None)

        self.menuitem = Gtk.MenuItem(None, False)
        self.menuitem.connect(u'activate', self._on_menuitem_activate)
        self.menuitem.show()

        Gdk.threads_leave()

        self.shelf = Shelf(self.id, offline, shelf_config_path, self.config_path_lock)

        self.database = Database(self.id, offline, database_config_path, self.config_path_lock)

        self.producer = Producer(self, proxy_controller, status_bar)

        self.identifier = self.build_identifier()

        self.backend.user_backend.block_updates()

        self._on_user_nick_changed()

        self._on_user_changed()

        self.user_nick_uid = self.backend.user_backend.nick_notifier.connect(self._on_user_nick_changed)

        self.user_uid = self.backend.user_backend.notifier.connect(self._on_user_changed)

        self.backend.user_backend.allow_updates()


    def isolate(self, completely):
        self.backend.user_backend.block_updates()

        self.backend.user_backend.notifier.disconnect(self.user_uid)

        self.backend.user_backend.nick_notifier.disconnect(self.user_nick_uid)

        self.backend.user_backend.allow_updates()

        if completely:
            try:
                os.remove(self.user_path)
            except OSError:
                pass

        self.identifier.isolate()

        self.producer.isolate()
        self.producer = None

        self.database.isolate(completely)

        self.shelf.isolate(completely)

        self.backend.isolate()


    def get_user_nick(self):
        self.user_nick_lock.acquire()

        user_nick = self.user_nick

        self.user_nick_lock.release()

        return user_nick


    def build_identifier(self):
        return self.identifier_factory.build(self.backend.user_backend)



class Manager(object):
    def _menu_position_func(self, menu, user_data):
        x, y = self.togglebutton.get_window().get_origin()

        allocation = self.togglebutton.get_allocation()

        x += allocation.x

        y += allocation.y + allocation.height

        return x, y, True


    def _add(self, id, data, offline):
        account = Account(id, data, offline, self.user_config_path, self.shelf_config_path, self.database_config_path, self.config_path_lock, self.proxy_controller, self.identifier_factory, self.status_bar, self)

        self.accounts.append(account)

        Gdk.threads_enter()

        if len(self.accounts) == 1:
            if self.sensitive:
                self.togglebutton.set_sensitive(True)

        account.identifier.hide()

        account.identifier.spinner.modify_fg(Gtk.STATE_NORMAL, self.normal_fg)

        account.label.modify_fg(Gtk.STATE_NORMAL, self.normal_fg)
        account.label.modify_fg(Gtk.STATE_PRELIGHT, self.prelight_fg)
        account.label.modify_fg(Gtk.STATE_ACTIVE, self.active_fg)

        self.identifier_box.pack_start(account.identifier, False, False, 0)

        self.label_box.pack_start(account.label, False, False, 0)

        self.menu.append(account.menuitem)

        Gdk.threads_leave()

        return account


    def _remove(self, previous_thread, stream_thread, account):
        if previous_thread is not None:
            previous_thread.join()

        if stream_thread is not None:
            stream_thread.join()

        account.producer.stop()

        account.isolate(True)

        Gdk.threads_enter()

        account.identifier.destroy()

        account.menuitem.destroy()

        account.label.destroy()

        Gdk.threads_leave()


    def _save(self):
        dataset = []

        for account in self.accounts:
            dataset.append([bytes(account.id).decode(ENCODING), account.backend.data[SUBTOKEN_KEY]])

        try:
            self.config_path_lock.acquire()

            try:
                if not os.path.isdir(self.config_path):
                    os.makedirs(self.config_path)
            finally:
                self.config_path_lock.release()

            if os.path.isdir(self.path):
                shutil.rmtree(self.path)

            with open(self.path, u'w') as file:
                json.dump(dataset, file)
        except (OSError, IOError):
            pass


    def _on_togglebutton_size_allocate(self, widget, allocation):
        self.menu.set_size_request(allocation.width, -1)


    def _on_togglebutton_enter_notify_event(self, widget, event):
        self.togglebutton.set_relief(Gtk.RELIEF_NORMAL)

        return False


    def _on_togglebutton_leave_notify_event(self, widget, event):
        self._on_togglebutton_unrealize(widget)

        return False


    def _on_togglebutton_unrealize(self, widget):
        self.togglebutton.set_relief(Gtk.RELIEF_NONE)


    def _on_togglebutton_toggled(self, togglebutton):
        if self.togglebutton.get_active():
            if self.account is None or len(self.accounts) > 1:
                time = Gtk.get_current_event_time()

                self.menu.popup(None, None, self._menu_position_func, 1, time, None)
            else:
                self.togglebutton.set_active(False)


    def _on_menu_selection_done(self, menushell):
        self.togglebutton.set_active(False)


    def _on_id_changed(self):
        id = self.id_setting.get_value()

        account = self.get_account(id)

        self.set_account(account)


    def _on_font_changed(self):
        unit = self.font_scheme.unit

        char = NICK_DIMENSION * self.font_scheme.char

        Gdk.threads_enter()

        self.identifier_box.set_size_request(unit, unit)

        self.label_box.set_size_request(char, -1)

        Gdk.threads_leave()


    def _update_loop(self):
        while True:
            timestamp = time()

            if timestamp - self.timestamp > 86400.0:
                try:
                    lengths = self.backend.get_lengths()
                except ConnectionError as error:
                    if self.success == -1:
                        self.timeout = min(self.timeout + 0.25, 16.0)

                        if self.timeout > 15.9:
                            self.status_bar.popup(error.message)
                    else:
                        self.success = -1

                        self.timeout = 0.25
                except Error as error:
                    if self.success == 0:
                        self.timeout = min(2.0 * self.timeout, float(randint(240, 300)))

                        if self.timeout > 239.0:
                            self.status_bar.popup(error.message)
                    else:
                        self.success = 0

                        self.timeout = float(randint(20, 40))
                else:
                    self.success = 1

                    self.timeout = 3600.0

                    self.timestamp = timestamp

                    self.lengths_lock.acquire()

                    self.lengths = lengths

                    self.lengths_lock.release()

            self.update_condition.acquire()

            if not self.update_finished:
                self.update_condition.wait(self.timeout)

            update_finished = self.update_finished

            self.update_condition.release()

            if update_finished:
                break


    def __init__(self, name, config_path, config_path_lock, setting_factory, proxy_controller, font_scheme, identifier_factory, status_bar):
        self.name = name

        self.config_path = config_path

        self.config_path_lock = config_path_lock

        self.proxy_controller = proxy_controller

        self.font_scheme = font_scheme

        self.identifier_factory = identifier_factory

        self.status_bar = status_bar

        self.sensitive = True

        self.update_finished = False

        self.success = 1

        self.timeout = 0.0

        self.timestamp = 0.0

        self.lengths = [20, 21]

        self.path = os.path.join(self.config_path, u'accounts')

        self.user_config_path = os.path.join(config_path, u'users')

        self.shelf_config_path = os.path.join(config_path, u'shelves')

        self.database_config_path = os.path.join(config_path, u'databases')

        self.lengths_lock = Lock()

        self.update_condition = Condition()

        self.normal_fg = None

        self.prelight_fg = None

        self.active_fg = None

        self.identifier_box = Gtk.HBox(False, 0)
        self.identifier_box.show()

        self.label_box = Gtk.HBox(False, 0)
        self.label_box.show()

        self.arrow = Gtk.Arrow(Gtk.ARROW_DOWN, Gtk.SHADOW_NONE)
        self.arrow.show()

        button_box = Gtk.HBox(False, DEFAULT_BORDER)
        button_box.pack_start(self.label_box, True, True, 0)
        button_box.pack_start(self.arrow, False, False, 0)
        button_box.show()

        self.togglebutton = Gtk.ToggleButton(None, False)
        self.togglebutton.set_relief(Gtk.RELIEF_NONE)
        self.togglebutton.set_sensitive(False)
        self.togglebutton.connect(u'size-allocate', self._on_togglebutton_size_allocate)
        self.togglebutton.connect(u'enter-notify-event', self._on_togglebutton_enter_notify_event)
        self.togglebutton.connect(u'leave-notify-event', self._on_togglebutton_leave_notify_event)
        self.togglebutton.connect(u'unrealize', self._on_togglebutton_unrealize)
        self.togglebutton.connect(u'toggled', self._on_togglebutton_toggled)
        self.togglebutton.add(button_box)
        self.togglebutton.show()

        self.menu = Gtk.Menu()
        self.menu.connect(u'selection-done', self._on_menu_selection_done)

        self.id_setting = setting_factory.build_integer(u'account', u'id', self._on_id_changed)

        self.backend = ManagerBackend()

        self.account = None

        self.accounts = []

        try:
            with open(self.path) as file:
                dataset = json.load(file)

            for spec in dataset:
                try:
                    key, subtoken = spec

                    id = long(key)

                    if id < 0:
                        raise ValueError

                    subtoken_secret = keyring.get_password(self.name, subtoken)

                    if subtoken_secret is None:
                        raise ValueError

                    data = {SUBTOKEN_KEY: subtoken,
                            SUBTOKEN_SECRET_KEY: subtoken_secret.decode(ENCODING)}

                    self._add(id, data, True)
                except (TypeError, ValueError, KeyError, IOError):
                    pass
        except (IOError, ValueError, TypeError):
            pass

        self._save()

        self._on_id_changed()

        self.font_scheme.block_updates()

        self._on_font_changed()

        self.font_uid = self.font_scheme.notifier.connect(self._on_font_changed)

        self.font_scheme.allow_updates()

        self.remove_thread = None


    def isolate(self):
        if self.remove_thread is not None:
            self.remove_thread.join()

        self.font_scheme.block_updates()

        self.font_scheme.notifier.disconnect(self.font_uid)

        self.font_scheme.allow_updates()

        for account in self.accounts:
            account.isolate(False)

        del self.accounts[:]

        self.account = None

        self.id_setting.isolate()

        self.menu.destroy()


    def start_all(self):
        self.update_thread = dispatch(self._update_loop)

        for account in self.accounts:
            account.producer.start()


    def stop_all(self):
        for account in self.accounts:
            account.producer.stop()

        self.update_condition.acquire()

        self.update_finished = True

        self.update_condition.notify()

        self.update_condition.release()

        self.update_thread.join()


    def get_lengths(self):
        self.lengths_lock.acquire()

        lengths = self.lengths

        self.lengths_lock.release()

        return lengths


    def get_account(self, id):
        for account in self.accounts:
            if account.id == id:
                return account

        return None


    def get_current_account(self):
        if self.account is None:
            toplevel = self.togglebutton.get_toplevel()

            run_generic_error_dialog(toplevel, u'Please select an account and try again.')

        return self.account


    def set_account(self, account):
        if self.account != account:
            if self.account is not None:
                self.account.identifier.hide()

                self.account.label.hide()

                self.account.menuitem.show()

            self.account = account

            if self.account is None:
                self.id_setting.set_value(-1)
            else:
                self.account.menuitem.hide()

                self.account.label.show()

                self.account.identifier.show()

                self.id_setting.set_value(self.account.id)


    def set_sensitive(self, sensitive):
        if self.sensitive != sensitive:
            self.sensitive = sensitive

            if self.accounts:
                self.togglebutton.set_sensitive(self.sensitive)


    def set_colors(self, fg, style):
        self.normal_fg = fg

        self.prelight_fg = style.fg[Gtk.STATE_PRELIGHT]

        self.active_fg = style.fg[Gtk.STATE_ACTIVE]

        for identifier in self.identifier_box.get_children():
            identifier.spinner.modify_fg(Gtk.STATE_NORMAL, self.normal_fg)

        for label in self.label_box.get_children():
            label.modify_fg(Gtk.STATE_NORMAL, self.normal_fg)
            label.modify_fg(Gtk.STATE_PRELIGHT, self.prelight_fg)
            label.modify_fg(Gtk.STATE_ACTIVE, self.active_fg)

        self.arrow.modify_fg(Gtk.STATE_NORMAL, self.normal_fg)
        self.arrow.modify_fg(Gtk.STATE_PRELIGHT, self.prelight_fg)
        self.arrow.modify_fg(Gtk.STATE_ACTIVE, self.active_fg)

        self.togglebutton.modify_bg(Gtk.STATE_PRELIGHT, style.bg[Gtk.STATE_PRELIGHT])
        self.togglebutton.modify_bg(Gtk.STATE_ACTIVE, style.bg[Gtk.STATE_ACTIVE])


    def add(self, data):
        ids = []

        for account in self.accounts:
            ids.append(account.id)

        if ids:
            ids.sort()

            if ids[0] > 0:
                id = 0
            else:
                current = ids[0]

                for id in ids[1:]:
                    if id - current > 1:
                        break

                    current = id

                id = current + 1
        else:
            id = 0

        account = self._add(id, data, False)

        keyring.set_password(self.name.encode(ENCODING), data[SUBTOKEN_KEY].encode(ENCODING), data[SUBTOKEN_SECRET_KEY].encode(ENCODING))

        self._save()

        Gdk.threads_enter()

        self.set_account(account)

        Gdk.threads_leave()

        account.producer.start()

        return account


    def remove(self, account, stream_thread):
        index = self.accounts.index(account)

        if len(self.accounts) == 1:
            self.set_account(None)

            if self.sensitive:
                self.togglebutton.set_sensitive(False)
        else:
            if index == len(self.accounts) - 1:
                self.set_account(self.accounts[index - 1])
            else:
                self.set_account(self.accounts[index + 1])

        account.menuitem.hide()

        self.accounts.remove(account)

        self._save()

        self.remove_thread = dispatch(self._remove, self.remove_thread, stream_thread, account)


    def move_up(self):
        if self.accounts:
            if self.account is not None:
                index = self.accounts.index(self.account)

                if index > 0:
                    self.set_account(self.accounts[index - 1])


    def move_down(self):
        if self.accounts:
            if self.account is None:
                self.set_account(self.accounts[0])
            else:
                index = self.accounts.index(self.account)

                if index < len(self.accounts) - 1:
                    self.set_account(self.accounts[index + 1])
