# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import os, shutil, sqlite3

from threading import Lock


from polly import ENCODING

from polly.twitter import ID_KEY, REQUEST_TIMESTAMP_KEY, CREATION_TIMESTAMP_KEY, ENTITIES_KEY, MENTIONS_KEY

from polly.twitter.lead import user



CREATE_SQL = u'CREATE TABLE users (id INTEGER PRIMARY KEY, timestamp REAL, nick TEXT)'



class Database(object):
    def _insert(self, cursor, id, timestamp, nick):
        cursor.execute(u'SELECT timestamp, nick FROM users WHERE id=?', (id,))

        row = cursor.fetchone()

        if row is None:
            cursor.execute(u'INSERT INTO users (id, timestamp, nick) VALUES (?, ?, ?)', (id, timestamp, nick))
        else:
            if timestamp > row[0] and nick != row[1]:
                cursor.execute(u'UPDATE users SET timestamp=?, nick=? WHERE id=?', (timestamp, nick, id))


    def __init__(self, id, offline, config_path, config_path_lock):
        self.path = os.path.join(config_path, bytes(id).decode(ENCODING))

        self.lock = Lock()

        success = False

        if offline:
            try:
                connection = sqlite3.connect(self.path)
            except sqlite3.Error:
                pass
            else:
                try:
                    cursor = connection.cursor()
                except sqlite3.Error:
                    pass
                else:
                    try:
                        cursor.execute(u'SELECT * FROM sqlite_master WHERE sql=?', (CREATE_SQL,))

                        if cursor.fetchone() is not None:
                            success = True
                    except sqlite3.Error:
                        pass
                    finally:
                        try:
                            cursor.close()
                        except sqlite3.Error:
                            pass
                finally:
                    try:
                        connection.close()
                    except sqlite3.Error:
                        pass

        if not success:
            config_path_lock.acquire()

            try:
                if not os.path.isdir(config_path):
                    os.makedirs(config_path)
            except OSError:
                pass
            finally:
                config_path_lock.release()

            if os.path.exists(self.path):
                try:
                    if os.path.isfile(self.path):
                        os.remove(self.path)
                    elif os.path.isdir(self.path):
                        shutil.rmtree(self.path)
                except OSError:
                    pass

            try:
                connection = sqlite3.connect(self.path)
            except sqlite3.Error:
                pass
            else:
                try:
                    connection.execute(CREATE_SQL)

                    connection.commit()
                except sqlite3.Error:
                    pass
                finally:
                    try:
                        connection.close()
                    except sqlite3.Error:
                        pass


    def isolate(self, completely):
        if completely:
            try:
                os.remove(self.path)
            except OSError:
                pass


    def insert(self, dataset):
        self.lock.acquire()

        try:
            connection = sqlite3.connect(self.path)
        except sqlite3.Error:
            pass
        else:
            try:
                cursor = connection.cursor()

                try:
                    for data in dataset:
                        if data[1] is None:
                            user_data = data[2]
                        else:
                            user_data = data[1]

                        if not user_data[user.OWN_KEY]:
                            id = user_data[ID_KEY]

                            timestamp = user_data[REQUEST_TIMESTAMP_KEY]

                            nick = user_data[user.NICK_KEY]

                            self._insert(cursor, id, timestamp, nick)

                        if data[2][user.OWN_KEY]:
                            try:
                                for mention in data[3][ENTITIES_KEY][MENTIONS_KEY]:
                                    try:
                                        id = mention[ID_KEY]

                                        timestamp = data[3][CREATION_TIMESTAMP_KEY]

                                        nick = mention[user.NICK_KEY]

                                        self._insert(cursor, id, timestamp, nick)
                                    except (TypeError, KeyError):
                                        pass
                            except (TypeError, KeyError):
                                pass
                except sqlite3.Error:
                    pass
                finally:
                    try:
                        cursor.close()
                    except sqlite3.Error:
                        pass

                connection.commit()
            except sqlite3.Error:
                pass
            finally:
                try:
                    connection.close()
                except sqlite3.Error:
                    pass

        self.lock.release()


    def select(self, prefix, nicks):
        self.lock.acquire()

        try:
            connection = sqlite3.connect(self.path)
        except sqlite3.Error:
            pass
        else:
            try:
                cursor = connection.cursor()

                try:
                    nick = u'{}%'.format(prefix.lower().replace(u'_', u'\_'))

                    cursor.execute(u'SELECT nick FROM users WHERE LOWER(nick) LIKE ? ESCAPE "\\"', (nick,))

                    for row in cursor:
                        nicks.append(row[0])
                except sqlite3.Error:
                    pass
                finally:
                    try:
                        cursor.close()
                    except sqlite3.Error:
                        pass
            except sqlite3.Error:
                pass
            finally:
                try:
                    connection.close()
                except sqlite3.Error:
                    pass

        self.lock.release()
