# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import gtk as Gtk

from gtk import gdk as Gdk


from polly import ENCODING, dispatch

from polly.twitter import ID_KEY

from polly.twitter.account import Error

from polly.gui import DEFAULT_BORDER, run_generic_error_dialog

from polly.gui.dialog import Dialog



class GroupDialog(Dialog):
    def _except(self, error):
        Gdk.threads_enter()

        self.stop(error.message)

        self.ok_button.set_sensitive(True)

        self.mode_box.set_sensitive(True)

        self.description_textview.set_sensitive(True)

        self.name_entry.set_sensitive(True)

        Gdk.threads_leave()


    def _get_parameters(self):
        name_bytes = self.name_entry.get_text()

        if not name_bytes.strip():
            run_generic_error_dialog(self, u'The name cannot be blank.')

            return None

        name = name_bytes.decode(ENCODING)

        start = self.description_textbuffer.get_start_iter()

        end = self.description_textbuffer.get_end_iter()

        description_bytes = self.description_textbuffer.get_text(start, end)

        description = description_bytes.decode(ENCODING)

        if len(description) > 100:
            run_generic_error_dialog(self, u'The description cannot exceed 100 characters.')

            return None

        mode = u'public' if self.mode_radiobuttons[0].get_active() else u'private'

        return name, description, mode


    def _on_cancel_button_clicked(self, button):
        self.response(Gtk.RESPONSE_REJECT)


    def _on_ok_button_clicked(self, button):
        parameters = self._get_parameters()

        if parameters is not None:
            self.name_entry.set_sensitive(False)

            self.description_textview.set_sensitive(False)

            self.mode_box.set_sensitive(False)

            self.ok_button.set_sensitive(False)

            self.start()

            self.thread = dispatch(self._task, self.thread, parameters)


    def __init__(self, title, toplevel, account):
        super(GroupDialog, self).__init__(title, toplevel)

        self.account = account

        name_label = Gtk.Label(None)
        name_label.set_alignment(1.0, 0.5)
        name_label.set_markup(u'<b>Name:</b>')
        name_label.show()

        self.name_entry = Gtk.Entry(25)
        self.name_entry.show()

        name_alignment = Gtk.Alignment(0.0, 0.5, 0.0, 0.0)
        name_alignment.add(self.name_entry)
        name_alignment.show()

        description_label = Gtk.Label(None)
        description_label.set_alignment(1.0, 0.0)
        description_label.set_markup(u'<b>Description:</b>')
        description_label.show()

        self.description_textview = Gtk.TextView(None)
        self.description_textview.set_wrap_mode(Gtk.WRAP_WORD_CHAR)
        self.description_textview.show()

        self.description_textbuffer = self.description_textview.get_buffer()

        description_scrolledwindow = Gtk.ScrolledWindow(None, None)
        description_scrolledwindow.set_policy(Gtk.POLICY_AUTOMATIC, Gtk.POLICY_AUTOMATIC)
        description_scrolledwindow.set_shadow_type(Gtk.SHADOW_IN)
        description_scrolledwindow.set_size_request(-1, 10 * DEFAULT_BORDER)
        description_scrolledwindow.add(self.description_textview)
        description_scrolledwindow.show()

        mode_label = Gtk.Label(None)
        mode_label.set_alignment(1.0, 0.5)
        mode_label.set_markup(u'<b>Mode:</b>')
        mode_label.show()

        self.mode_radiobuttons = []

        self.mode_radiobuttons.append(Gtk.RadioButton(None,
                                                      u'Public', False))
        self.mode_radiobuttons[0].show()

        self.mode_radiobuttons.append(Gtk.RadioButton(self.mode_radiobuttons[0],
                                                      u'Private', False))
        self.mode_radiobuttons[1].show()

        self.mode_box = Gtk.HBox(False, DEFAULT_BORDER)
        self.mode_box.pack_start(self.mode_radiobuttons[0], False, False, 0)
        self.mode_box.pack_start(self.mode_radiobuttons[1], False, False, 0)
        self.mode_box.show()

        table = Gtk.Table(3, 2, False)
        table.set_border_width(2 * DEFAULT_BORDER)
        table.set_row_spacings(2 * DEFAULT_BORDER)
        table.set_col_spacings(2 * DEFAULT_BORDER)
        table.attach(name_label, 0, 1, 0, 1, Gtk.EXPAND | Gtk.FILL, Gtk.EXPAND | Gtk.FILL, 0, 0)
        table.attach(name_alignment, 1, 2, 0, 1, Gtk.EXPAND | Gtk.FILL, Gtk.EXPAND | Gtk.FILL, 0, 0)
        table.attach(description_label, 0, 1, 1, 2, Gtk.EXPAND | Gtk.FILL, Gtk.EXPAND | Gtk.FILL, 0, 0)
        table.attach(description_scrolledwindow, 1, 2, 1, 2, Gtk.EXPAND | Gtk.FILL, Gtk.EXPAND | Gtk.FILL, 0, 0)
        table.attach(mode_label, 0, 1, 2, 3, Gtk.EXPAND | Gtk.FILL, Gtk.EXPAND | Gtk.FILL, 0, 0)
        table.attach(self.mode_box, 1, 2, 2, 3, Gtk.EXPAND | Gtk.FILL, Gtk.EXPAND | Gtk.FILL, 0, 0)
        table.show()

        self.content_frame.add(table)

        cancel_button = Gtk.Button(None, Gtk.STOCK_CANCEL, False)
        cancel_button.connect(u'clicked', self._on_cancel_button_clicked)
        cancel_button.show()

        self.ok_button = Gtk.Button(None, Gtk.STOCK_OK, False)
        self.ok_button.connect(u'clicked', self._on_ok_button_clicked)
        self.ok_button.show()

        self.action_box.pack_start(cancel_button, False, False, 0)
        self.action_box.pack_start(self.ok_button, False, False, 0)



class CreateGroupDialog(GroupDialog):
    def _task(self, previous_thread, parameters):
        if previous_thread is not None:
            previous_thread.join()

        name, description, mode = parameters

        try:
            data = self.account.backend.send_group_create(name, description, mode)
        except Error as error:
            self._except(error)
        else:
            Gdk.threads_enter()

            if self.running:
                self.stream_manager.new_tweets_home_of_group(self.account, [data[0][ID_KEY], data[1][ID_KEY], True], None)

                self.response(Gtk.RESPONSE_ACCEPT)

            Gdk.threads_leave()


    def __init__(self, toplevel, account, stream_manager):
        super(CreateGroupDialog, self).__init__(u'Create list', toplevel, account)

        self.stream_manager = stream_manager



class EditGroupDialog(GroupDialog):
    def _task(self, previous_thread, parameters):
        if previous_thread is not None:
            previous_thread.join()

        name, description, mode = parameters

        try:
            data = self.account.backend.send_group_update(self.group_backend.id, name, description, mode)
        except Error as error:
            self._except(error)
        else:
            self.user_backend.set_data(data[0])

            self.group_backend.set_data(data[1])

            Gdk.threads_enter()

            if self.running:
                self.response(Gtk.RESPONSE_ACCEPT)

            Gdk.threads_leave()


    def __init__(self, toplevel, account, group):
        super(EditGroupDialog, self).__init__(u'Edit list', toplevel, account)

        self.user_backend = group.user_backend

        self.group_backend = group.group_backend

        name, description, index = group.get_parameters()

        self.name_entry.set_text(name)

        start = self.description_textbuffer.get_start_iter()

        self.description_textbuffer.insert(start, description)

        self.mode_radiobuttons[index].set_active(True)
