# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import gtk as Gtk

from gtk import gdk as Gdk


from polly import ENCODING, dispatch

from polly.twitter import ID_KEY

from polly.twitter.lead.group import NICK_KEY

from polly.twitter.account import Error

from polly.gui import DEFAULT_BORDER, CONNECTING_TEXT

from polly.gui.dialog import Dialog



class MembershipCheckButton(Gtk.CheckButton):
    def _toggle_task(self, previous_thread, active):
        if previous_thread is not None:
            previous_thread.join()

        try:
            if active:
                data = self.dialog.account.backend.send_group_membership(self.group_id, self.dialog.user.user_backend.id)
            else:
                data = self.dialog.account.backend.delete_group_membership(self.group_id, self.dialog.user.user_backend.id)
        except Error as error:
            Gdk.threads_enter()

            self.updating = True

            self.set_active(not active)

            self.updating = False

            self.dialog.stop(error.message)

            self.dialog.inner_box.set_sensitive(True)

            Gdk.threads_leave()
        else:
            self.dialog.user.user_backend.set_data(data[0])

            self.dialog.account.backend.group_backend_cache.set_data(data[1])

            Gdk.threads_enter()

            self.dialog.stop(None)

            self.dialog.inner_box.set_sensitive(True)

            Gdk.threads_leave()


    def _on_toggled(self, togglebutton):
        if not self.updating:
            self.dialog.inner_box.set_sensitive(False)

            self.dialog.start()

            self.dialog.thread = dispatch(self._toggle_task, self.dialog.thread, self.get_active())


    def __init__(self, dialog, group_data, membership_ids):
        super(MembershipCheckButton, self).__init__(group_data[NICK_KEY], False)
        self.connect(u'toggled', self._on_toggled)
        self.show()

        self.dialog = dialog

        self.group_id = group_data[ID_KEY]

        if self.group_id in membership_ids:
            self.updating = True

            self.set_active(True)

        self.updating = False



class MembershipsDialog(Dialog):
    def _try(self):
        self.label.set_text(CONNECTING_TEXT)

        self.start()

        self.thread = dispatch(self._try_task, self.thread)


    def _except(self, error):
        Gdk.threads_enter()

        self.stop(error.message)

        self.label.set_markup(u'Could not get memberships, <a href="try">try again</a>?')

        Gdk.threads_leave()


    def _try_task(self, previous_thread):
        if previous_thread is not None:
            previous_thread.join()

        try:
            dataset = self.account.backend.get_group_dataset_groups(self.account.backend.user_backend.id, -1)
        except Error as error:
            self._except(error)
        else:
            if dataset[2]:
                try:
                    subdataset = self.account.backend.get_group_dataset_memberships(self.user.user_backend.id, -1, 1)
                except Error as error:
                    self._except(error)
                else:
                    membership_ids = set()

                    for subdata in subdataset[2]:
                        membership_ids.add(subdata[1][ID_KEY])

                    Gdk.threads_enter()

                    self.stop(None)

                    self.label.set_alignment(0.0, 0.5)
                    self.label.set_markup(u'<big><b>Memberships of {}</b></big>'.format(self.user.user_nick))

                    self.inner_box = Gtk.VBox(True, 0)
                    self.inner_box.show()

                    for data in dataset[2]:
                        checkbutton = MembershipCheckButton(self, data[1], membership_ids)

                        self.inner_box.pack_start(checkbutton, True, True, 0)

                    self.outer_box.pack_start(self.inner_box, True, True, 0)

                    Gdk.threads_leave()
            else:
                Gdk.threads_enter()

                self.stop(None)

                self.label.set_markup(u'You have no lists, <a href="create">create one</a>?')

                Gdk.threads_leave()
        finally:
            Gdk.threads_enter()

            self.label.set_size_request(-1, -1)

            Gdk.threads_leave()


    def _on_label_activate_link(self, label, uri):
        uri = uri.decode(ENCODING)

        if uri == u'try':
            allocation = self.label.get_allocation()

            self.label.set_size_request(allocation.width, allocation.height)

            self._try()
        elif uri == u'create':
            self.hide()

            self.response(Gtk.RESPONSE_CLOSE)

            self.user.main_panel.create_group(self.user.stream.manager, self.user.stream.account)

        return True


    def _on_label_populate_popup(self, label, menu):
        menuitems = menu.get_children()

        menuitems[0].set_sensitive(False)
        menuitems[1].set_sensitive(False)


    def _on_close_button_clicked(self, button):
        self.response(Gtk.RESPONSE_CLOSE)


    def __init__(self, title, toplevel, account, user):
        super(MembershipsDialog, self).__init__(title, toplevel)

        self.account = account

        self.user = user

        self.label = Gtk.Label(None)
        self.label.connect(u'activate-link', self._on_label_activate_link)
        self.label.connect(u'populate-popup', self._on_label_populate_popup)
        self.label.show()

        self.outer_box = Gtk.VBox(False, 2 * DEFAULT_BORDER)
        self.outer_box.set_border_width(2 * DEFAULT_BORDER)
        self.outer_box.pack_start(self.label, False, False, 0)
        self.outer_box.show()

        self.content_frame.add(self.outer_box)

        close_button = Gtk.Button(None, Gtk.STOCK_CLOSE, False)
        close_button.connect(u'clicked', self._on_close_button_clicked)
        close_button.show()

        self.action_box.pack_start(close_button, False, False, 0)

        self._try()
