# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from gtk import gdk as Gdk


from polly import escape, dispatch

from polly.twitter.lead import user

from polly.twitter.account import Error

from polly.gui import show_url, run_confirmation_dialog

from polly.gui.twitter import DELETE_BOX_TOOLTIP, DELETE_SPINNER_TOOLTIP, COMPOSE_MESSAGE_TOOLTIP, CANCEL_MESSAGE_TOOLTIP, Link

from polly.gui.twitter.post import Post



class Message(Post):
    def _delete_task(self, previous_thread):
        if previous_thread is not None:
            previous_thread.join()

        try:
            data = self.stream.account.backend.delete_message(self.real_id)
        except Error as error:
            self.status_bar.popup(error.message)

            Gdk.threads_enter()

            self.delete_button.set_spinning(False)

            Gdk.threads_leave()
        else:
            if data is not None:
                self.recipient_backend.set_data(data[1])

                self.user_backend.set_data(data[2])


    def _set_normal_mg(self, color):
        super(Message, self)._set_normal_mg(color)

        self.recipient_link.set_color(color)


    def _set_font(self, unit, void):
        super(Message, self)._set_font(unit, void)

        self.recipient_link.box.set_size_request(-1, unit)

        self.footer_link.set_size_request(-1, unit + void)


    def _on_info_button_clicked(self):
        super(Message, self)._on_info_button_clicked()

        if self.info_button.toggled:
            self.recipient_avatar.box.set_tooltip_markup(self.recipient_name)

            self.recipient_link.set_markup(self.recipient_name)
        else:
            self.recipient_avatar.box.set_tooltip_markup(self.recipient_nick)

            self.recipient_link.set_markup(self.recipient_nick)


    def _on_footer_clicked(self):
        show_url(u'http://twitter.com/messages')


    def _on_message_button_clicked(self):
        if self.message_button.toggled:
            self.main_panel.cancel_message_to_message()
        else:
            self.main_panel.compose_message_to_message(self)


    def _on_delete_button_clicked(self):
        if run_confirmation_dialog(self.get_toplevel(), u'Delete message',
                                   u'Are you sure you want to delete this message?',
                                   u'You will not be able to restore it.'):
            self.delete_button.set_spinning(True)

            self.delete_thread = dispatch(self._delete_task, self.delete_thread)


    def _on_recipient_clicked(self):
        self.stream.manager.new_tweets_home_of_user(self.stream.account, [self.recipient_backend.id, True], self.stream)


    def _on_recipient_nick_changed(self):
        Gdk.threads_enter()

        self.recipient_nick = self.recipient_backend.data[user.NICK_KEY]

        if not self.info_button.toggled:
            self.recipient_avatar.box.set_tooltip_markup(self.recipient_nick)

            self.recipient_link.set_markup(self.recipient_nick)

        Gdk.threads_leave()


    def _on_recipient_name_changed(self):
        recipient_name = self.recipient_backend.data[user.NAME_KEY]

        if isinstance(recipient_name, unicode):
            recipient_name = recipient_name.strip()
        else:
            recipient_name = u''

        Gdk.threads_enter()

        if recipient_name:
            self.recipient_name = escape(recipient_name)
        else:
            self.recipient_name = self.recipient_nick

        if self.info_button.toggled:
            self.recipient_avatar.box.set_tooltip_markup(self.recipient_name)

            self.recipient_link.set_markup(self.recipient_name)

        Gdk.threads_leave()


    def __init__(self, real_id, recipient_data, user_data, message_data, stream, user_backend_cache, color_scheme, font_scheme, button_factory, avatar_factory, preview_factory, status_bar, main_panel):
        super(Message, self).__init__(real_id, user_data, message_data, stream, user_backend_cache, color_scheme, font_scheme, button_factory, avatar_factory, preview_factory, status_bar, main_panel)

        self.logged = False

        self.recipient_backend = user_backend_cache.acquire(recipient_data)

        self.recipient_avatar = avatar_factory.build(self.recipient_backend, 1, 0, self._on_recipient_clicked)

        self.message_button = button_factory.build(u'message', self._on_message_button_clicked)

        self.delete_button = button_factory.build(u'delete', self._on_delete_button_clicked)

        Gdk.threads_enter()

        self.spinners.append(self.recipient_avatar.spinner)

        self.recipient_avatar.box.connect(u'button-press-event', self._on_child_button_press_event)
        self.recipient_avatar.box.connect(u'focus-in-event', self._on_child_focus_in_event)

        self.recipient_link = Link(u'<small>{}</small>', self._on_recipient_clicked)
        self.recipient_link.box.connect(u'button-press-event', self._on_child_button_press_event)
        self.recipient_link.box.connect(u'focus-in-event', self._on_child_focus_in_event)

        self.footer_box.pack_end(self.recipient_link, True, True, 0)
        self.footer_box.pack_end(self.recipient_avatar, False, False, 0)

        self.vertical_box.pack_start(self.footer_link, False, False, 0)
        self.vertical_box.reorder_child(self.footer_link, 2)

        self.message_button.box.set_tooltip_text(COMPOSE_MESSAGE_TOOLTIP)

        self.delete_button.set_toggled(True)
        self.delete_button.box.set_tooltip_text(DELETE_BOX_TOOLTIP)
        self.delete_button.spinner.set_tooltip_text(DELETE_SPINNER_TOOLTIP)

        self.action_box.pack_end(self.message_button, True, True, 0)
        self.action_box.pack_end(self.delete_button, True, True, 0)

        self.spinners.append(self.delete_button.spinner)

        Gdk.threads_leave()

        self._connect()

        self.recipient_backend.block_updates()

        self._on_recipient_nick_changed()

        self._on_recipient_name_changed()

        self.recipient_nick_uid = self.recipient_backend.nick_notifier.connect(self._on_recipient_nick_changed)

        self.recipient_name_uid = self.recipient_backend.name_notifier.connect(self._on_recipient_name_changed)

        self.recipient_backend.allow_updates()

        self.delete_thread = None


    def isolate(self):
        if self.logged:
            self.stream.manager.message_log.decrement(self.real_id)

        if self.delete_thread is not None:
            self.delete_thread.join()

        self.recipient_backend.block_updates()

        self.recipient_backend.name_notifier.disconnect(self.recipient_name_uid)

        self.recipient_backend.nick_notifier.disconnect(self.recipient_nick_uid)

        self.recipient_backend.allow_updates()

        self._disconnect()

        self.recipient_link.isolate()

        self.delete_button.isolate()

        self.message_button.isolate()

        self.recipient_avatar.isolate()

        self.recipient_backend.release()

        super(Message, self).isolate()


    def set_messaging(self, messaging):
        self.message_button.set_toggled(messaging)

        if messaging:
            self.popup()

            self.message_button.box.set_tooltip_text(CANCEL_MESSAGE_TOOLTIP)
        else:
            self.message_button.box.set_tooltip_text(COMPOSE_MESSAGE_TOOLTIP)

        self._set_selected(messaging)


    def log(self):
        self.logged = True

        return self.stream.manager.message_log.increment(self.real_id)


    def notify(self, policy):
        self.stream.activate_indicator()

        self._notify()
