# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import gtk as Gtk

from gtk import gdk as Gdk


from polly import escape, dispatch

from polly.external.wraplabel import WrapLabel

from polly.twitter import CREATION_TIMESTAMP_KEY

from polly.twitter.lead import user

from polly.twitter.post import TEXT_KEY

from polly.twitter.post.result import USER_NICK_KEY

from polly.twitter.account import Error

from polly.gui import DEFAULT_BORDER, show_url

from polly.gui.twitter import Link, Item

from polly.gui.twitter.post import AVATAR_DIMENSION, Post



class Result(Post):
    def _expand_task(self, previous_thread):
        if previous_thread is not None:
            previous_thread.join()

        try:
            data = self.stream.account.backend.get_tweet_data(self.real_id)
        except Error as error:
            self.status_bar.popup(error.message)

            Gdk.threads_enter()

            self.expand_button.set_spinning(False)

            Gdk.threads_leave()
        else:
            self.stream.replace(self, data)


    def _set_font(self, unit, void):
        self.header.set_size_request(-1, unit)

        self.footer_link.box.set_size_request(-1, unit)

        self.vertical_box.set_spacing(void)


    def _on_footer_clicked(self):
        show_url(u'http://twitter.com/{}/status/{}'.format(self.header.get_text(), self.real_id))


    def _on_expand_button_clicked(self):
        self.expand_button.set_spinning(True)

        self.expand_thread = dispatch(self._expand_task, self.expand_thread)


    def _on_enter_notify_event(self, widget, event):
        self.enter()

        return False


    def _on_leave_notify_event(self, widget, event):
        allocation = self.outer_box.get_allocation()

        if (event.x < 0 or event.x >= allocation.width or
            event.y < 0 or event.y >= allocation.height):
            self.leave()

        return False


    def _on_button_press_event(self, widget, event):
        if event.button == 1:
            if self.get_focus_child() is None:
                self.footer_link.box.grab_focus()

        return False


    def _on_key_press_event(self, widget, event):
        if event.keyval == Gtk.keysyms.Tab:
            self.popup()

        return False


    def _on_unrealize(self, widget):
        if self.inside:
            self.leave()


    def _on_color_normal_fg_changed(self):
        normal_fg = self.color_scheme.normal_fg

        Gdk.threads_enter()

        self.expand_button.spinner.modify_fg(Gtk.STATE_NORMAL, normal_fg)

        Gdk.threads_leave()


    def _on_color_normal_bg_changed(self):
        normal_bg = self.color_scheme.normal_bg

        Gdk.threads_enter()

        self.outer_box.modify_bg(Gtk.STATE_NORMAL, normal_bg)

        Gdk.threads_leave()


    def _on_color_normal_mg_changed(self):
        normal_mg = self.color_scheme.normal_mg

        Gdk.threads_enter()

        if self.stamp.spinner is not None:
            self.stamp.spinner.modify_fg(Gtk.STATE_NORMAL, normal_mg)

        self.header.modify_fg(Gtk.STATE_NORMAL, normal_mg)

        self.wraplabel.modify_fg(Gtk.STATE_NORMAL, normal_mg)

        self.footer_link.set_color(normal_mg)

        Gdk.threads_leave()


    def __init__(self, real_id, result_data, stream, color_scheme, font_scheme, button_factory, stamp_factory, status_bar):
        self.real_id = real_id

        self.stream = stream

        self.color_scheme = color_scheme

        self.font_scheme = font_scheme

        self.status_bar = status_bar

        self.inside = False

        self.timestamp = result_data[CREATION_TIMESTAMP_KEY]

        self.stamp = stamp_factory.build(result_data[user.IMAGE_URL_KEY], AVATAR_DIMENSION)

        self.expand_button = button_factory.build(u'expand', self._on_expand_button_clicked)

        Gdk.threads_enter()

        self.header = Gtk.Label(None)
        self.header.set_alignment(0.0, 0.5)
        self.header.set_markup(u'<b>{}</b>'.format(result_data[USER_NICK_KEY]))
        self.header.show()

        self.wraplabel = WrapLabel(None)
        self.wraplabel.set_markup(escape(result_data[TEXT_KEY].strip()))
        self.wraplabel.show()

        self.footer_link = Link(u'<small>{}</small>', self._on_footer_clicked)
        self.footer_link.box.connect(u'button-press-event', self._on_child_button_press_event)
        self.footer_link.box.connect(u'focus-in-event', self._on_child_focus_in_event)

        self.expand_button.hide()

        footer_box = Gtk.HBox(False, DEFAULT_BORDER)
        footer_box.pack_start(self.footer_link, True, True, 0)
        footer_box.pack_start(self.expand_button, False, False, 0)
        footer_box.show()

        self.vertical_box = Gtk.VBox(False, 0)
        self.vertical_box.pack_start(self.header, False, False, 0)
        self.vertical_box.pack_start(self.wraplabel, True, True, 0)
        self.vertical_box.pack_start(footer_box, False, False, 0)
        self.vertical_box.show()

        horizontal_box = Gtk.HBox(False, DEFAULT_BORDER)
        horizontal_box.set_border_width(DEFAULT_BORDER)
        horizontal_box.pack_start(self.stamp, False, False, 0)
        horizontal_box.pack_start(self.vertical_box, True, True, 0)
        horizontal_box.show()

        self.outer_box = Gtk.EventBox()
        self.outer_box.connect(u'enter-notify-event', self._on_enter_notify_event)
        self.outer_box.connect(u'leave-notify-event', self._on_leave_notify_event)
        self.outer_box.connect(u'button-press-event', self._on_button_press_event)
        self.outer_box.connect(u'key-press-event', self._on_key_press_event)
        self.outer_box.connect(u'unrealize', self._on_unrealize)
        self.outer_box.add(horizontal_box)
        self.outer_box.show()

        separator = Gtk.HSeparator()
        separator.show()

        super(Item, self).__init__(False, 0)
        self.pack_start(self.outer_box, True, True, 0)
        self.pack_start(separator, False, False, 0)
        self.show()

        self.updating = True

        Gdk.threads_leave()

        self.color_scheme.block_updates()

        self.font_scheme.block_updates()

        self._on_color_normal_fg_changed()

        self._on_color_normal_bg_changed()

        self._on_color_normal_mg_changed()

        self._on_font_changed()

        self.color_normal_fg_uid = self.color_scheme.normal_fg_notifier.connect(self._on_color_normal_fg_changed)

        self.color_normal_bg_uid = self.color_scheme.normal_bg_notifier.connect(self._on_color_normal_bg_changed)

        self.color_normal_mg_uid = self.color_scheme.normal_mg_notifier.connect(self._on_color_normal_mg_changed)

        self.font_uid = self.font_scheme.notifier.connect(self._on_font_changed)

        self.color_scheme.allow_updates()

        self.font_scheme.allow_updates()

        self.expand_thread = None

        self.child_pressed = False


    def isolate(self):
        if self.expand_thread is not None:
            self.expand_thread.join()

        self.font_scheme.block_updates()

        self.color_scheme.block_updates()

        self.font_scheme.notifier.disconnect(self.font_uid)

        self.color_scheme.normal_mg_notifier.disconnect(self.color_normal_mg_uid)

        self.color_scheme.normal_bg_notifier.disconnect(self.color_normal_bg_uid)

        self.color_scheme.normal_fg_notifier.disconnect(self.color_normal_fg_uid)

        self.font_scheme.allow_updates()

        self.color_scheme.allow_updates()

        Gdk.threads_enter()

        self.updating = False

        Gdk.threads_leave()

        self.footer_link.isolate()

        self.expand_button.isolate()

        self.stamp.isolate()


    def enter(self):
        if not self.inside:
            if self.stream.current_item is not None:
                self.stream.current_item.leave()

            self.stream.current_item = self

            self.inside = True

            self.expand_button.show()


    def leave(self):
        self.expand_button.hide()

        self.inside = False

        if self.stream.current_item == self:
            self.stream.current_item = None


    def synchronize(self, timestamp):
        Gdk.threads_enter()

        updating = self.updating

        Gdk.threads_leave()

        if updating:
            small_time = self._get_small_time(timestamp)

            Gdk.threads_enter()

            self.footer_link.set_markup(small_time)

            Gdk.threads_leave()
