# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from urllib import quote


import gtk as Gtk

from gtk import gdk as Gdk


from polly import ENCODING, escape, dispatch

from polly.twitter import ID_KEY

from polly.twitter.lead import user

from polly.twitter.post.tweet import REPLIED_TWEET_ID_KEY, COORDINATES_KEY, SOURCE_KEY, PLACE_KEY, RETWEETER_COUNT_KEY, RETWEET_ID_KEY, FAVORITED_KEY

from polly.twitter.account import Error

from polly.gui import DEFAULT_BORDER, show_url, run_confirmation_dialog

from polly.gui.twitter import DELETE_BOX_TOOLTIP, DELETE_SPINNER_TOOLTIP, Link

from polly.gui.twitter.post import Post



COMPOSE_REPLY_TOOLTIP = u'Compose reply'

CANCEL_REPLY_TOOLTIP = u'Cancel reply'



class Tweet(Post):
    def _build_menu(self):
        super(Tweet, self)._build_menu()

        retweet_menuitem = Gtk.MenuItem(u'Edit and retweet', False)
        retweet_menuitem.connect(u'activate', self._on_retweet_menuitem_activate)
        retweet_menuitem.show()

        self.menu.append(retweet_menuitem)


    def _show_action_buttons(self):
        self.action_box.show()

        if self.retweet_button is not None:
            self.retweet_button.image.show()

        if self.favorite_button is not None:
            self.favorite_button.image.show()


    def _hide_action_buttons(self):
        if self.retweet_button is not None:
            if not self.retweet_button.toggled:
                if self.retweet_button.box.is_focus():
                    ancestor = self.retweet_button.box.get_ancestor(Gtk.Window)

                    if ancestor is not None:
                        ancestor.child_focus(Gtk.DIR_TAB_BACKWARD)

                self.retweet_button.image.hide()

        if self.favorite_button is not None:
            if not self.favorite_button.toggled:
                if self.favorite_button.box.is_focus():
                    ancestor = self.favorite_button.box.get_ancestor(Gtk.Window)

                    if ancestor is not None:
                        ancestor.child_focus(Gtk.DIR_TAB_BACKWARD)

                self.favorite_button.image.hide()

        if self.action_box.get_focus_child() is not None:
            ancestor = self.action_box.get_ancestor(Gtk.Window)

            if ancestor is not None:
                ancestor.child_focus(Gtk.DIR_TAB_BACKWARD)

        self.action_box.hide()


    def _add_action_buttons(self):
        if self.user_own:
            self.delete_button = self.button_factory.build(u'delete', self._on_delete_button_clicked)

            Gdk.threads_enter()

            self.delete_button.set_toggled(True)
            self.delete_button.box.set_tooltip_text(DELETE_BOX_TOOLTIP)
            self.delete_button.spinner.set_tooltip_text(DELETE_SPINNER_TOOLTIP)

            self.spinners.append(self.delete_button.spinner)

            Gdk.threads_leave()
        else:
            self.delete_button = None

            retweet_button = self.button_factory.build(u'retweet', self._on_retweet_button_clicked)

            Gdk.threads_enter()

            self.retweet_button = retweet_button
            self.retweet_button.box.connect(u'button-press-event', self._on_child_button_press_event)
            self.retweet_button.box.connect(u'focus-in-event', self._on_child_focus_in_event)

            self.permanent_box.pack_end(self.retweet_button, True, True, 0)

            self.spinners.append(self.retweet_button.spinner)

            Gdk.threads_leave()

        favorite_button = self.button_factory.build(u'favorite', self._on_favorite_button_clicked)

        self.reply_button = self.button_factory.build(u'reply', self._on_reply_button_clicked)

        Gdk.threads_enter()

        self.favorite_button = favorite_button
        self.favorite_button.box.connect(u'button-press-event', self._on_child_button_press_event)
        self.favorite_button.box.connect(u'focus-in-event', self._on_child_focus_in_event)

        self.reply_button.box.set_tooltip_text(COMPOSE_REPLY_TOOLTIP)

        self.permanent_box.pack_end(self.favorite_button, True, True, 0)
        self.permanent_box.show()

        self.action_box.pack_end(self.reply_button, True, True, 0)

        if self.user_own:
            self.action_box.pack_end(self.delete_button, True, True, 0)

        self.spinners.append(self.favorite_button.spinner)

        Gdk.threads_leave()


    def _connect_to_tweet_backend(self):
        self.tweet_backend.block_updates()

        self._on_tweet_retweeter_count_changed()

        self._on_tweet_favorited_changed()

        self.tweet_retweeter_count_uid = self.tweet_backend.retweeter_count_notifier.connect(self._on_tweet_retweeter_count_changed)

        self.tweet_favorited_uid = self.tweet_backend.notifiers[FAVORITED_KEY].connect(self._on_tweet_favorited_changed)

        if not self.user_own:
            self._on_tweet_retweet_id_changed()

            self.tweet_retweet_id_uid = self.tweet_backend.notifiers[RETWEET_ID_KEY].connect(self._on_tweet_retweet_id_changed)

        self.tweet_backend.allow_updates()


    def _actions_task(self, previous_thread):
        if previous_thread is not None:
            previous_thread.join()

        try:
            data = self.stream.account.backend.get_tweet_data(self.tweet_id)
        except Error as error:
            self.status_bar.popup(error.message)

            Gdk.threads_enter()

            self.actions_button.set_spinning(False)

            Gdk.threads_leave()
        else:
            self.user_backend.set_data(data[2])

            self.tweet_backend = self.tweet_backend_cache.acquire(data[3])

            Gdk.threads_enter()

            self.action_box.remove(self.actions_button)

            Gdk.threads_leave()

            self._add_action_buttons()

            self._connect_to_tweet_backend()

            self.actions_button.isolate()

            Gdk.threads_enter()

            self.spinners.remove(self.actions_button.spinner)

            self.actions_button.destroy()
            self.actions_button = None

            Gdk.threads_leave()


    def _retweet_task(self, previous_thread, retweet_id):
        if previous_thread is not None:
            previous_thread.join()

        try:
            if retweet_id is None:
                data = self.stream.account.backend.send_tweet_retweet(self.tweet_backend.id)
            else:
                data = self.stream.account.backend.delete_tweet_retweet(retweet_id, self.tweet_backend.id)
        except Error as error:
            self.status_bar.popup(error.message)
        else:
            self.stream.account.backend.user_backend.set_data(data[1])

            self.user_backend.set_data(data[2])

            if retweet_id is None:
                self.tweet_backend.set_value(RETWEET_ID_KEY, data[3][RETWEET_ID_KEY])
            else:
                self.tweet_backend.set_value(RETWEET_ID_KEY, None)

            self.tweet_backend.set_data(data[3])

        Gdk.threads_enter()

        self.retweet_button.set_spinning(False)

        Gdk.threads_leave()


    def _favorite_task(self, previous_thread, toggled):
        if previous_thread is not None:
            previous_thread.join()

        try:
            if toggled:
                data = self.stream.account.backend.delete_tweet_favorite(self.tweet_backend.id)
            else:
                data = self.stream.account.backend.send_tweet_favorite(self.tweet_backend.id)
        except Error as error:
            self.status_bar.popup(error.message)
        else:
            self.user_backend.set_data(data[2])

            if toggled:
                self.tweet_backend.set_value(FAVORITED_KEY, False)
            else:
                self.tweet_backend.set_value(FAVORITED_KEY, True)

            self.tweet_backend.set_data(data[3])

        Gdk.threads_enter()

        self.favorite_button.set_spinning(False)

        Gdk.threads_leave()


    def _delete_task(self, previous_thread):
        if previous_thread is not None:
            previous_thread.join()

        try:
            data = self.stream.account.backend.delete_tweet(self.tweet_backend.id)
        except Error as error:
            self.status_bar.popup(error.message)

            Gdk.threads_enter()

            self.delete_button.set_spinning(False)

            Gdk.threads_leave()
        else:
            if data is not None:
                self.user_backend.set_data(data[2])


    def _set_normal_mg(self, color):
        super(Tweet, self)._set_normal_mg(color)

        self.source_link.set_color(color)

        if self.place_link is not None:
            self.place_link.set_color(color)

        self.retweeter_count_link.set_color(color)

        if self.retweeter_backend is not None:
            self.retweeter_link.set_color(color)


    def _set_font(self, unit, void):
        super(Tweet, self)._set_font(unit, void)

        if self.retweeter_backend is not None:
            self.retweeter_link.box.set_size_request(-1, unit)

            self.footer_link.set_size_request(-1, unit + void)


    def _on_info_button_clicked(self):
        super(Tweet, self)._on_info_button_clicked()

        if self.info_button.toggled:
            self.wraplabel.hide()

            self.info_box.show()

            if self.retweeter_backend is not None:
                self.retweeter_avatar.box.set_tooltip_markup(self.retweeter_name)

                self.retweeter_link.set_markup(self.retweeter_name)
        else:
            self.info_box.hide()

            self.wraplabel.show()

            if self.retweeter_backend is not None:
                self.retweeter_avatar.box.set_tooltip_markup(self.retweeter_nick)

                self.retweeter_link.set_markup(self.retweeter_nick)


    def _on_context_button_clicked(self):
        self.stream.manager.new_tweets_context(self.stream.account, self.tweet_id, self.stream)


    def _on_coordinates_button_clicked(self):
        show_url(self.coordinates_url)


    def _on_source_clicked(self):
        show_url(self.source_url)


    def _on_place_clicked(self):
        show_url(self.place_url)


    def _on_retweeter_count_clicked(self):
        self.stream.manager.new_users_retweeters(self.stream.account, self.tweet_id, self.stream)


    def _on_footer_clicked(self):
        show_url(u'http://twitter.com/{}/status/{}'.format(self.user_nick, self.tweet_id))


    def _on_retweeter_clicked(self):
        self.stream.manager.new_tweets_home_of_user(self.stream.account, [self.retweeter_backend.id, True], self.stream)


    def _on_actions_button_clicked(self):
        self.actions_button.set_spinning(True)

        self.actions_thread = dispatch(self._actions_task, self.actions_thread)


    def _on_retweet_button_clicked(self):
        self.retweet_button.set_spinning(True)

        self.retweet_thread = dispatch(self._retweet_task, self.retweet_thread, self.retweet_id)


    def _on_favorite_button_clicked(self):
        self.favorite_button.set_spinning(True)

        self.favorite_thread = dispatch(self._favorite_task, self.favorite_thread, self.favorite_button.toggled)


    def _on_reply_button_clicked(self):
        if self.reply_button.toggled:
            self.main_panel.cancel_reply()
        else:
            self.main_panel.compose_reply(self)


    def _on_delete_button_clicked(self):
        if run_confirmation_dialog(self.get_toplevel(), u'Delete tweet',
                                   u'Are you sure you want to delete this tweet?',
                                   u'You will not be able to restore it.'):
            self.delete_button.set_spinning(True)

            self.delete_thread = dispatch(self._delete_task, self.delete_thread)


    def _on_retweet_menuitem_activate(self, menuitem):
        if self.retweeter_backend is None:
            retweeter_nick = None
        else:
            retweeter_nick = self.retweeter_nick

        self.main_panel.compose_retweet(retweeter_nick, self.user_nick, self)


    def _on_user_protected_changed(self):
        Gdk.threads_enter()

        if self.retweet_button is not None:
            if self.user_backend.data[user.PROTECTED_KEY]:
                self.retweet_button.hide()
            else:
                self.retweet_button.show()

        Gdk.threads_leave()


    def _on_retweeter_nick_changed(self):
        Gdk.threads_enter()

        self.retweeter_nick = self.retweeter_backend.data[user.NICK_KEY]

        if not self.info_button.toggled:
            self.retweeter_avatar.box.set_tooltip_markup(self.retweeter_nick)

            self.retweeter_link.set_markup(self.retweeter_nick)

        Gdk.threads_leave()


    def _on_retweeter_name_changed(self):
        retweeter_name = self.retweeter_backend.data[user.NAME_KEY]

        if isinstance(retweeter_name, unicode):
            retweeter_name = retweeter_name.strip()
        else:
            retweeter_name = u''

        Gdk.threads_enter()

        if retweeter_name:
            self.retweeter_name = escape(retweeter_name)
        else:
            self.retweeter_name = self.retweeter_nick

        if self.info_button.toggled:
            self.retweeter_avatar.box.set_tooltip_markup(self.retweeter_name)

            self.retweeter_link.set_markup(self.retweeter_name)

        Gdk.threads_leave()


    def _on_tweet_retweeter_count_changed(self):
        retweeter_count = self.tweet_backend.data[RETWEETER_COUNT_KEY]

        Gdk.threads_enter()

        self.retweeter_count_link.set_markup(u'Retweeters: {}'.format(retweeter_count))

        Gdk.threads_leave()


    def _on_tweet_retweet_id_changed(self):
        Gdk.threads_enter()

        self.retweet_id = self.tweet_backend.data[RETWEET_ID_KEY]

        if self.retweet_id is None:
            self.retweet_button.set_toggled(False)
            self.retweet_button.box.set_tooltip_text(u'Retweet')
            self.retweet_button.spinner.set_tooltip_text(u'Retweeting\u2026')

            if not self.inside:
                self.retweet_button.image.hide()
        else:
            self.retweet_button.set_toggled(True)
            self.retweet_button.box.set_tooltip_text(u'Unretweet')
            self.retweet_button.spinner.set_tooltip_text(u'Unretweeting\u2026')

            self.retweet_button.image.show()

        Gdk.threads_leave()


    def _on_tweet_favorited_changed(self):
        Gdk.threads_enter()

        self.favorite_button.set_toggled(self.tweet_backend.data[FAVORITED_KEY])

        if self.favorite_button.toggled:
            self.favorite_button.box.set_tooltip_text(u'Unfavorite')
            self.favorite_button.spinner.set_tooltip_text(u'Unfavoriting\u2026')

            self.favorite_button.image.show()
        else:
            self.favorite_button.box.set_tooltip_text(u'Favorite')
            self.favorite_button.spinner.set_tooltip_text(u'Favoriting\u2026')

            if not self.inside:
                self.favorite_button.image.hide()

        Gdk.threads_leave()


    def __init__(self, real_id, retweeter_data, user_data, tweet_data, stream, user_backend_cache, tweet_backend_cache, color_scheme, font_scheme, button_factory, avatar_factory, preview_factory, status_bar, main_panel):
        super(Tweet, self).__init__(real_id, user_data, tweet_data, stream, user_backend_cache, color_scheme, font_scheme, button_factory, avatar_factory, preview_factory, status_bar, main_panel)

        self.tweet_backend_cache = tweet_backend_cache

        self.button_factory = button_factory

        self.logged = False

        self.tweet_id = tweet_data[ID_KEY]

        if tweet_data[REPLIED_TWEET_ID_KEY] is None:
            self.context_button = None
        else:
            self.context_button = button_factory.build(u'context', self._on_context_button_clicked)

            Gdk.threads_enter()

            self.context_button.set_toggled(True)
            self.context_button.box.set_tooltip_text(u'Show context')

            self.indicator_box.pack_end(self.context_button, True, True, 0)

            Gdk.threads_leave()

        coordinates = tweet_data[COORDINATES_KEY]

        if coordinates is not None:
            try:
                if coordinates[u'type'] == u'Point':
                    coordinates = coordinates[u'coordinates']

                    if not isinstance(coordinates, list):
                        raise TypeError

                    if len(coordinates) != 2:
                        raise TypeError

                    if (not isinstance(coordinates[0], float) or
                        not isinstance(coordinates[1], float)):
                        raise TypeError
                else:
                    raise TypeError
            except (TypeError, KeyError):
                coordinates = None

        if coordinates is None:
            self.coordinates_button = None
        else:
            self.coordinates_url = u'http://maps.google.com/maps?q={},{}'.format(coordinates[1], coordinates[0])

            self.coordinates_button = button_factory.build(u'coordinates', self._on_coordinates_button_clicked)

            Gdk.threads_enter()

            self.coordinates_button.set_toggled(True)
            self.coordinates_button.box.set_tooltip_text(u'Show coordinates')

            self.indicator_box.pack_end(self.coordinates_button, True, True, 0)

            Gdk.threads_leave()

        place = tweet_data[PLACE_KEY]

        if place is not None:
            try:
                place = place[u'full_name'].strip()

                if not isinstance(place, unicode):
                    raise TypeError
            except (TypeError, KeyError):
                place = None

        if place is None:
            self.place_link = None
        else:
            processed_place = quote(place.encode(ENCODING), b'').decode(ENCODING)

            self.place_url = u'http://maps.google.com/maps?q={}'.format(processed_place)

            Gdk.threads_enter()

            self.place_link = Link(u'<small>{}</small>', self._on_place_clicked)
            self.place_link.set_markup(u'From {}'.format(escape(place)))

            Gdk.threads_leave()

        source = tweet_data[SOURCE_KEY]

        if source == u'web':
            self.source_url = u'http://www.twitter.com'

            source_name = source
        else:
            index = source.find(u'"') + 1

            self.source_url = source[index:source.find(u'"', index)]

            index = source.find(u'>') + 1

            source_name = escape(source[index:source.find(u'<', index)].strip())

        Gdk.threads_enter()

        self.source_link = Link(u'<small>{}</small>', self._on_source_clicked)
        self.source_link.set_markup(u'Via {}'.format(source_name))

        self.retweeter_count_link = Link(u'<small>{}</small>', self._on_retweeter_count_clicked)
        self.retweeter_count_link.set_markup(u'View retweeters')

        self.info_box = Gtk.VBox(True, 0)

        if self.place_link is not None:
            self.info_box.pack_start(self.place_link, True, True, 0)

        self.info_box.pack_start(self.source_link, True, True, 0)
        self.info_box.pack_start(self.retweeter_count_link, True, True, 0)

        self.inner_box.pack_start(self.info_box, True, True, 0)
        self.inner_box.reorder_child(self.info_box, 1)

        self.permanent_box = Gtk.HBox(True, DEFAULT_BORDER)

        self.footer_box.pack_end(self.permanent_box, False, False, 0)
        self.footer_box.reorder_child(self.permanent_box, 0)

        Gdk.threads_leave()

        self.retweet_button = None

        self.favorite_button = None

        if retweeter_data is None:
            self.retweeter_backend = None

            self.tweet_backend = self.tweet_backend_cache.acquire(tweet_data)

            self.actions_button = None

            Gdk.threads_enter()

            self.footer_link.box.connect(u'button-press-event', self._on_child_button_press_event)
            self.footer_link.box.connect(u'focus-in-event', self._on_child_focus_in_event)

            self.footer_box.pack_end(self.footer_link, True, True, 0)

            Gdk.threads_leave()

            self._add_action_buttons()
        else:
            self.retweeter_backend = user_backend_cache.acquire(retweeter_data)

            self.tweet_backend = None

            self.retweeter_avatar = avatar_factory.build(self.retweeter_backend, 1, 0, self._on_retweeter_clicked)

            self.actions_button = self.button_factory.build(u'actions', self._on_actions_button_clicked)

            Gdk.threads_enter()

            self.spinners.append(self.retweeter_avatar.spinner)

            self.retweeter_avatar.box.connect(u'button-press-event', self._on_child_button_press_event)
            self.retweeter_avatar.box.connect(u'focus-in-event', self._on_child_focus_in_event)

            self.retweeter_link = Link(u'<small>{}</small>', self._on_retweeter_clicked)
            self.retweeter_link.box.connect(u'button-press-event', self._on_child_button_press_event)
            self.retweeter_link.box.connect(u'focus-in-event', self._on_child_focus_in_event)

            self.footer_box.pack_end(self.retweeter_link, True, True, 0)
            self.footer_box.pack_end(self.retweeter_avatar, False, False, 0)

            self.vertical_box.pack_start(self.footer_link, False, False, 0)
            self.vertical_box.reorder_child(self.footer_link, 2)

            self.actions_button.set_toggled(True)
            self.actions_button.box.set_tooltip_text(u'Open actions')
            self.actions_button.spinner.set_tooltip_text(u'Opening actions\u2026')

            self.action_box.pack_end(self.actions_button, True, True, 0)

            self.spinners.append(self.actions_button.spinner)

            Gdk.threads_leave()

        self._connect()

        if not self.user_own:
            self.user_backend.block_updates()

            self._on_user_protected_changed()

            self.user_protected_uid = self.user_backend.protected_notifier.connect(self._on_user_protected_changed)

            self.user_backend.allow_updates()

        if self.retweeter_backend is None:
            self._connect_to_tweet_backend()
        else:
            self.retweeter_backend.block_updates()

            self._on_retweeter_nick_changed()

            self._on_retweeter_name_changed()

            self.retweeter_nick_uid = self.retweeter_backend.nick_notifier.connect(self._on_retweeter_nick_changed)

            self.retweeter_name_uid = self.retweeter_backend.name_notifier.connect(self._on_retweeter_name_changed)

            self.retweeter_backend.allow_updates()

        self.actions_thread = None

        self.retweet_thread = None

        self.favorite_thread = None

        self.delete_thread = None


    def isolate(self):
        if self.logged:
            self.stream.manager.tweet_log.decrement(self.tweet_id)

        if self.delete_thread is not None:
            self.delete_thread.join()

        if self.favorite_thread is not None:
            self.favorite_thread.join()

        if self.retweet_thread is not None:
            self.retweet_thread.join()

        if self.actions_thread is not None:
            self.actions_thread.join()

        if self.tweet_backend is not None:
            self.tweet_backend.block_updates()

            if not self.user_own:
                self.tweet_backend.notifiers[RETWEET_ID_KEY].disconnect(self.tweet_retweet_id_uid)

            self.tweet_backend.notifiers[FAVORITED_KEY].disconnect(self.tweet_favorited_uid)

            self.tweet_backend.retweeter_count_notifier.disconnect(self.tweet_retweeter_count_uid)

            self.tweet_backend.allow_updates()

        if self.retweeter_backend is not None:
            self.retweeter_backend.block_updates()

            self.retweeter_backend.name_notifier.disconnect(self.retweeter_name_uid)

            self.retweeter_backend.nick_notifier.disconnect(self.retweeter_nick_uid)

            self.retweeter_backend.allow_updates()

        if not self.user_own:
            self.user_backend.block_updates()

            self.user_backend.protected_notifier.disconnect(self.user_protected_uid)

            self.user_backend.allow_updates()

        self._disconnect()

        if self.actions_button is None:
            self.reply_button.isolate()

            self.favorite_button.isolate()

            if self.user_own:
                self.delete_button.isolate()
            else:
                self.retweet_button.isolate()
        else:
            self.actions_button.isolate()

        if self.tweet_backend is not None:
            self.tweet_backend.release()

        if self.retweeter_backend is not None:
            self.retweeter_link.isolate()

            self.retweeter_avatar.isolate()

            self.retweeter_backend.release()

        self.retweeter_count_link.isolate()

        self.source_link.isolate()

        if self.place_link is not None:
            self.place_link.isolate()

        if self.coordinates_button is not None:
            self.coordinates_button.isolate()

        if self.context_button is not None:
            self.context_button.isolate()

        super(Tweet, self).isolate()


    def set_replying(self, replying):
        self.reply_button.set_toggled(replying)

        if replying:
            self.popup()

            self.reply_button.box.set_tooltip_text(CANCEL_REPLY_TOOLTIP)
        else:
            self.reply_button.box.set_tooltip_text(COMPOSE_REPLY_TOOLTIP)

        self._set_selected(replying)


    def log(self):
        self.logged = True

        return self.stream.manager.tweet_log.increment(self.tweet_id)


    def notify(self, policy):
        if (policy > 3 or
           (policy > 2 and self.retweeter_backend is None) or
           (policy > 1 and self.mentioned)):
            self._notify()
