# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import os, shutil, tarfile

from tarfile import TarError


import gio as Gio


from polly import dispatch, Notifier, Updatable



FAMILY_NAME = u'iconsets'


DEFAULT_NAME = u'default16'



class Error(Exception):
    def __init__(self, message):
        self.message = message



class ControllerBackend(Updatable):
    def _notify(self):
        thread = dispatch(self.notifier.notify)

        thread.join()


    def _add_names(self, data_path, unordered_names):
        if os.path.isdir(data_path):
            try:
                for name in os.listdir(data_path):
                    if os.path.isdir(os.path.join(data_path, name)):
                        unordered_names.add(name)
            except OSError:
                pass


    def _ensure_name(self):
        if self.name in self.names:
            return False

        length = len(self.names)

        if length == 0:
            self.name = DEFAULT_NAME
        else:
            for index in xrange(length):
                if self.names[index] > self.name:
                    break

            if index == length:
                index = length - 1

            self.name = self.names[index]

        return True


    def _update_added_names(self):
        added_names = set()

        self._add_names(self.data_paths[0], added_names)

        if self.added_names == added_names:
            return False

        self.added_names = added_names

        return True


    def _update_names(self):
        unordered_names = self.added_names.copy()

        for data_path in self.data_paths[1:]:
            self._add_names(data_path, unordered_names)

        names = list(unordered_names)

        names.sort()

        if self.names == names:
            return False

        self.names = names

        return True


    def _update_name(self):
        name = self.name_setting.get_value()

        if name not in self.names:
            name = DEFAULT_NAME

        if self.name == name:
            return False

        self.name = name

        return True


    def _update_strings(self):
        self.strings = []

        for data_path in self.data_paths:
            self.strings.append(os.path.join(data_path, self.name, u'{}.svg'))


    def _update_sub_monitors(self):
        for uid in self.sub_monitors.viewkeys():
            self.sub_monitors[uid].disconnect(uid)

            self.sub_monitors[uid].cancel()

        self.sub_monitors.clear()

        for data_path in self.data_paths:
            sub_file = Gio.File(os.path.join(data_path, self.name))

            sub_monitor = sub_file.monitor_directory(Gio.FILE_MONITOR_NONE, None)

            uid = sub_monitor.connect(u'changed', self._on_sub_monitor_changed)

            self.sub_monitors[uid] = sub_monitor


    # NOTE: The G_IS_OBJECT assertion fails on g_object_set_qdata when
    # this method is called. Seems to be a mostly harmless bug in PyGTK.


    def _on_sub_monitor_changed(self, monitor, file, other_file, event_type):
        self._notify()


    def _on_monitor_changed(self, monitor, file, other_file, event_type):
        self._acquire_update()

        added_names_changed = self._update_added_names()

        if self._update_names():
            name_changed = self._ensure_name()

            self.names_callback()

            if name_changed:
                self.updating_names = True

                self.name_setting.set_value(self.name)

                self.updating_names = False

                self._update_strings()

                self._update_sub_monitors()

                self._notify()
        else:
            if added_names_changed:
                self.added_names_callback()

        self._release_update()


    def _on_name_changed(self):
        if not self.updating_names:
            self._acquire_update()

            if self._update_name():
                self.name_callback()

                self._update_strings()

                self._update_sub_monitors()

                self._notify()

            self._release_update()


    def __init__(self, added_names_callback, names_callback, name_callback, cache_path, data_paths, cache_path_lock, setting_factory):
        super(ControllerBackend, self).__init__()

        self.added_names_callback = added_names_callback

        self.names_callback = names_callback

        self.name_callback = name_callback

        self.cache_path = os.path.join(cache_path, FAMILY_NAME)

        self.data_paths = []

        for data_path in data_paths:
            self.data_paths.append(os.path.join(data_path, FAMILY_NAME))

        self.cache_path_lock = cache_path_lock

        self.updating_names = False

        self.name = DEFAULT_NAME

        self.names = []

        self.added_names = set()

        self.monitors = {}

        for data_path in self.data_paths:
            file = Gio.File(data_path)

            monitor = file.monitor_directory(Gio.FILE_MONITOR_NONE, None)

            uid = monitor.connect(u'changed', self._on_monitor_changed)

            self.monitors[uid] = monitor

        self.sub_monitors = {}

        self.notifier = Notifier()

        self.name_setting = setting_factory.build_string(u'iconset', u'name', self._on_name_changed)

        self._update_added_names()

        self._update_names()

        self._update_name()

        self._update_strings()

        self._update_sub_monitors()


    def isolate(self):
        self.name_setting.isolate()

        for uid in self.sub_monitors.viewkeys():
            self.sub_monitors[uid].disconnect(uid)

            self.sub_monitors[uid].cancel()

        for uid in self.monitors.viewkeys():
            self.monitors[uid].disconnect(uid)

            self.monitors[uid].cancel()

        self.name_callback = None

        self.names_callback = None

        self.added_names_callback = None


    def set_name(self, name):
        if self.name != name:
            self.name_setting.set_value(name)


    def add_new(self, path):
        basename = os.path.basename(path)

        if basename.endswith(u'2'):
            name = basename[0:-8]
        else:
            name = basename[0:-7]

        if not name:
            raise Error(u'The iconset must have a name.')

        data_home = self.data_paths[0]

        permanent_path = os.path.join(data_home, name)

        if os.path.isdir(permanent_path):
            raise Error(u'An iconset with the name <i>{}</i> has been already added.'.format(name))

        temporary_path = os.path.join(self.cache_path, name)

        try:
            self.cache_path_lock.acquire()

            try:
                if os.path.exists(temporary_path):
                    if os.path.isfile(temporary_path):
                        os.remove(temporary_path)
                    elif os.path.isdir(temporary_path):
                        shutil.rmtree(temporary_path)

                os.makedirs(temporary_path)
            finally:
                self.cache_path_lock.release()
        except OSError:
            raise Error(u'Check if you have permission to write the folder <i>{}</i>.'.format(temporary_path))

        try:
            with tarfile.open(path) as file:
                for member in file.getmembers():
                    if not member.isfile() and not member.issym():
                        raise Error(u'Members of the iconset must be files or symbolic links.')

                    if member.name != os.path.basename(member.name):
                        raise Error(u'No member of the iconset can be inside a folder.')

                file.extractall(temporary_path)

            if not os.path.isdir(data_home):
                os.makedirs(data_home)

            if os.path.isfile(permanent_path):
                os.remove(permanent_path)

            shutil.copytree(temporary_path, permanent_path, True)
        except (IOError, TarError, OSError):
            raise Error(u'Check if the file is valid, if you have permission to read it, and if you have permission to read and write to the folders <i>{}</i> and <i>{}</i>.'.
                        format(temporary_path, permanent_path))
        finally:
            if os.path.isdir(temporary_path):
                try:
                    shutil.rmtree(temporary_path)
                except OSError:
                    pass

        self._acquire_update()

        self._update_added_names()

        self._update_names()

        self.name = name

        self.names_callback()

        self.updating_names = True

        self.name_setting.set_value(self.name)

        self.updating_names = False

        self._update_strings()

        self._update_sub_monitors()

        self._notify()

        self._release_update()


    def remove_current(self):
        path = os.path.join(self.data_paths[0], self.name)

        if os.path.isdir(path):
            try:
                shutil.rmtree(path)
            except OSError:
                raise Error(u'Check if you have permission to write the folder <i>{}</i>.'.format(path))
