# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



# NOTE: Due to a bug in httplib2, which is already fixed
# in trunk, if an user attempts to connect to an invalid
# proxy, the connection breaks and a restart is needed.



import socket

from threading import Lock


import socks



PROXY_DIR = u'proxy'


DEFAULT_PORT = 8080

LOWER_PORT = 0

UPPER_PORT = 65535

DEFAULT_TYPE = 1


TYPES = [socks.PROXY_TYPE_SOCKS4,
         socks.PROXY_TYPE_SOCKS5,
         socks.PROXY_TYPE_HTTP]



class ControllerBackend(object):
    def _update_enabled(self):
        enabled = self.enabled_setting.get_value()

        if self.enabled == enabled:
            return False

        self.lock.acquire()

        self.enabled = enabled

        self.lock.release()

        # NOTE: I am not entirely sure this is thread-safe,
        # particularly considering third-party libraries.

        if self.enabled:
            socket.socket = self.proxied_socket
        else:
            socket.socket = self.regular_socket

        return True


    def _update_host(self):
        host = self.host_setting.get_value()

        if self.host == host:
            return False

        self.lock.acquire()

        self.host = host

        self.lock.release()

        return True


    def _update_port(self):
        port = self.port_setting.get_value()

        if port < LOWER_PORT or port > UPPER_PORT:
            port = DEFAULT_PORT

        if self.port == port:
            return False

        self.lock.acquire()

        self.port = port

        self.lock.release()

        return True


    def _update_type(self):
        type = self.type_setting.get_value()

        if type < 0 or type > 2:
            type = DEFAULT_TYPE

        if self.type == type:
            return False

        self.lock.acquire()

        self.type = type

        self.lock.release()

        return True


    def _update_proxy(self):
        socks.setdefaultproxy(TYPES[self.type], self.host, self.port, True)


    def _on_enabled_changed(self):
        if self._update_enabled():
            self.enabled_callback()


    def _on_host_changed(self):
        if self._update_host():
            self._update_proxy()

            self.host_callback()


    def _on_port_changed(self):
        if self._update_port():
            self._update_proxy()

            self.port_callback()


    def _on_type_changed(self):
        if self._update_type():
            self._update_proxy()

            self.type_callback()


    def __init__(self, enabled_callback, host_callback, port_callback, type_callback, setting_factory):
        self.enabled_callback = enabled_callback

        self.host_callback = host_callback

        self.port_callback = port_callback

        self.type_callback = type_callback

        self.enabled = False

        self.host = u'localhost'

        self.port = DEFAULT_PORT

        self.type = DEFAULT_TYPE

        self.lock = Lock()

        self.regular_socket = socket.socket

        self.proxied_socket = socks.socksocket

        self.enabled_setting = setting_factory.build_boolean(PROXY_DIR, u'enabled', self._on_enabled_changed)

        self.host_setting = setting_factory.build_string(PROXY_DIR, u'host', self._on_host_changed)

        self.port_setting = setting_factory.build_integer(PROXY_DIR, u'port', self._on_port_changed)

        self.type_setting = setting_factory.build_integer(PROXY_DIR, u'type', self._on_type_changed)

        self._update_enabled()

        self._update_host()

        self._update_port()

        self._update_type()

        self._update_proxy()


    def isolate(self):
        self.type_setting.isolate()

        self.port_setting.isolate()

        self.host_setting.isolate()

        self.enabled_setting.isolate()

        self.type_callback = None

        self.port_callback = None

        self.host_callback = None

        self.enabled_callback = None


    def get_values(self):
        self.lock.acquire()

        enabled = self.enabled

        host = self.host

        port = self.port

        type = self.type

        self.lock.release()

        return enabled, host, port, type


    def set_enabled(self, enabled):
        if self.enabled != enabled:
            self.enabled_setting.set_value(enabled)


    def set_host(self, host):
        if self.host != host:
            self.host_setting.set_value(host)


    def set_port(self, port):
        if self.port != port:
            self.port_setting.set_value(port)


    def set_type(self, type):
        if self.type != type:
            self.type_setting.set_value(type)
