# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import json, socket

from collections import OrderedDict

from httplib import HTTPException

from threading import Lock

from urllib import urlencode


import httplib2

from httplib2 import HttpLib2Error

from socks import ProxyError


from polly import STATUS_KEY, ENCODING



SHORTENER_DIR = u'shortener'


DEFAULT_MINIMUM = 30

LOWER_MINIMUM = 20

UPPER_MINIMUM = 140

DEFAULT_NAME = u'is.gd'


RATE_MESSAGE = u'You have exceeded the request limit for the {} shortener. Please wait some time before trying again.'

CODE_MESSAGE = u'The {0} shortener responded with error {1}. Please try again.'



class Exit(Exception):
    pass



class Error(Exception):
    def __init__(self, message):
        self.message = message



class Shortener(object):
    def _request(self, url, method, body, headers):
        http_client = httplib2.Http()

        # NOTE: The usage of AttributeError is actually due to
        # Issue 96 in httplib2, which is already fixed in trunk.

        try:
            response, content = http_client.request(url, method, body, headers)
        except (HttpLib2Error, HTTPException, ProxyError, socket.error, AttributeError):
            raise Error(u'Could not connect to the {} shortener. Check your internet connection and try again.'.format(self.name))

        status = long(response[STATUS_KEY])

        if status != 200:
            if status in self.messages.viewkeys():
                raise Error(self.messages[status])
            else:
                raise Error(CODE_MESSAGE.format(self.name, status))

        try:
            data = json.loads(content)

            processed_data = self._process(data)

            if not isinstance(processed_data, unicode):
                raise TypeError
        except (ValueError, TypeError, KeyError):
            raise Error(u'Unexpected response from the {} shortener. Please try again.'.format(self.name))

        return processed_data


    def _request_get(self, url, parameters):
        return self._request(b'{}?{}'.format(url, urlencode(parameters)), b'GET', None, None)


    def _request_post(self, url, body, headers):
        return self._request(url, b'POST', body, headers)


    def __init__(self, name):
        self.name = name



class ISGDShortener(Shortener):
    def _process(self, data):
        return data[u'shorturl']


    def __init__(self):
        super(ISGDShortener, self).__init__(u'is.gd')

        self.messages = {502: RATE_MESSAGE.format(self.name),
                         503: u'The {} shortener seems to be down. Please wait a few minutes and try again.'.format(self.name)}


    def shorten(self, url):
        parameters = {b'format': b'json',
                      b'url': url.encode(ENCODING)}

        return self._request_get(b'http://is.gd/create.php', parameters)



class BITLYShortener(Shortener):
    def _process(self, data):
        status = long(data[u'status_code'])

        if status == 200:
            return data[u'data'][u'url']
        elif status == 403:
            raise Error(RATE_MESSAGE.format(self.name))
        else:
            raise Error(CODE_MESSAGE.format(self.name, status))


    def __init__(self):
        super(BITLYShortener, self).__init__(u'bit.ly')

        self.messages = {}


    def shorten(self, url):
        parameters = {b'login': b'pollyclient',
                      b'apiKey': b'R_5547dabb0f1fc372428fb93b5223f28b',
                      b'format': b'json',
                      b'longUrl': url.encode(ENCODING)}

        return self._request_get(b'http://api.bitly.com/v3/shorten', parameters)



class GOOGLShortener(Shortener):
    def _process(self, data):
        return data[u'id']


    def __init__(self):
        super(GOOGLShortener, self).__init__(u'goo.gl')

        self.messages = {}


    def shorten(self, url):
        parameters = {b'longUrl': url.encode(ENCODING)}

        return self._request_post(b'https://www.googleapis.com/urlshortener/v1/url?key=AIzaSyBJBaLGNJXHh7wVH0beHabz-OxasC9vsWI', json.dumps(parameters), {b'Content-Type': b'application/json'})



class ControllerBackend(object):
    def _update_enabled(self):
        enabled = self.enabled_setting.get_value()

        if self.enabled == enabled:
            return False

        self.lock.acquire()

        self.enabled = enabled

        self.lock.release()

        return True


    def _update_minimum(self):
        minimum = self.minimum_setting.get_value()

        if minimum < LOWER_MINIMUM or minimum > UPPER_MINIMUM:
            minimum = DEFAULT_MINIMUM

        if self.minimum == minimum:
            return False

        self.lock.acquire()

        self.minimum = minimum

        self.lock.release()

        return True


    def _update_name(self):
        name = self.name_setting.get_value()

        if name not in self.shorteners.viewkeys():
            name = DEFAULT_NAME

        if self.name == name:
            return False

        self.lock.acquire()

        self.name = name

        self.lock.release()

        return True


    def _on_enabled_changed(self):
        if self._update_enabled():
            self.enabled_callback()


    def _on_minimum_changed(self):
        if self._update_minimum():
            self.minimum_callback()


    def _on_name_changed(self):
        if self._update_name():
            self.name_callback()


    def __init__(self, enabled_callback, minimum_callback, name_callback, setting_factory):
        self.enabled_callback = enabled_callback

        self.minimum_callback = minimum_callback

        self.name_callback = name_callback

        self.enabled = True

        self.minimum = DEFAULT_MINIMUM

        self.name = DEFAULT_NAME

        self.lock = Lock()

        self.enabled_setting = setting_factory.build_boolean(SHORTENER_DIR, u'enabled', self._on_enabled_changed)

        self.minimum_setting = setting_factory.build_integer(SHORTENER_DIR, u'minimum', self._on_minimum_changed)

        self.name_setting = setting_factory.build_string(SHORTENER_DIR, u'name', self._on_name_changed)

        self.shorteners = OrderedDict()

        for Class in [ISGDShortener, BITLYShortener, GOOGLShortener]:
            shortener = Class()

            self.shorteners[shortener.name] = shortener

        self._update_enabled()

        self._update_minimum()

        self._update_name()


    def isolate(self):
        self.name_setting.isolate()

        self.minimum_setting.isolate()

        self.enabled_setting.isolate()

        self.name_callback = None

        self.minimum_callback = None

        self.enabled_callback = None


    def set_enabled(self, enabled):
        if self.enabled != enabled:
            self.enabled_setting.set_value(enabled)


    def set_minimum(self, minimum):
        if self.minimum != minimum:
            self.minimum_setting.set_value(minimum)


    def set_name(self, name):
        if self.name != name:
            self.name_setting.set_value(name)


    def shorten(self, url):
        self.lock.acquire()

        enabled = self.enabled

        minimum = self.minimum

        shortener = self.shorteners[self.name]

        self.lock.release()

        if enabled and len(url) >= minimum:
            return shortener.shorten(url)
        else:
            raise Exit
