# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



# NOTE: Inclusion of entities will
# be the API default in the future.



import json, socket

from calendar import timegm

from email.utils import parsedate

from httplib import HTTPException

from threading import Lock

from urllib import urlencode

from urlparse import parse_qs


import httplib2

from httplib2 import HttpLib2Error

from socks import ProxyError


from polly import STATUS_KEY, ENCODING, Updatable, Notifier

from polly.external import oauth2

from polly.twitter import ID_KEY, REQUEST_TIMESTAMP_KEY, CREATION_TIMESTAMP_KEY, ENTITIES_KEY, MENTIONS_KEY, DISPLAY_URL_KEY, EXPANDED_URL_KEY

from polly.twitter.lead import user, relationship, group

from polly.twitter.post import tweet, message, result

from polly.twitter.topic import category, trend



PREVIOUS_CURSOR_KEY = u'previous_cursor'

NEXT_CURSOR_KEY = u'next_cursor'

USER_KEY = u'user'

USERS_KEY = u'users'

RELATIONSHIP_KEY = u'relationship'

RETWEET_KEY = u'retweeted_status'

MENTION_ONLY_KEY = u'mention_only'


SUBTOKEN_KEY = u'oauth_token'

SUBTOKEN_SECRET_KEY = u'oauth_token_secret'


CRITERIA = {SUBTOKEN_KEY: unicode,
            SUBTOKEN_SECRET_KEY: unicode}


VERIFICATION_URL = b'https://api.twitter.com/1/account/verify_credentials.json'

GET_USER_URL = b'https://api.twitter.com/1/users/show.json'

POST_RELATIONSHIP_URL = b'https://api.twitter.com/1/friendships/update.json'

GET_TWEET_URL = b'https://api.twitter.com/1/statuses/show/{}.json'

DELETE_TWEET_URL = b'https://api.twitter.com/1/statuses/destroy/{}.json'

GET_TWEETS_HOME = b'https://api.twitter.com/1/statuses/home_timeline.json'

GET_TWEETS_MENTIONS = b'https://api.twitter.com/1/statuses/mentions.json'

GET_MESSAGES_RECEIVED = b'https://api.twitter.com/1/direct_messages.json'

GET_MESSAGES_SENT = b'https://api.twitter.com/1/direct_messages/sent.json'


CONNECTION_MESSAGE = u'Could not connect to Twitter. Check your internet connection and try again.'

PROTECTED_MESSAGE = u'Twitter refused the request. If the user is protected, check if you are following and try again.'

REFUSED_MESSAGE = u'Twitter refused the request. Please wait a few minutes and try again.'

DOWN_MESSAGE = u'Twitter seems to be down. Please wait a few minutes and try again.'

CODE_304_MESSAGE = u'Twitter did not return any data. Please try again.'

CODE_401_MESSAGE = u'Twitter did not authorize the request. Please try again.'

CODE_404_MESSAGE = u'This data seems to have been deleted.'

CODE_MESSAGE = u'Twitter responded with error {}. Please try again.'

UNEXPECTED_MESSAGE = u'Unexpected response from Twitter. Please try again.'



def filter(parameters):
    parameters = {key: value
                  for key, value in parameters.viewitems()
                  if value is not None}

    return urlencode(parameters)


def validate(data, criteria):
    for key, classinfo in criteria.viewitems():
        if not isinstance(data[key], classinfo):
            raise TypeError



class Exit(Exception):
    pass



class Error(Exception):
    def __init__(self, message):
        self.message = message



class ConnectionError(Error):
    pass



class UnauthorizedError(Error):
    pass



class ForbiddenError(Error):
    pass



class NotFoundError(Error):
    pass



class RateError(Error):
    pass



class Backend(Updatable):
    # NOTE: The presence of long will not be necessary
    # after the planned port to Python 3 is completed.


    def _process_paged_user_ids(self, data, timestamp):
        data = data[u'ids']

        for subdata in data:
            if not isinstance(subdata, (int, long)):
                raise TypeError

        return data


    def _process_user_ids(self, data, timestamp):
        for subdata in data:
            if not isinstance(subdata, (int, long)):
                raise TypeError

        return data


    def _process_users(self, data, timestamp):
        for subdata in data:
            self.process_user(subdata, timestamp)

        return data


    # NOTE: The presence of long will not be necessary
    # after the planned port to Python 3 is completed.

    def _process_user_page(self, data, timestamp):
        previous_cursor = data[PREVIOUS_CURSOR_KEY]

        if not isinstance(previous_cursor, (int, long)):
            raise TypeError

        next_cursor = data[NEXT_CURSOR_KEY]

        if not isinstance(next_cursor, (int, long)):
            raise TypeError

        for subdata in data[USERS_KEY]:
            self.process_user(subdata, timestamp)

        return (previous_cursor, next_cursor, data[USERS_KEY])


    def _process_relationship(self, data, timestamp):
        id = data[RELATIONSHIP_KEY][u'target'][ID_KEY]

        data = data[RELATIONSHIP_KEY][u'source']

        data[ID_KEY] = id

        data[REQUEST_TIMESTAMP_KEY] = timestamp

        validate(data, relationship.CRITERIA)

        return data


    def _process_tweets(self, data, timestamp):
        processed_data = []

        for subdata in data:
            processed_data.append(self.process_tweet(subdata, timestamp))

        return processed_data


    def _process_messages(self, data, timestamp):
        processed_data = []

        for subdata in data:
            processed_data.append(self.process_message(subdata, timestamp))

        return processed_data


    # NOTE: The presence of long will not be necessary
    # after the planned port to Python 3 is completed.

    def _process_group_page(self, data, timestamp):
        previous_cursor = data[PREVIOUS_CURSOR_KEY]

        if not isinstance(previous_cursor, (int, long)):
            raise TypeError

        next_cursor = data[NEXT_CURSOR_KEY]

        if not isinstance(next_cursor, (int, long)):
            raise TypeError

        processed_data = []

        for subdata in data[u'lists']:
            processed_data.append(self.process_group(subdata, timestamp))

        return (previous_cursor, next_cursor, processed_data)


    def _process_categories(self, data, timestamp):
        for subdata in data:
            validate(subdata, category.CRITERIA)

        return data


    def _process_suggestions(self, data, timestamp):
        data = data[USERS_KEY]

        for subdata in data:
            self.process_user(subdata, timestamp)

        return data


    def _process_trends(self, data, timestamp):
        data = data[0][u'trends']

        for subdata in data:
            validate(subdata, trend.CRITERIA)

        return data


    def _request(self, url, method, body, callback):
        oauth_client = oauth2.Client(self.oauth_consumer, self.oauth_token)

        # NOTE: The usage of AttributeError is actually due to
        # Issue 96 in httplib2, which is already fixed in trunk.

        try:
            response, content = oauth_client.request(url, method, body)
        except (HttpLib2Error, HTTPException, ProxyError, socket.error, AttributeError):
            raise ConnectionError(CONNECTION_MESSAGE)

        status = long(response[STATUS_KEY])

        if status != 200:
            if status == 304:
                raise Error(CODE_304_MESSAGE)
            elif status == 400:
                raise Error(REFUSED_MESSAGE)
            elif status == 401:
                raise UnauthorizedError(CODE_401_MESSAGE)
            elif status == 403:
                raise ForbiddenError(REFUSED_MESSAGE)
            elif status == 404:
                raise NotFoundError(CODE_404_MESSAGE)
            elif status == 420:
                raise RateError(u'You have exceeded the request limit for Twitter. Please wait some time before trying again.')
            elif status == 500 or status == 502 or status == 503:
                raise Error(DOWN_MESSAGE)
            else:
                raise Error(CODE_MESSAGE.format(status))

        try:
            timestamp = float(timegm(parsedate(response[u'last-modified'])))

            data = json.loads(content)

            processed_data = callback(data, timestamp)
        except (ValueError, TypeError, KeyError, AttributeError):
            raise Error(UNEXPECTED_MESSAGE)

        return processed_data


    def _request_get(self, url, parameters, callback):
        return self._request(b'{}?{}'.format(url, filter(parameters)), b'GET', b'', callback)


    def _request_post(self, url, parameters, callback):
        return self._request(url, b'POST', filter(parameters), callback)


    def _get_user_data_own(self):
        parameters = {b'include_entities': 1,
                      b'skip_status': 1}

        return self._request_get(VERIFICATION_URL, parameters, self.process_user)


    def _acquire_social(self):
        self.connecting_lock.acquire()

        self.social_count_lock.acquire()

        previous = self.social_count

        self.social_count += 1

        self.social_count_lock.release()

        if previous == 0:
            self.social_lock.acquire()

        self.connecting_lock.release()


    def _release_social(self):
        self.social_count_lock.acquire()

        self.social_count -= 1

        current = self.social_count

        self.social_count_lock.release()

        if current == 0:
            self.social_lock.release()


    def __init__(self, data, user_data, oauth_consumer):
        super(Backend, self).__init__()

        self.data = data

        self.oauth_consumer = oauth_consumer

        self.social_count = 0

        self.social_count_lock = Lock()

        self.social_lock = Lock()

        self.connecting_lock = Lock()

        self.oauth_token = oauth2.Token(self.data[SUBTOKEN_KEY].encode(ENCODING),
                                        self.data[SUBTOKEN_SECRET_KEY].encode(ENCODING))

        self.delete_user_follow_notifier = Notifier()

        self.send_tweet_notifier = Notifier()

        self.delete_tweet_notifier = Notifier()

        self.send_tweet_retweet_notifier = Notifier()

        self.send_tweet_favorite_notifier = Notifier()

        self.delete_tweet_favorite_notifier = Notifier()

        self.send_message_notifier = Notifier()

        self.delete_message_notifier = Notifier()

        self.delete_group_notifier = Notifier()

        if user_data is None:
            self.user_id = None

            user_data = self._get_user_data_own()
        else:
            validate(user_data, user.CRITERIA)

        self.user_id = user_data[ID_KEY]

        self.user_backend_cache = user.BackendCache()

        self.relationship_backend_cache = relationship.BackendCache()

        self.tweet_backend_cache = tweet.BackendCache()

        self.group_backend_cache = group.BackendCache()

        self.user_backend = self.user_backend_cache.acquire(user_data)

        self.thread = None


    def isolate(self):
        if self.thread is not None:
            self.thread.join()

        self.user_backend.release()


    def process_user(self, data, timestamp):
        data[REQUEST_TIMESTAMP_KEY] = timestamp

        if self.user_id is None:
            data[user.OWN_KEY] = True
        else:
            data[user.OWN_KEY] = (self.user_id == data[ID_KEY])

        if ENTITIES_KEY not in data.viewkeys():
            data[ENTITIES_KEY] = None

        if DISPLAY_URL_KEY not in data.viewkeys():
            data[DISPLAY_URL_KEY] = None

        if EXPANDED_URL_KEY not in data.viewkeys():
            data[EXPANDED_URL_KEY] = None

        validate(data, user.CRITERIA)

        return data


    def process_tweet(self, data, timestamp):
        real_id = data[ID_KEY]

        if RETWEET_KEY in data.viewkeys():
            creation_timestamp = float(timegm(parsedate(data[CREATION_TIMESTAMP_KEY])))

            retweeter_data = data.pop(USER_KEY)

            self.process_user(retweeter_data, creation_timestamp)

            tweet_data = data.pop(RETWEET_KEY)

            if retweeter_data[user.OWN_KEY]:
                tweet_data[tweet.RETWEET_ID_KEY] = data
        else:
            retweeter_data = None

            tweet_data = data

        creation_timestamp = float(timegm(parsedate(tweet_data[CREATION_TIMESTAMP_KEY])))

        user_data = tweet_data.pop(USER_KEY)

        self.process_user(user_data, creation_timestamp)

        tweet_data[REQUEST_TIMESTAMP_KEY] = timestamp

        tweet_data[CREATION_TIMESTAMP_KEY] = creation_timestamp

        if tweet.RETWEET_ID_KEY in tweet_data.viewkeys():
            retweet_data = tweet_data[tweet.RETWEET_ID_KEY]

            tweet_data[tweet.RETWEET_ID_KEY] = retweet_data[ID_KEY]
        else:
            tweet_data[tweet.RETWEET_ID_KEY] = None

        validate(tweet_data, tweet.CRITERIA)

        return (real_id, retweeter_data, user_data, tweet_data)


    def process_message(self, data, timestamp):
        creation_timestamp = float(timegm(parsedate(data[CREATION_TIMESTAMP_KEY])))

        recipient_data = data.pop(u'recipient')

        self.process_user(recipient_data, creation_timestamp)

        user_data = data.pop(u'sender')

        self.process_user(user_data, creation_timestamp)

        data[CREATION_TIMESTAMP_KEY] = creation_timestamp

        validate(data, message.CRITERIA)

        return (data[ID_KEY], recipient_data, user_data, data)


    def process_group(self, data, timestamp):
        creation_timestamp = data[CREATION_TIMESTAMP_KEY]

        if creation_timestamp is None:
            creation_timestamp = data[USER_KEY][CREATION_TIMESTAMP_KEY]

        creation_timestamp = float(timegm(parsedate(creation_timestamp)))

        user_data = data.pop(USER_KEY)

        self.process_user(user_data, creation_timestamp)

        data[REQUEST_TIMESTAMP_KEY] = timestamp

        validate(data, group.CRITERIA)

        return (user_data, data)


    def notify_delete_tweet(self, deleted_id):
        self._acquire_update()

        self.deleted_id = deleted_id

        self.delete_tweet_notifier.notify()

        self._release_update()


    def notify_delete_tweet_favorite(self, deleted_id):
        self._acquire_update()

        self.deleted_id = deleted_id

        self.delete_tweet_favorite_notifier.notify()

        self._release_update()


    def notify_delete_message(self, deleted_id):
        self._acquire_update()

        self.deleted_id = deleted_id

        self.delete_message_notifier.notify()

        self._release_update()


    def notify_delete_group(self, deleted_id):
        self._acquire_update()

        self.deleted_id = deleted_id

        self.delete_group_notifier.notify()

        self._release_update()


    def get_oauth_headers(self, url):
        oauth_method = oauth2.SignatureMethod_HMAC_SHA1()

        oauth_request = oauth2.Request.from_consumer_and_token(self.oauth_consumer, self.oauth_token, b'GET', url)

        oauth_request.sign_request(oauth_method, self.oauth_consumer, self.oauth_token)

        oauth_headers = [b'{}: {}'.format(key, value) for key, value in oauth_request.to_header().viewitems()]

        oauth_headers.append(b'Expect:')

        return oauth_headers


    def get_oauth_headers_echo(self):
        oauth_method = oauth2.SignatureMethod_HMAC_SHA1()

        oauth_request = oauth2.Request.from_consumer_and_token(self.oauth_consumer, self.oauth_token, b'GET', VERIFICATION_URL)

        oauth_request.sign_request(oauth_method, self.oauth_consumer, self.oauth_token)

        oauth_headers = oauth_request.to_header(b'https://api.twitter.com/')

        oauth_echo_header = oauth_headers[u'Authorization'].decode(ENCODING)

        return [b'X-Auth-Service-Provider: {}'.format(VERIFICATION_URL),
                b'X-Verify-Credentials-Authorization: {}'.format(oauth_echo_header),
                b'Expect:']


    def get_user_data(self, user_id):
        parameters = {b'user_id': user_id,
                      b'include_entities': 1,
                      b'skip_status': 1}

        return self._request_get(GET_USER_URL, parameters, self.process_user)


    def get_user_ids_followers(self, user_id):
        parameters = {b'user_id': user_id,
                      b'cursor': -1}

        return self._request_get(b'https://api.twitter.com/1/followers/ids.json', parameters, self._process_paged_user_ids)


    def get_user_ids_friends(self, user_id):
        parameters = {b'user_id': user_id,
                      b'cursor': -1}

        return self._request_get(b'https://api.twitter.com/1/friends/ids.json', parameters, self._process_paged_user_ids)


    def get_user_ids_blocking(self):
        parameters = {}

        return self._request_get(b'https://api.twitter.com/1/blocks/blocking/ids.json', parameters, self._process_user_ids)


    def get_user_ids_retweeters(self, tweet_id):
        parameters = {b'count': 100}

        return self._request_get(b'https://api.twitter.com/1/statuses/{}/retweeted_by/ids.json'.format(tweet_id), parameters, self._process_user_ids)


    def get_user_dataset(self, user_ids):
        bytes_user_ids = []

        for user_id in user_ids:
            bytes_user_ids.append(bytes(user_id))

        parameters = {b'user_id': b','.join(bytes_user_ids),
                      b'include_entities': 1,
                      b'skip_status': 1}

        return self._request_get(b'https://api.twitter.com/1/users/lookup.json', parameters, self._process_users)


    def get_user_dataset_members(self, list_id, cursor):
        parameters = {b'list_id': list_id,
                      b'cursor': cursor,
                      b'include_entities': 1,
                      b'skip_status': 1}

        return self._request_get(b'https://api.twitter.com/1/lists/members.json', parameters, self._process_user_page)


    def get_user_dataset_subscribers(self, list_id, cursor):
        parameters = {b'list_id': list_id,
                      b'cursor': cursor,
                      b'include_entities': 1,
                      b'skip_status': 1}

        return self._request_get(b'https://api.twitter.com/1/lists/subscribers.json', parameters, self._process_user_page)


    def get_relationship_data(self, target_id):
        parameters = {b'target_id': target_id}

        return self._request_get(b'https://api.twitter.com/1/friendships/show.json', parameters, self._process_relationship)


    def send_relationship_retweets(self, user_id):
        parameters = {b'user_id': user_id,
                      b'retweets': b'true'}

        return self._request_post(POST_RELATIONSHIP_URL, parameters, self._process_relationship)


    def delete_relationship_retweets(self, user_id):
        parameters = {b'user_id': user_id,
                      b'retweets': b'false'}

        return self._request_post(POST_RELATIONSHIP_URL, parameters, self._process_relationship)


    def send_relationship_notifications(self, user_id):
        parameters = {b'user_id': user_id,
                      b'device': b'true'}

        return self._request_post(POST_RELATIONSHIP_URL, parameters, self._process_relationship)


    def delete_relationship_notifications(self, user_id):
        parameters = {b'user_id': user_id,
                      b'device': b'false'}

        return self._request_post(POST_RELATIONSHIP_URL, parameters, self._process_relationship)


    def send_user_report(self, user_id):
        parameters = {b'user_id': user_id,
                      b'include_entities': 1,
                      b'skip_status': 1}

        self._acquire_social()

        try:
            data = self._request_post(b'https://api.twitter.com/1/report_spam.json', parameters, self.process_user)
        finally:
            self._release_social()

        # NOTE: The data lacks pending info if the user is protected,
        # but considering the circunstances its value can be deduced.

        data[user.PENDING_KEY] = False

        return data


    def send_user_block(self, user_id):
        parameters = {b'user_id': user_id,
                      b'include_entities': 1,
                      b'skip_status': 1}

        self._acquire_social()

        try:
            data = self._request_post(b'https://api.twitter.com/1/blocks/create.json', parameters, self.process_user)
        finally:
            self._release_social()

        # NOTE: The data lacks pending info if the user is protected,
        # but considering the circunstances its value can be deduced.

        data[user.PENDING_KEY] = False

        return data


    def delete_user_block(self, user_id):
        parameters = {b'user_id': user_id,
                      b'include_entities': 1,
                      b'skip_status': 1}

        self._acquire_social()

        try:
            data = self._request_post(b'https://api.twitter.com/1/blocks/destroy.json', parameters, self.process_user)
        finally:
            self._release_social()

        # NOTE: The data lacks pending info if the user is protected,
        # but considering the circunstances its value can be deduced.

        data[user.PENDING_KEY] = False

        return data


    def send_user_follow(self, user_id):
        parameters = {b'user_id': user_id,
                      b'include_entities': 1,
                      b'skip_status': 1}

        self._acquire_social()

        try:
            data = self._request_post(b'https://api.twitter.com/1/friendships/create.json', parameters, self.process_user)
        finally:
            self._release_social()

        # NOTE: The data lacks pending info if the user is protected,
        # but considering the circunstances its value can be deduced.

        if data[user.PROTECTED_KEY]:
            data[user.PENDING_KEY] = True
        else:
            data[user.PENDING_KEY] = False

        return data


    def delete_user_follow(self, user_id):
        parameters = {b'user_id': user_id,
                      b'include_entities': 1,
                      b'skip_status': 1}

        self._acquire_social()

        try:
            self.unfollowed_data = self._request_post(b'https://api.twitter.com/1/friendships/destroy.json', parameters, self.process_user)
        finally:
            self._release_social()

        # NOTE: The data lacks pending info if the user is protected,
        # but considering the circunstances its value can be deduced.

        self.unfollowed_data[user.PENDING_KEY] = False

        self.delete_user_follow_notifier.notify()

        return self.unfollowed_data


    def send_tweet(self, status, in_reply_to_status_id):
        parameters = {b'status': status.encode(ENCODING),
                      b'in_reply_to_status_id': in_reply_to_status_id,
                      b'include_entities': 1}

        try:
            data = self._request_post(b'https://api.twitter.com/1/statuses/update.json', parameters, self.process_tweet)
        except NotFoundError:
            data = None

        self._acquire_update()

        self.send_tweet_notifier.notify()

        self._release_update()

        return data


    def get_tweet_data(self, id):
        parameters = {b'include_entities': 1,
                      b'include_my_retweet': 1}

        return self._request_get(GET_TWEET_URL.format(id), parameters, self.process_tweet)


    def get_tweet_dataset_home_mentions(self, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1,
                      b'include_my_retweet': 1}

        dataset = self._request_get(GET_TWEETS_HOME, parameters, self._process_tweets)

        dataset_home = []

        for data in dataset:
            if data[1] is None:
                dataset_home.append(data)
            else:
                if data[1][user.OWN_KEY]:
                    user_ids = []

                    try:
                        for mention in data[3][ENTITIES_KEY][MENTIONS_KEY]:
                            try:
                                user_ids.append(mention[ID_KEY])
                            except (TypeError, KeyError):
                                pass
                    except (TypeError, KeyError):
                        pass

                    if self.user_id not in user_ids:
                        dataset_home.append(data)
                else:
                    dataset_home.append(data)

        dataset_mentions = self._request_get(GET_TWEETS_MENTIONS, parameters, self._process_tweets)

        if not dataset_home:
            for data_mentions in dataset_mentions:
                data_mentions[3][MENTION_ONLY_KEY] = True

            return dataset_mentions

        if not dataset_mentions:
            for data_home in dataset_home:
                data_home[3][MENTION_ONLY_KEY] = False

            return dataset_home

        length_home = len(dataset_home)

        length_mentions = len(dataset_mentions)

        index_home = 0

        index_mentions = 0

        dataset = []

        while True:
            data_home = dataset_home[index_home]

            data_mentions = dataset_mentions[index_mentions]

            if data_home[0] > data_mentions[0]:
                data_home[3][MENTION_ONLY_KEY] = False

                dataset.append(data_home)

                index_home += 1
            else:
                if data_home[0] == data_mentions[0]:
                    index_home += 1

                    data_mentions[3][MENTION_ONLY_KEY] = False
                else:
                    data_mentions[3][MENTION_ONLY_KEY] = True

                dataset.append(data_mentions)

                index_mentions += 1

            if index_home == length_home:
                if since_id is not None:
                    if dataset_mentions[-1][0] == since_id + 1:
                        for data_mentions in dataset_mentions[index_mentions:]:
                            data_mentions[3][MENTION_ONLY_KEY] = True

                        dataset.extend(dataset_mentions[index_mentions:])

                break

            if index_mentions == length_mentions:
                if since_id is not None:
                    if dataset_home[-1][0] == since_id + 1:
                        for data_home in dataset_home[index_home:]:
                            data_home[3][MENTION_ONLY_KEY] = False

                        dataset.extend(dataset_home[index_home:])

                break

        return dataset[:count]


    def get_tweet_dataset_home(self, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1,
                      b'include_my_retweet': 1}

        return self._request_get(GET_TWEETS_HOME, parameters, self._process_tweets)


    def get_tweet_dataset_mentions(self, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1,
                      b'include_my_retweet': 1}

        return self._request_get(GET_TWEETS_MENTIONS, parameters, self._process_tweets)


    def get_tweet_dataset_sent(self, user_id, since_id, max_id, count, include_rts):
        parameters = {b'user_id': user_id,
                      b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_rts': include_rts,
                      b'include_entities': 1,
                      b'include_my_retweet': 1}

        return self._request_get(b'https://api.twitter.com/1/statuses/user_timeline.json', parameters, self._process_tweets)


    def get_tweet_dataset_favorited(self, user_id, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1,
                      b'include_my_retweet': 1}

        return self._request_get(b'https://api.twitter.com/1/favorites/{}.json'.format(user_id), parameters, self._process_tweets)


    def get_tweet_dataset_retweeted_by_me(self, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1,
                      b'include_my_retweet': 1}

        return self._request_get(b'https://api.twitter.com/1/statuses/retweeted_by_me.json', parameters, self._process_tweets)


    def get_tweet_dataset_retweeted_to_me(self, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1,
                      b'include_my_retweet': 1}

        return self._request_get(b'https://api.twitter.com/1/statuses/retweeted_to_me.json', parameters, self._process_tweets)


    def get_tweet_dataset_retweeted_from_me(self, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1,
                      b'include_my_retweet': 1}

        return self._request_get(b'https://api.twitter.com/1/statuses/retweets_of_me.json', parameters, self._process_tweets)


    def get_tweet_dataset_group(self, list_id, since_id, max_id, count, include_rts):
        parameters = {b'list_id': list_id,
                      b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_rts': include_rts,
                      b'include_entities': 1,
                      b'include_my_retweet': 1}

        return self._request_get(b'https://api.twitter.com/1/lists/statuses.json', parameters, self._process_tweets)


    def delete_tweet(self, id):
        parameters = {b'include_entities': 1}

        try:
            data = self._request_post(DELETE_TWEET_URL.format(id), parameters, self.process_tweet)
        except NotFoundError:
            data = None

        self.notify_delete_tweet(id)

        return data


    def send_tweet_retweet(self, tweet_id):
        parameters = {b'include_entities': 1}

        data = self._request_post(b'https://api.twitter.com/1/statuses/retweet/{}.json'.format(tweet_id), parameters, self.process_tweet)

        self._acquire_update()

        self.send_tweet_retweet_notifier.notify()

        self._release_update()

        return data


    def delete_tweet_retweet(self, retweet_id, tweet_id):
        parameters = {b'include_entities': 1}

        data = self._request_post(DELETE_TWEET_URL.format(retweet_id), parameters, self.process_tweet)

        self.notify_delete_tweet(retweet_id)

        return data


    def send_tweet_favorite(self, tweet_id):
        parameters = {b'include_entities': 1}

        data = self._request_post(b'https://api.twitter.com/1/favorites/create/{}.json'.format(tweet_id), parameters, self.process_tweet)

        self._acquire_update()

        self.favorited_data = data

        self.send_tweet_favorite_notifier.notify()

        self._release_update()

        return data


    def delete_tweet_favorite(self, tweet_id):
        parameters = {b'include_entities': 1}

        data = self._request_post(b'https://api.twitter.com/1/favorites/destroy/{}.json'.format(tweet_id), parameters, self.process_tweet)

        self.notify_delete_tweet_favorite(tweet_id)

        return data


    def send_message(self, user_id, text):
        parameters = {b'user_id': user_id,
                      b'text': text.encode(ENCODING),
                      b'include_entities': 1}

        try:
            data = self._request_post(b'https://api.twitter.com/1/direct_messages/new.json', parameters, self.process_message)
        except NotFoundError:
            data = None

        self._acquire_update()

        self.send_message_notifier.notify()

        self._release_update()

        return data


    def get_message_dataset_all(self, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1}

        dataset_received = self._request_get(GET_MESSAGES_RECEIVED, parameters, self._process_messages)

        dataset_sent = self._request_get(GET_MESSAGES_SENT, parameters, self._process_messages)

        if not dataset_received:
            return dataset_sent

        if not dataset_sent:
            return dataset_received

        length_received = len(dataset_received)

        length_sent = len(dataset_sent)

        index_received = 0

        index_sent = 0

        dataset = []

        while True:
            data_received = dataset_received[index_received]

            data_sent = dataset_sent[index_sent]

            if data_received[0] > data_sent[0]:
                dataset.append(data_received)

                index_received += 1
            else:
                if data_received[0] == data_sent[0]:
                    index_received += 1

                dataset.append(data_sent)

                index_sent += 1

            if index_received == length_received:
                if since_id is not None:
                    if dataset_sent[-1][0] == since_id + 1:
                        dataset.extend(dataset_sent[index_sent:])

                break

            if index_sent == length_sent:
                if since_id is not None:
                    if dataset_received[-1][0] == since_id + 1:
                        dataset.extend(dataset_received[index_received:])

                break

        return dataset[:count]


    def get_message_dataset_received(self, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1}

        return self._request_get(GET_MESSAGES_RECEIVED, parameters, self._process_messages)


    def get_message_dataset_sent(self, since_id, max_id, count):
        parameters = {b'since_id': since_id,
                      b'max_id': max_id,
                      b'count': count,
                      b'include_entities': 1}

        return self._request_get(GET_MESSAGES_SENT, parameters, self._process_messages)


    def delete_message(self, id):
        parameters = {b'include_entities': 1}

        try:
            data = self._request_post(b'https://api.twitter.com/1/direct_messages/destroy/{}.json'.format(id), parameters, self.process_message)
        except NotFoundError:
            data = None

        self.notify_delete_message(id)

        return data


    def get_group_data(self, list_id):
        parameters = {b'list_id': list_id}

        return self._request_get(b'https://api.twitter.com/1/lists/show.json', parameters, self.process_group)


    def get_group_dataset_memberships(self, user_id, cursor, filter_to_owned_lists):
        parameters = {b'user_id': user_id,
                      b'cursor': cursor,
                      b'filter_to_owned_lists': filter_to_owned_lists}

        return self._request_get(b'https://api.twitter.com/1/lists/memberships.json', parameters, self._process_group_page)


    def get_group_dataset_subscriptions(self, user_id, cursor):
        parameters = {b'user_id': user_id,
                      b'cursor': cursor}

        return self._request_get(b'https://api.twitter.com/1/lists/subscriptions.json', parameters, self._process_group_page)


    def get_group_dataset_groups(self, user_id, cursor):
        parameters = {b'user_id': user_id,
                      b'cursor': cursor}

        return self._request_get(b'https://api.twitter.com/1/lists.json', parameters, self._process_group_page)


    def send_group_create(self, name, description, mode):
        parameters = {b'name': name.encode(ENCODING),
                      b'description': description.encode(ENCODING),
                      b'mode': mode.encode(ENCODING)}

        return self._request_post(b'https://api.twitter.com/1/lists/create.json', parameters, self.process_group)


    def send_group_update(self, list_id, name, description, mode):
        parameters = {b'list_id': list_id,
                      b'name': name.encode(ENCODING),
                      b'description': description.encode(ENCODING),
                      b'mode': mode.encode(ENCODING)}

        return self._request_post(b'https://api.twitter.com/1/lists/update.json', parameters, self.process_group)


    def delete_group(self, list_id):
        parameters = {b'list_id': list_id}

        try:
            data = self._request_post(b'https://api.twitter.com/1/lists/destroy.json', parameters, self.process_group)
        except NotFoundError:
            data = None

        self.notify_delete_group(list_id)

        return data


    def send_group_membership(self, list_id, user_id):
        parameters = {b'list_id': list_id,
                      b'user_id': user_id}

        return self._request_post(b'https://api.twitter.com/1/lists/members/create.json', parameters, self.process_group)


    def delete_group_membership(self, list_id, user_id):
        parameters = {b'list_id': list_id,
                      b'user_id': user_id}

        return self._request_post(b'https://api.twitter.com/1/lists/members/destroy.json', parameters, self.process_group)


    def send_group_subscription(self, list_id):
        parameters = {b'list_id': list_id}

        return self._request_post(b'https://api.twitter.com/1/lists/subscribers/create.json', parameters, self.process_group)


    def delete_group_subscription(self, list_id):
        parameters = {b'list_id': list_id}

        return self._request_post(b'https://api.twitter.com/1/lists/subscribers/destroy.json', parameters, self.process_group)


    def get_category_dataset(self):
        parameters = {}

        return self._request_get(b'https://api.twitter.com/1/users/suggestions.json', parameters, self._process_categories)


    def get_suggestion_dataset(self, slug):
        parameters = {b'include_entities': 1,
                      b'skip_status': 1}

        return self._request_get(b'https://api.twitter.com/1/users/suggestions/{}.json'.format(slug), parameters, self._process_suggestions)


    def get_trend_dataset(self):
        parameters = {}

        return self._request_get(b'https://api.twitter.com/1/trends/1.json', parameters, self._process_trends)


    def acquire_connecting(self):
        self.connecting_lock.acquire()

        self.social_lock.acquire()
 
        self.social_lock.release()


    def release_connecting(self):
        self.connecting_lock.release()



class ManagerBackend(object):
    def _uncurtain(self, codepoints):
        chars = []

        for codepoint in reversed(codepoints):
            chars.append(chr(255 - codepoint))

        return b''.join(chars)


    def _process_lengths(self, data):
        return [data[u'short_url_length'], data[u'short_url_length_https']]


    def _process_results(self, data):
        processed_data = []

        for subdata in data[u'results']:
            creation_timestamp = float(timegm(parsedate(subdata[CREATION_TIMESTAMP_KEY])))

            subdata[CREATION_TIMESTAMP_KEY] = creation_timestamp

            validate(subdata, result.CRITERIA)

            processed_data.append((subdata[ID_KEY], subdata))

        return processed_data


    def _http_request(self, url, parameters, callback):
        http_client = httplib2.Http()

        # NOTE: The usage of AttributeError is actually due to
        # Issue 96 in httplib2, which is already fixed in trunk.

        try:
            response, content = http_client.request(b'{}?{}'.format(url, filter(parameters)), b'GET')
        except (HttpLib2Error, HTTPException, ProxyError, socket.error, AttributeError):
            raise ConnectionError(CONNECTION_MESSAGE)

        status = long(response[STATUS_KEY])

        if status != 200:
            if status == 304:
                raise Error(CODE_304_MESSAGE)
            elif status == 400 or status == 403:
                raise ForbiddenError(REFUSED_MESSAGE)
            elif status == 404:
                raise NotFoundError(CODE_404_MESSAGE)
            elif status == 406:
                raise UnauthorizedError(u'Twitter refused the search. The format seems to be invalid.')
            elif status == 500 or status == 502 or status == 503:
                raise Error(DOWN_MESSAGE)
            else:
                raise Error(CODE_MESSAGE.format(status))

        try:
            data = json.loads(content)

            processed_data = callback(data)
        except (ValueError, TypeError, KeyError):
            raise Error(UNEXPECTED_MESSAGE)

        return processed_data


    def _oauth_request(self, oauth_client, url):
        # NOTE: The usage of AttributeError is actually due to
        # Issue 96 in httplib2, which is already fixed in trunk.

        try:
            response, content = oauth_client.request(url, b'GET')
        except (HttpLib2Error, HTTPException, ProxyError, socket.error, AttributeError):
            raise ConnectionError(CONNECTION_MESSAGE)

        status = long(response[STATUS_KEY])

        if status != 200:
            if status == 304:
                raise Error(CODE_304_MESSAGE)
            elif status == 400:
                raise Error(REFUSED_MESSAGE)
            elif status == 401:
                raise UnauthorizedError(CODE_401_MESSAGE)
            elif status == 403:
                raise ForbiddenError(REFUSED_MESSAGE)
            elif status == 404:
                raise NotFoundError(CODE_404_MESSAGE)
            elif status == 500 or status == 502 or status == 503:
                raise Error(DOWN_MESSAGE)
            else:
                raise Error(CODE_MESSAGE.format(status))

        try:
            data = parse_qs(content, True, True)

            processed_data = {SUBTOKEN_KEY: data[SUBTOKEN_KEY][0].decode(ENCODING),
                              SUBTOKEN_SECRET_KEY: data[SUBTOKEN_SECRET_KEY][0].decode(ENCODING)}
        except (ValueError, TypeError, KeyError, AttributeError):
            raise Error(UNEXPECTED_MESSAGE)

        return processed_data


    def __init__(self):
        self.oauth_consumer = oauth2.Consumer(self._uncurtain([190, 137, 178, 199, 176, 167, 148, 190, 155, 175, 173, 206, 202, 146, 168, 135, 141, 145, 170, 173, 153, 198]),
                                              self._uncurtain([207, 181, 188, 145, 133, 202, 188, 171, 172, 175, 135, 149, 170, 144, 165, 141, 183, 158, 204, 200, 153, 183, 168, 207, 135, 179, 170, 165, 139, 168, 179, 148, 207, 204, 207, 138, 187, 203, 133, 174, 147, 135]))

 
    def get_lengths(self):
        parameters = {}

        return self._http_request(b'https://api.twitter.com/1/help/configuration.json', parameters, self._process_lengths)


    def get_result_dataset(self, q, since_id, max_id, rpp, result_type):
        parameters = {b'q': q.encode(ENCODING),
                      b'since_id': since_id,
                      b'max_id': max_id,
                      b'rpp': rpp,
                      b'result_type': result_type.encode(ENCODING)}

        return self._http_request(b'https://search.twitter.com/search.json', parameters, self._process_results)


    def get_request_data(self):
        oauth_client = oauth2.Client(self.oauth_consumer)

        return self._oauth_request(oauth_client, b'https://twitter.com/oauth/request_token')


    def get_authorize_url(self, data):
        return b'https://twitter.com/oauth/authorize?oauth_token={}'.format(data[SUBTOKEN_KEY].encode(ENCODING))


    def get_access_data(self, data, pin):
        oauth_token = oauth2.Token(data[SUBTOKEN_KEY].encode(ENCODING),
                                   data[SUBTOKEN_SECRET_KEY].encode(ENCODING))

        oauth_token.set_verifier(pin)

        oauth_client = oauth2.Client(self.oauth_consumer, oauth_token)

        return self._oauth_request(oauth_client, b'https://twitter.com/oauth/access_token')


    def build(self, data, user_data):
        validate(data, CRITERIA)

        return Backend(data, user_data, self.oauth_consumer)
