# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



from polly import Notifier

from polly.twitter import ID_KEY, REQUEST_TIMESTAMP_KEY, ItemBackend, BaseBackendCache



SPAM_KEY = u'marked_spam'

BLOCKING_KEY = u'blocking'

FRIEND_KEY = u'following'

FOLLOWER_KEY = u'followed_by'

RETWEETS_KEY = u'want_retweets'

NOTIFICATIONS_KEY = u'notifications_enabled'


# NOTE: The presence of long will not be necessary
# after the planned port to Python 3 is completed.

CRITERIA = {ID_KEY: (int, long),
            REQUEST_TIMESTAMP_KEY: float,
            SPAM_KEY: (type(None), bool),
            BLOCKING_KEY: (type(None), bool),
            FRIEND_KEY: (type(None), bool),
            FOLLOWER_KEY: (type(None), bool),
            RETWEETS_KEY: (type(None), bool),
            NOTIFICATIONS_KEY: (type(None), bool)}



class Backend(ItemBackend):
    def _notify(self):
        self.data[SPAM_KEY] = self.previous_data[SPAM_KEY]

        self.data[BLOCKING_KEY] = self.previous_data[BLOCKING_KEY]

        self.data[FRIEND_KEY] = self.previous_data[FRIEND_KEY]

        if self.previous_data[FOLLOWER_KEY] != self.data[FOLLOWER_KEY]:
            self.follower_notifier.notify()

        self.data[RETWEETS_KEY] = self.previous_data[RETWEETS_KEY]

        self.data[NOTIFICATIONS_KEY] = self.previous_data[NOTIFICATIONS_KEY]


    def __init__(self, id, data, friendship_backend_cache):
        super(Backend, self).__init__(id, data, friendship_backend_cache)

        self.social_notifier = Notifier()

        self.follower_notifier = Notifier()

        self.notifiers[RETWEETS_KEY] = Notifier()

        self.notifiers[NOTIFICATIONS_KEY] = Notifier()


    def refresh(self):
        self._acquire_update()

        self.social_notifier.notify()

        self._release_update()


    def set_blocking(self, spam):
        self._acquire_update()

        if spam:
            self.data[SPAM_KEY] = True

        self.data[BLOCKING_KEY] = True

        self.data[FRIEND_KEY] = False

        self.notifier.notify()

        self.social_notifier.notify()

        self._release_update()


    def unset_blocking(self):
        self._acquire_update()

        if self.data[BLOCKING_KEY] != False:
            self.data[BLOCKING_KEY] = False

            self.notifier.notify()

            self.social_notifier.notify()

        self._release_update()


    def update_friend(self, friend):
        self._acquire_update()

        if self.data[FRIEND_KEY] != friend:
            self.data[FRIEND_KEY] = friend

            self.notifier.notify()

            self.social_notifier.notify()

        self._release_update()



class BackendCache(BaseBackendCache):
    def __init__(self):
        super(BackendCache, self).__init__(Backend)


    def unset_blocking(self, id):
        self.lock.acquire()

        if id in self.cache.viewkeys():
            self.cache[id].unset_blocking()

        self.lock.release()


    def set_friend(self, id):
        self.lock.acquire()

        if id in self.cache.viewkeys():
            self.cache[id].update_friend(True)

        self.lock.release()
