# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import json, socket

from collections import OrderedDict

from httplib import HTTPException


import httplib2, pycurl

from httplib2 import HttpLib2Error

from socks import ProxyError


from polly import STATUS_KEY, ENCODING, split

from polly.twitter import PROXIES

from polly.twitter.account import Exit, Error



DEFAULT_NAME = u'img.ly'


FILE_MESSAGE = u'The {} provider did not accept the file. Check if the media is valid and try again.'



class Provider(object):
    def _append_to_header(self, line):
        self.header_lines.append(line)

        return len(line)


    def _append_to_body(self, slice):
        self.body_slices.append(slice)

        return len(slice)


    def _send(self, post_url, account, parameters, text_spec, path_spec, callback):
        curl = pycurl.Curl()

        curl.setopt(curl.URL, post_url)

        curl.setopt(curl.HTTPHEADER, account.backend.get_oauth_headers_echo())

        parameters.append((text_spec[0], text_spec[1].encode(ENCODING)))

        parameters.append((path_spec[0], (pycurl.FORM_FILE, path_spec[1].encode(ENCODING))))

        curl.setopt(curl.HTTPPOST, parameters)

        enabled, host, port, type = self.proxy_controller.backend.get_values()

        if enabled:
            curl.setopt(pycurl.PROXY, host.encode(ENCODING))

            curl.setopt(pycurl.PROXYPORT, port)

            curl.setopt(pycurl.PROXYTYPE, PROXIES[type])

        curl.setopt(curl.CONNECTTIMEOUT, 5)

        curl.setopt(curl.NOSIGNAL, 1)

        curl.setopt(curl.NOPROGRESS, 0)

        curl.setopt(curl.PROGRESSFUNCTION, callback)

        curl.setopt(curl.HEADERFUNCTION, self._append_to_header)

        curl.setopt(curl.WRITEFUNCTION, self._append_to_body)

        self.header_lines = []

        self.body_slices = []

        try:
            curl.perform()
        except pycurl.error as error:
            if error[0] == pycurl.E_ABORTED_BY_CALLBACK:
                raise Exit
            else:
                raise Error(u'Could not upload the file to the {} provider. Check your internet connection and try again.'.format(self.name))
        finally:
            try:
                curl.close()
            except pycurl.error:
                pass

        for line in self.header_lines:
            if line.startswith(b'HTTP'):
                words = split(line.decode(ENCODING), self.pattern)

                status = long(words[1])

                if status != 100:
                    if status != 200:
                        if status in self.messages.viewkeys():
                            raise Error(self.messages[status])
                        else:
                            raise Error(u'The {0} provider responded to the upload with error {1}. Please try again.'.format(self.name, status))

                    break

        content = b''.join(self.body_slices).decode(ENCODING)

        try:
            data = json.loads(content)

            processed_data = self._process(data)
        except (ValueError, TypeError, KeyError):
            raise Error(u'Unexpected response to the upload from the {} provider. Please try again.'.format(self.name))

        return processed_data


    def __init__(self, name, proxy_controller):
        self.name = name

        self.proxy_controller = proxy_controller

        self.pattern = u'[\s]+'



class IMGLYProvider(Provider):
    def _process(self, data):
        return data[u'url']


    def send(self, account, text, path, callback):
        parameters = []

        text_spec = (b'message', text)

        path_spec = (b'media', path)

        return self._send(b'http://img.ly/api/2/upload.json', account, parameters, text_spec, path_spec, callback)


    def __init__(self, proxy_controller):
        super(IMGLYProvider, self).__init__(u'img.ly', proxy_controller)

        self.messages = {400: FILE_MESSAGE.format(self.name)}



class TWITPICProvider(Provider):
    def _process(self, data):
        id = data[u'id']

        if id == u'0' or id == self.last_id:
            raise Error(FILE_MESSAGE.format(self.name))

        return data[u'url']


    def send(self, account, text, path, callback):
        # NOTE: Twitpic can reject a file and give 200 anyway. To work
        # around this, we first make a request for the user information,
        # which either gives the id of the last accepted file or a 404
        # error. The id of a rejected file is either 0 or this last one.

        http_client = httplib2.Http()

        user_nick = account.get_user_nick().encode(ENCODING)

        url = b'http://api.twitpic.com/2/users/show.json?username={}'.format(user_nick)

        # NOTE: The usage of AttributeError is actually due to
        # Issue 96 in httplib2, which is already fixed in trunk.

        try:
            response, content = http_client.request(url)
        except (HttpLib2Error, HTTPException, ProxyError, socket.error, AttributeError):
            raise Error(u'Could not connect to the {} provider. Check your internet connection and try again.'.format(self.name))

        status = long(response[STATUS_KEY])

        if status == 200:
            try:
                data = json.loads(content)

                self.last_id = data[u'images'][0][u'short_id']
            except (ValueError, TypeError, KeyError):
                raise Error(u'Unexpected response from the {} provider. Please try again.'.format(self.name))
        else:
            if status == 404:
                self.last_id = None
            else:
                raise Error(u'The {0} provider responded with error {1}. Please try again.'.format(self.name, status))

        parameters = [(b'key', b'7a78abba0755f4913dec28d8c3ab3b62')]

        text_spec = (b'message', text)

        path_spec = (b'media', path)

        return self._send(b'http://api.twitpic.com/2/upload.json', account, parameters, text_spec, path_spec, callback)


    def __init__(self, proxy_controller):
        super(TWITPICProvider, self).__init__(u'TwitPic', proxy_controller)

        self.messages = {400: FILE_MESSAGE.format(self.name)}



class TWITVIDProvider(Provider):
    def _process(self, data):
        return data[u'rsp'][u'media_url']


    def __init__(self, proxy_controller):
        super(TWITVIDProvider, self).__init__(u'TwitVid', proxy_controller)

        self.messages = {}


    def send(self, account, text, path, callback):
        parameters = [(b'title', b'Sent via Polly'),
                      (b'format', b'json')]

        text_spec = (b'description', text)

        path_spec = (b'media', path)

        return self._send(b'https://im.twitvid.com/api/upload', account, parameters, text_spec, path_spec, callback)



class RunnerBackend(object):
    def _update_name(self):
        name = self.name_setting.get_value()

        if name not in self.providers.viewkeys():
            name = DEFAULT_NAME

        if self.name == name:
            return False

        self.name = name

        return True


    def _on_name_changed(self):
        if self._update_name():
            self.name_callback()


    def __init__(self, name_callback, setting_factory, proxy_controller):
        self.name_callback = name_callback

        self.name = DEFAULT_NAME

        self.name_setting = setting_factory.build_string(u'media', u'name', self._on_name_changed)

        self.providers = OrderedDict()

        for Class in [IMGLYProvider, TWITPICProvider, TWITVIDProvider]:
            provider = Class(proxy_controller)

            self.providers[provider.name] = provider

        self._update_name()


    def isolate(self):
        self.name_setting.isolate()

        self.name_callback = None


    def set_name(self, name):
        if self.name != name:
            self.name_setting.set_value(name)
