# #########################################################################
# Copyright (C) 2011 Marcelo Hashimoto
#
# This file is part of Polly.
#
# Polly is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polly is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polly. If not, see <http://www.gnu.org/licenses/>.
# #########################################################################



import json

from collections import deque

from random import randint

from threading import Condition

from time import time


import pycurl


from polly import ENCODING, split, dispatch, Updatable, Notifier

from polly.twitter import ID_KEY, REQUEST_TIMESTAMP_KEY, ENTITIES_KEY, MENTIONS_KEY, PROXIES

from polly.twitter.lead import user, group

from polly.twitter.post import tweet

from polly.twitter.account import Error, ConnectionError, RateError



EVENT_KEY = u'event'

SOURCE_USER_KEY = u'source'

TARGET_USER_KEY = u'target'

TARGET_ITEM_KEY = u'target_object'


DELETE_KEY = u'delete'

TWEET_KEY = u'status'

MESSAGE_KEY = u'direct_message'


UNFOLLOW_EVENT = u'unfollow'


STREAMING_URL = b'https://userstream.twitter.com/2/user.json?delimited=length'



class Producer(Updatable):
    def _produce(self, content):
        # NOTE: The first JSON data has an extra newline character
        # in the end, but the documentation says nothing about it.

        if self.streaming_starting:
            content = content[:-3].decode(ENCODING)
        else:
            content = content[:-2].decode(ENCODING)

        try:
            data = json.loads(content)
        except ValueError:
            return False

        if self.streaming_starting:
            try:
                data = data.pop(u'friends')

                for subdata in data:
                    # NOTE: The presence of long will not be necessary
                    # after the planned port to Python 3 is completed.

                    if not isinstance(subdata, (int, long)):
                        raise TypeError
            except (AttributeError, KeyError, TypeError):
                return False

            self.streaming_starting = False

        self.buffer_condition.acquire()

        self.queue.append((data, self.timestamp))

        self.buffer_condition.notify()

        self.buffer_condition.release()

        return True


    def _on_delete_follow(self):
        self.buffer_condition.acquire()

        if self.open:
            timestamp = self.account.backend.unfollowed_data[REQUEST_TIMESTAMP_KEY]

            data = {EVENT_KEY: UNFOLLOW_EVENT,
                    TARGET_USER_KEY: self.account.backend.unfollowed_data}

            self.queue.append((data, timestamp))

            self.buffer_condition.notify()

        self.buffer_condition.release()


    def _callback(self, download_total, download_current, upload_total, upload_current):
        self.dead = (time() - self.timestamp > 35.0)

        if self.dead:
            return True
        else:
            self.streaming_condition.acquire()

            streaming_finished = self.streaming_finished

            self.streaming_condition.release()

            return streaming_finished


    def _process_header(self, line):
        self.timestamp = time()

        if line == b'\r\n':
            if self.status == 200:
                self.success = 1

                self.idle = True

                self.streaming_starting = True

                self.buffer_finished = False

                self.buffer_thread = dispatch(self._buffer_loop)
        else:
            if line.startswith(b'HTTP'):
                words = split(line.decode(ENCODING), self.pattern)

                self.status = long(words[1])

        return len(line)


    def _process_body(self, slice):
        self.timestamp = time()

        if self.status == 200:
            if self.idle:
                if slice != b'\r\n':
                    index = slice.find(b'\r\n')

                    self.remaining = long(slice[:index])

                    content = slice[(index + 2):]

                    if len(content) == self.remaining:
                        if not self._produce(content):
                            return 0
                    else:
                        self.idle = False

                        self.slices = [content]

                        self.remaining -= len(content)
            else:
                self.slices.append(slice)

                if len(slice) == self.remaining:
                    content = b''.join(self.slices)

                    del self.slices[:]

                    if self._produce(content):
                        self.idle = True
                    else:
                        return 0
                else:
                    self.remaining -= len(slice)

        return len(slice)


    def _streaming_loop(self):
        self.account.identifier.set_spinning(True)

        while True:
            curl = pycurl.Curl()

            curl.setopt(curl.URL, STREAMING_URL)

            curl.setopt(curl.HTTPHEADER, self.account.backend.get_oauth_headers(STREAMING_URL))

            enabled, host, port, type = self.proxy_controller.backend.get_values()

            if enabled:
                curl.setopt(pycurl.PROXY, host.encode(ENCODING))

                curl.setopt(pycurl.PROXYPORT, port)

                curl.setopt(pycurl.PROXYTYPE, PROXIES[type])

            curl.setopt(curl.CONNECTTIMEOUT, 5)

            curl.setopt(curl.NOSIGNAL, 1)

            curl.setopt(curl.NOPROGRESS, 0)

            curl.setopt(curl.PROGRESSFUNCTION, self._callback)

            curl.setopt(curl.HEADERFUNCTION, self._process_header)

            curl.setopt(curl.WRITEFUNCTION, self._process_body)

            self.dead = False

            self.connected = False

            self.timestamp = time()

            self.buffer_thread = None

            self.account.backend.acquire_connecting()

            try:
                curl.perform()
            except pycurl.error as error:
                if error[0] == pycurl.E_ABORTED_BY_CALLBACK:
                    if self.dead:
                        print u'DEBUG: {} disconnected because of timeout'.format(self.account.get_user_nick())

                        self.success = -1

                        self.timeout = 0.25
                    else:
                        print u'DEBUG: {} disconnected successfully'.format(self.account.get_user_nick())

                        self.success = 1
                elif error[0] == pycurl.E_WRITE_ERROR:
                    print u'DEBUG: {} disconnected because of invalid format'.format(self.account.get_user_nick())

                    self.success = 0

                    self.timeout = float(randint(20, 40))
                else:
                    print u'DEBUG: {} disconnected because of curl error {}'.format(self.account.get_user_nick(), error[0])

                    if self.success == -1:
                        self.timeout = min(self.timeout + 0.25, 16.0)
                    else:
                        self.success = -1

                        self.timeout = 0.25
            else:
                print u'DEBUG: {} disconnected because of http error {}'.format(self.account.get_user_nick(), self.status)

                if self.status == 420:
                    self.success = 0

                    self.timeout = 3.0 * randint(240, 300)
                else:
                    if self.success == 0:
                        self.timeout = min(2.0 * self.timeout, float(randint(240, 300)))
                    else:
                        self.success = 0

                        self.timeout = float(randint(20, 40))
            finally:
                try:
                    curl.close()
                except pycurl.error:
                    pass

            if self.buffer_thread is not None:
                self.buffer_condition.acquire()

                self.buffer_finished = True

                self.buffer_condition.notify()

                self.buffer_condition.release()

                self.buffer_thread.join()

            if self.connected:
                self.account.identifier.set_spinning(True)
            else:
                self.account.backend.release_connecting()

            if self.success == 1:
                break
            else:
                self.streaming_condition.acquire()

                if not self.streaming_finished:
                    self._acquire_update()

                    self.broken_notifier.notify()

                    if self.success == -1:
                        if self.timeout > 15.9:
                            self.backup_notifier.notify()
                    else:
                        if self.timeout > 239.0:
                            self.backup_notifier.notify()

                    self._release_update()

                    print u'DEBUG: {} will reconnect in {} seconds'.format(self.account.get_user_nick(), self.timeout)

                    self.streaming_condition.wait(self.timeout)

                streaming_finished = self.streaming_finished

                self.streaming_condition.release()

                if streaming_finished:
                    break

        self.account.identifier.set_spinning(False)


    def _buffer_loop(self):
        self.account.shelf.clear()

        subsuccess = 1

        subtimeout = 0.0

        while True:
            try:
                data = self.account.backend.get_user_ids_blocking()
            except ConnectionError as error:
                if subsuccess == -1:
                    subtimeout = min(subtimeout + 0.25, 16.0)

                    if subtimeout > 15.9:
                        self.status_bar.popup(error.message)
                else:
                    subsuccess = -1

                    subtimeout = 0.25
            except RateError as error:
                subsuccess = 0

                subtimeout = 3.0 * randint(240, 300)

                self.status_bar.popup(error.message)
            except Error as error:
                if subsuccess == 0:
                    subtimeout = min(2.0 * subtimeout, float(randint(240, 300)))

                    if subtimeout > 239.0:
                        self.status_bar.popup(error.message)
                else:
                    subsuccess = 0

                    subtimeout = float(randint(20, 40))
            else:
                for subdata in data:
                    self.account.shelf.set_blocking(subdata)

                del data[:]

                break

            self.buffer_condition.acquire()

            if not self.buffer_finished:
                self.buffer_condition.wait(subtimeout)

            buffer_finished = self.buffer_finished

            self.buffer_condition.release()

            if buffer_finished:
                break

        buffer_starting = True

        while True:
            self.buffer_condition.acquire()

            while not self.buffer_finished and not self.queue:
                self.buffer_condition.wait()

            buffer_finished = self.buffer_finished

            if buffer_finished:
                self.open = False

                self.queue.clear()
            else:
                data, timestamp = self.queue.popleft()

                if buffer_starting:
                    for subdata in data:
                        self.account.shelf.set_friend(subdata)

                    del data[:]

                    self.open = True

                    self.account.identifier.set_spinning(False)

                    self.account.backend.release_connecting()

                    self.connected = True

                    print u'DEBUG: {} connected successfully'.format(self.account.get_user_nick())

                    self._acquire_update()

                    self.backup_notifier.notify()

                    self._release_update()

                    buffer_starting = False
                else:
                    try:
                        keys = data.viewkeys()

                        if u'limit' in keys:
                            pass
                        elif u'scrub_geo' in keys:
                            pass
                        elif EVENT_KEY in keys:
                            event = data.pop(EVENT_KEY)

                            if event == u'user_update':
                                target_user_data = self.account.backend.process_user(data.pop(TARGET_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(target_user_data)
                            elif event == u'block':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                target_user_data = self.account.backend.process_user(data.pop(TARGET_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(target_user_data)

                                self.account.shelf.set_blocking(target_user_data[ID_KEY])
                            elif event == u'unblock':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                target_user_data = self.account.backend.process_user(data.pop(TARGET_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(target_user_data)

                                self.account.shelf.unset(target_user_data[ID_KEY])

                                self.account.backend.relationship_backend_cache.unset_blocking(target_user_data[ID_KEY])
                            elif event == u'follow':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                target_user_data = self.account.backend.process_user(data.pop(TARGET_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(target_user_data)

                                if source_user_data[user.OWN_KEY]:
                                    self.account.shelf.set_friend(target_user_data[ID_KEY])

                                    self.account.backend.relationship_backend_cache.set_friend(target_user_data[ID_KEY])
                            elif event == UNFOLLOW_EVENT:
                                self.account.shelf.unset(data[TARGET_USER_KEY][ID_KEY])
                            elif event == u'favorite':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                # NOTE: The tweet included in this event does not
                                # contain entities, so it cannot be used directly.

                                target_user_data = self.account.backend.process_user(data.pop(TARGET_USER_KEY), timestamp)

                                if source_user_data[ID_KEY] != target_user_data[ID_KEY]:
                                    self.account.backend.user_backend_cache.set_data(target_user_data)

                                if source_user_data[user.OWN_KEY]:
                                    self.account.backend.tweet_backend_cache.set_value(data[TARGET_ITEM_KEY][ID_KEY], tweet.FAVORITED_KEY, True)
                            elif event == u'unfavorite':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                # NOTE: The tweet included in this event does not
                                # contain entities, so it cannot be used directly.

                                target_user_data = self.account.backend.process_user(data.pop(TARGET_USER_KEY), timestamp)

                                if source_user_data[ID_KEY] != target_user_data[ID_KEY]:
                                    self.account.backend.user_backend_cache.set_data(target_user_data)

                                if source_user_data[user.OWN_KEY]:
                                    tweet_id = data[TARGET_ITEM_KEY][ID_KEY]

                                    self.account.backend.notify_delete_tweet_favorite(tweet_id)

                                    self.account.backend.tweet_backend_cache.set_value(tweet_id, tweet.FAVORITED_KEY, False)
                            elif event == u'list_created':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)
                            elif event == u'list_updated':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                target_item_data = self.account.backend.process_group(data.pop(TARGET_ITEM_KEY), timestamp)

                                self.account.backend.group_backend_cache.set_data(target_item_data[1])
                            elif event == u'list_destroyed':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                self.account.backend.notify_delete_group(data[TARGET_ITEM_KEY][ID_KEY])
                            elif event == u'list_member_added':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                target_item_data = self.account.backend.process_group(data.pop(TARGET_ITEM_KEY), timestamp)

                                if source_user_data[ID_KEY] != target_item_data[0][ID_KEY]:
                                    self.account.backend.user_backend_cache.set_data(target_item_data[0])

                                self.account.backend.group_backend_cache.set_data(target_item_data[1])
                            elif event == u'list_member_removed':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                target_item_data = self.account.backend.process_group(data.pop(TARGET_ITEM_KEY), timestamp)

                                if source_user_data[ID_KEY] != target_item_data[0][ID_KEY]:
                                    self.account.backend.user_backend_cache.set_data(target_item_data[0])

                                self.account.backend.group_backend_cache.set_data(target_item_data[1])
                            elif event == u'list_user_subscribed':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                target_item_data = self.account.backend.process_group(data.pop(TARGET_ITEM_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(target_item_data[0])

                                if source_user_data[user.OWN_KEY]:
                                    self.account.backend.group_backend_cache.set_value(target_item_data[1][ID_KEY], group.SUBSCRIBED_KEY, True)

                                self.account.backend.group_backend_cache.set_data(target_item_data[1])
                            elif event == u'list_user_unsubscribed':
                                source_user_data = self.account.backend.process_user(data.pop(SOURCE_USER_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(source_user_data)

                                target_item_data = self.account.backend.process_group(data.pop(TARGET_ITEM_KEY), timestamp)

                                self.account.backend.user_backend_cache.set_data(target_item_data[0])

                                if source_user_data[user.OWN_KEY]:
                                    self.account.backend.group_backend_cache.set_value(target_item_data[1][ID_KEY], group.SUBSCRIBED_KEY, False)

                                self.account.backend.group_backend_cache.set_data(target_item_data[1])
                        elif DELETE_KEY in keys:
                            data = data.pop(DELETE_KEY)

                            keys = data.viewkeys()

                            if MESSAGE_KEY in keys:
                                self.account.backend.notify_delete_message(data[MESSAGE_KEY][ID_KEY])
                            elif TWEET_KEY in keys:
                                deleted_id = data[TWEET_KEY][ID_KEY]

                                self.account.backend.notify_delete_tweet(deleted_id)

                                self.account.backend.tweet_backend_cache.unset_retweet_id(deleted_id)
                        elif MESSAGE_KEY in keys:
                            data = self.account.backend.process_message(data[MESSAGE_KEY], timestamp)

                            self._acquire_update()

                            self.new_data = data

                            self.new_message_all_notifier.notify()

                            if data[1][user.OWN_KEY]:
                                self.new_message_received_notifier.notify()

                            if data[2][user.OWN_KEY]:
                                self.new_message_sent_notifier.notify()

                            self._release_update()
                        else:
                            data = self.account.backend.process_tweet(data, timestamp)

                            if data[1] is None:
                                if data[2][user.OWN_KEY]:
                                    user_id = self.account.backend.user_backend.id

                                    mention_also = False

                                    try:
                                        for mention in data[3][ENTITIES_KEY][MENTIONS_KEY]:
                                            try:
                                                if user_id == mention[ID_KEY]:
                                                    mention_also = True

                                                    break
                                            except (TypeError, KeyError):
                                                pass
                                    except (TypeError, KeyError):
                                        pass

                                    self.account.database.insert([data])

                                    self._acquire_update()

                                    self.new_data = data

                                    self.new_tweet_home_mentions_notifier.notify()

                                    self.new_tweet_home_notifier.notify()

                                    if mention_also:
                                        self.new_tweet_mentions_notifier.notify()

                                    self.new_tweet_sent_by_me_notifier.notify()

                                    self._release_update()
                                else:
                                    level = self.account.shelf.get(data[2][ID_KEY])

                                    if level == 1:
                                        self._acquire_update()

                                        self.new_data = data

                                        self.new_tweet_home_mentions_notifier.notify()

                                        self.new_tweet_mentions_notifier.notify()

                                        self._release_update()
                                    elif level == 2:
                                        user_id = self.account.backend.user_backend.id

                                        mention_also = False

                                        try:
                                            for mention in data[3][ENTITIES_KEY][MENTIONS_KEY]:
                                                try:
                                                    if user_id == mention[ID_KEY]:
                                                        mention_also = True

                                                        break
                                                except (TypeError, KeyError):
                                                    pass
                                        except (TypeError, KeyError):
                                            pass

                                        self.account.database.insert([data])

                                        self._acquire_update()

                                        self.new_data = data

                                        self.new_tweet_home_mentions_notifier.notify()

                                        self.new_tweet_home_notifier.notify()

                                        if mention_also:
                                            self.new_tweet_mentions_notifier.notify()

                                        self._release_update()
                            else:
                                if data[1][user.OWN_KEY]:
                                    not_following = (self.account.shelf.get(data[2][ID_KEY]) == 1)

                                    self._acquire_update()

                                    self.new_data = data

                                    if not_following:
                                        self.new_tweet_home_mentions_notifier.notify()

                                        self.new_tweet_home_notifier.notify()

                                    self.new_tweet_retweeted_by_me_notifier.notify()

                                    self._release_update()

                                    self.account.backend.tweet_backend_cache.set_value(data[3][ID_KEY], tweet.RETWEET_ID_KEY, data[0])
                                else:
                                    if self.account.shelf.get(data[1][ID_KEY]) == 2 and not data[2][user.OWN_KEY]:
                                        not_following = (self.account.shelf.get(data[2][ID_KEY]) == 1)

                                        if not_following:
                                            self.account.database.insert([data])

                                        self._acquire_update()

                                        self.new_data = data

                                        if not_following:
                                            self.new_tweet_home_mentions_notifier.notify()

                                            self.new_tweet_home_notifier.notify()

                                        self.new_tweet_retweeted_to_me_notifier.notify()

                                        self._release_update()
                    except (AttributeError, TypeError, KeyError):
                        pass

            self.buffer_condition.release()

            if buffer_finished:
                break


    def __init__(self, account, proxy_controller, status_bar):
        super(Producer, self).__init__()

        self.account = account

        self.proxy_controller = proxy_controller

        self.status_bar = status_bar

        self.open = False

        self.success = 1

        self.timeout = 0.0

        self.pattern = u'[\s]+'

        self.queue = deque()

        self.streaming_condition = Condition()

        self.buffer_condition = Condition()

        self.broken_notifier = Notifier()

        self.backup_notifier = Notifier()

        self.new_message_all_notifier = Notifier()

        self.new_message_received_notifier = Notifier()

        self.new_message_sent_notifier = Notifier()

        self.new_tweet_home_mentions_notifier = Notifier()

        self.new_tweet_home_notifier = Notifier()

        self.new_tweet_mentions_notifier = Notifier()

        self.new_tweet_sent_by_me_notifier = Notifier()

        self.new_tweet_retweeted_by_me_notifier = Notifier()

        self.new_tweet_retweeted_to_me_notifier = Notifier()

        self.delete_follow_uid = self.account.backend.delete_user_follow_notifier.connect(self._on_delete_follow)


    def isolate(self):
        self.account.backend.delete_user_follow_notifier.disconnect(self.delete_follow_uid)


    def start(self):
        self.streaming_finished = False

        self.streaming_thread = dispatch(self._streaming_loop)


    def stop(self):
        self.streaming_condition.acquire()

        self.streaming_finished = True

        self.streaming_condition.notify()

        self.streaming_condition.release()

        self.streaming_thread.join()
